# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


"""Tasks to migrate, validate, query"""

from oslo_config import cfg

from pypowervm import adapter as adpt
import pypowervm.const as c
import pypowervm.exceptions as pexc
import pypowervm.log as lgc
from pypowervm.wrappers import job
import pypowervm.wrappers.managed_system as wsys
import six

CONF = cfg.CONF
CONF.import_opt('pypowervm_job_request_timeout', 'pypowervm.wrappers.job')

_SUFFIX_PARM_MIGRATION = 'InternalMigrate'
_OPERATION_VALIDATION = "validation"
_OPERATION_MIGRATION = "migration"
_OPERATION_RECOVERY = "recovery"
_OPERATION_QUERY = "query"
_OPERATION_ABORT = "abort"

JOB_INPUT_PARAM_OPERATION = "operation"
JOB_INPUT_PARAM_INPUTXML = "inputXML"
JOB_OUTPUT_PARAM_OUTPUTXML = "outputXML"


@lgc.logcall
def validate(xml_data, add_parms=None):
    """Pass validation XML to target system
    :param xml_data: migration XML data from the source system
    :param add_parms: dict of parameters to pass directly to the job template
    """
    return _migration_task(xml_data, _OPERATION_VALIDATION,
                           add_parms=add_parms)


def migrate(xml_data, add_parms=None):
    """Pass migration XML to target system
    :param xml_data: migration XML data from the source system
    :param add_parms: dict of parameters to pass directly to the job template
    """
    return _migration_task(xml_data, _OPERATION_MIGRATION,
                           add_parms=add_parms)


def recovery(xml_data, add_parms=None):
    """Pass recovery XML to target system
    :param xml_data: recovery XML data from the source system
    :param add_parms: dict of parameters to pass directly to the job template
    """
    return _migration_task(xml_data, _OPERATION_RECOVERY,
                           add_parms=add_parms)


def abort(xml_data, add_parms=None):
    """Pass abort XML to target system
    :param xml_data: abort XML data from the source system
    :param add_parms: dict of parameters to pass directly to the job template
    """
    return _migration_task(xml_data, _OPERATION_ABORT,
                           add_parms=add_parms)


def query(xml_data, add_parms=None):
    """Pass query XML to target system
    :param xml_data: query XML data from the source system
    :param add_parms: dict of parameters to pass directly to the job template
    """
    return _migration_task(xml_data, _OPERATION_QUERY,
                           add_parms=add_parms)


def _process_migration_result(result):
    """Processes the Output XML returned by migration.
    :return xml_resp: XML output from migration
    """
    if result is None:
        return None

    xml_resp = result.get(JOB_OUTPUT_PARAM_OUTPUTXML)
    # If still none, nothing to do.
    return xml_resp


def get_uuid():
    session_adapter = adpt.Adapter()
    sys_list = session_adapter.read(wsys.System.schema_type)

    managed_sys = wsys.System.wrap(sys_list)[0]
    uuid = managed_sys.uuid
    return uuid, session_adapter


def _migration_task(xml_data, operation,
                    timeout=CONF.pypowervm_job_request_timeout,
                    add_parms=None):
    """Internal function to power on or off an instance.

    :param xml_data: String.  XML data for validation/migration
    :param operation: migration option - 'migration' or 'validation'.
    :param timeout: Value in seconds for specifying
                    how long to wait for the LPAR to stop
                    (for PowerOff suffix only)
    :param add_parms: dict of parameters to pass directly to the job template
    :return result: XML output from migration
    """

    uuid, adapter = get_uuid()
    complete = False
    while not complete:
        resp = adapter.read(wsys.System.schema_type, uuid,
                            suffix_type=c.SUFFIX_TYPE_DO,
                            suffix_parm=_SUFFIX_PARM_MIGRATION)
        job_wrapper = job.Job.wrap(resp.entry)
        job_parms = []

        job_parms.append(
            job_wrapper.create_job_parameter(JOB_INPUT_PARAM_OPERATION,
                                             operation))
        job_parms.append(
            job_wrapper.create_job_parameter(JOB_INPUT_PARAM_INPUTXML,
                                             xml_data))

        # Add add_parms to the job
        if add_parms is not None:
            for kw in add_parms.keys():
                job_parms.append(
                    job_wrapper.create_job_parameter(
                        kw, str(add_parms[kw])))
        try:
            job_wrapper.run_job(uuid, job_parms=job_parms, timeout=timeout)
            complete = True
            job_result = job_wrapper.get_job_results_as_dict()
            result = _process_migration_result(job_result)
            return result
        except pexc.JobRequestTimedOut as error:
            emsg = six.text_type(error)
            raise pexc.MigrationFailed(error=emsg)
        except pexc.JobRequestFailed as error:
            emsg = six.text_type(error)
            raise pexc.MigrationFailed(error=emsg)
