# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import argparse
import getpass
import logging
import sys
from migtools import migration
import pypowervm.exceptions as pexc
from pvmcli.util.i18n import _
import six
import pvmcli.util.util as util


FORMAT = ("%%(asctime)s - %(user)s - %%(message)s" % {
    'user': getpass.getuser()})
LOG_FILE = '/tmp/migrremote.log'


def validate(xml_data):
    """Internal function for migration validation.

    :param xml_data: String.  XML data for validation/migration
    :return write the migration output XML to stdout
    """
    try:
        output_xml = migration.validate(xml_data, None)
    except pexc.MigrationFailed as error:
        util.print_(six.text_type(error))
        sys.exit(1)

    # write to stdout if there is valid data
    if output_xml is not None:
        sys.stdout.write(output_xml)
    else:
        sys.stdout.write("No output XML data available from target!")
        sys.exit(1)


def migrate(xml_data):
    """Internal function for migration.

    :param xml_data: String.  XML data for validation/migration
    :return write the migration output XML to stdout
    """
    try:
        output_xml = migration.migrate(xml_data, None)
    except pexc.MigrationFailed as error:
        util.print_(six.text_type(error))
        sys.exit(1)

    # write to stdout if there is valid data
    if output_xml is not None:
        sys.stdout.write(output_xml)
    else:
        sys.stdout.write("No output XML data available from target!")
        sys.exit(1)


def recovery(xml_data):
    """Internal function for migration.

    :param xml_data: String.  XML data for recovery
    :return write the recovery output XML to stdout
    """
    try:
        output_xml = migration.recovery(xml_data, None)
    except pexc.MigrationFailed as error:
        util.print_(six.text_type(error))
        sys.exit(1)

    # write to stdout if there is valid data
    if output_xml is not None:
        sys.stdout.write(output_xml)
    else:
        sys.stdout.write("No output XML data available from target!")
        sys.exit(1)


def abort(xml_data):
    """Internal function for migration.

    :param xml_data: String.  XML data for abort
    :return write the abort output XML to stdout
    """
    try:
        output_xml = migration.abort(xml_data, None)
    except pexc.MigrationFailed as error:
        util.print_(six.text_type(error))
        sys.exit(1)

    # write to stdout if there is valid data
    if output_xml is not None:
        sys.stdout.write(output_xml)
    else:
        sys.stdout.write("No output XML data available from target!")
        sys.exit(1)


def query(xml_data):
    """Internal function for migration.

    :param xml_data: String.  XML data for query
    :return write the query output XML to stdout
    """
    try:
        output_xml = migration.query(xml_data, None)
    except pexc.MigrationFailed as error:
        util.print_(six.text_type(error))
        sys.exit(1)

    # write to stdout if there is valid data
    if output_xml is not None:
        sys.stdout.write(output_xml)
    else:
        sys.stdout.write("No output XML data available from target!")
        sys.exit(1)

operation = {
    "v": validate,
    "m": migrate,
    "s": abort,
    "q": query,
    "r": recovery,
}


def invoke_operation(oper, args):
    # convert file content to string
    xml_data = args.f.read()

    return operation[oper](xml_data)


def create_parser():
    parser = argparse.ArgumentParser(prog="migrremote")

    # Target system name
    parser.add_argument("-m",
                        help=_("system name"))

    # Migration operation to run
    parser.add_argument("-o",
                        choices=["m", "q", "r", "s", "v"],
                        required=True)

    # Input XML file name, default is stdin
    parser.add_argument("-f",
                        type=argparse.FileType("r"),
                        default="-",
                        help=_("input/output data file name"),
                        required=False)

    # Migration output detail level
    parser.add_argument("-d",
                        help=_("detail level"))

    # Wait time
    parser.add_argument("-w",
                        help=_("wait time"))

    return parser


def main():
    # Set up migrremote logging to use a FileHandler to the log file
    # specified in migrremote bash file. Rotation of the file should be handled
    # on each call to the bash file.
    try:
        logging.basicConfig(level=logging.DEBUG,
                            filename=LOG_FILE,
                            format=FORMAT)
    except IOError:
        # Do not fail migration operation for IO errors
        # while setting up logging
        pass
    parser = create_parser()
    args = parser.parse_args()
    invoke_operation(args.o, args)

if __name__ == "__main__":
    main()
