# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

from abc import ABCMeta
import six

from pypowervm.tasks import power as pwr
from pypowervm.tasks import power_opts as pwr_opts
from pypowervm.wrappers.base_partition \
    import DedicatedSharingMode as ds_modes
from pypowervm.wrappers.base_partition import SharingMode as s_modes
from pypowervm.wrappers import base_partition as bp_w
from pypowervm.wrappers import logical_partition as lp_w
from pypowervm.wrappers import managed_system as ms_w
from pypowervm.wrappers import shared_proc_pool as spp
from pypowervm.wrappers import virtual_io_server as vios_w
from pypowervm.utils import lpar_builder as lb

import pvmcli.command.object_type as obj_type
import pvmcli.command.field as field
import pvmcli.command.io_slot as io_slot
import pvmcli.command.operation as op
import pvmcli.util.exceptions as e
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.output_processor as out_proc
import pvmcli.util.util as u


POWER_OFF_OP_NAME = 'power-off'
POWER_ON_OP_NAME = 'power-on'
RESTART_OP_NAME = 'restart'
DUMPRESTART_OP_NAME = 'dumprestart'

PARTITION_NAME = 'Partition'

PO_NAME = _("Name")
PO_ID = _("ID")
PO_STATE = _("State")
PO_RMC = _("RMC")
PO_ENV = _("Env")
PO_REFCODE = _("Ref Code")
PO_MEM = _("Mem")
PO_CPU = _("CPU")
PO_ENT = _("Ent")

POOL_IDS = {'name': field.Field('name'),
            'id': field.Field('id', field_type=int)}

DED_PROC = 'dedicated'
SHR_PROC = 'shared'

NOT_APP = "----"


class ShrModeArgs(object):
    """Processor sharing mode arguments. """
    NORMAL = 'normal'
    H_NORMAL = 'share_idle_procs'
    KEEP = 'keep'
    H_KEEP = 'keep_idle_procs'
    ACTIVE = 'active'
    H_ACTIVE = 'share_idle_procs_active'
    ALWAYS = 'always'
    H_ALWAYS = 'share_idle_procs_always'
    DED_VALUES = (NORMAL, KEEP, ACTIVE, ALWAYS,
                  H_NORMAL, H_KEEP, H_ACTIVE, H_ALWAYS)

    CAPPED = 'capped'
    H_CAPPED = 'cap'
    UNCAPPED = 'uncapped'
    H_UNCAPPED = 'uncap'
    SHR_VALUES = (CAPPED, UNCAPPED, H_CAPPED, H_UNCAPPED)


SHR_MODE_MAP = {
    ShrModeArgs.NORMAL: ds_modes.SHARE_IDLE_PROCS,
    ShrModeArgs.H_NORMAL: ds_modes.SHARE_IDLE_PROCS,
    ShrModeArgs.KEEP: ds_modes.KEEP_IDLE_PROCS,
    ShrModeArgs.H_KEEP: ds_modes.KEEP_IDLE_PROCS,
    ShrModeArgs.ACTIVE: ds_modes.SHARE_IDLE_PROCS_ACTIVE,
    ShrModeArgs.H_ACTIVE: ds_modes.SHARE_IDLE_PROCS_ACTIVE,
    ShrModeArgs.ALWAYS: ds_modes.SHARE_IDLE_PROCS_ALWAYS,
    ShrModeArgs.H_ALWAYS: ds_modes.SHARE_IDLE_PROCS_ALWAYS,
    ShrModeArgs.CAPPED: s_modes.CAPPED,
    ShrModeArgs.H_CAPPED: s_modes.CAPPED,
    ShrModeArgs.UNCAPPED: s_modes.UNCAPPED,
    ShrModeArgs.H_UNCAPPED: s_modes.UNCAPPED}


def lpar_type(type_str):
    """Map any case LPAR type to its pypowervm wrapper equivalent.

    :param type_str: Any case version of 'aix/linux' or 'os400'
    :returns: The LPARType associated with type_str, or type_str if invalid."""
    if str.lower(type_str) == str.lower(bp_w.LPARType.AIXLINUX):
        return bp_w.LPARType.AIXLINUX
    elif str.lower(type_str) == str.lower(bp_w.LPARType.OS400):
        return bp_w.LPARType.OS400
    else:
        return type_str


@six.add_metaclass(ABCMeta)
class BasePartition(obj_type.ObjectType):
    def __init__(self, child_disp_models=[], excluded_dfields=None):
        super(BasePartition, self).__init__()

        # Boolean determining if this partition type has an env field
        self.has_env = True

        bp_field_models = [bp_w.PartitionCapabilities,
                           bp_w.PartitionProcessorConfiguration,
                           bp_w.PartitionMemoryConfiguration,
                           bp_w.SharedProcessorConfiguration,
                           bp_w.DedicatedProcessorConfiguration,
                           bp_w.PartitionIOConfiguration,
                           bp_w.IOSlot,
                           bp_w.IOSlot.AssociatedIOSlot,
                           bp_w.IOAdapter,
                           bp_w.PhysFCAdapter,
                           bp_w.PhysFCPort,
                           bp_w.TaggedIO] + child_disp_models

        self.add_field_models(bp_field_models)

        self.exc_disp_fields.append('ProfileIOSlot.adapter')
        self.exc_disp_fields.append('SharedEthernetAdapter.is_primary')
        if excluded_dfields:
            self.exc_disp_fields.extend(excluded_dfields)

        d_op = self.add_display_operation()
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_class',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_subsys_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_subsys_vendor_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_vendor_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'drc_index',
                                  out_func=out_proc.hex_no_prefix)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(bp_w.IOSlot, 'pci_rev_id',
                                  out_func=out_proc.two_char_hex)
        d_op.set_d_field_out_func(bp_w.IOAdapter, 'id',
                                  out_func=out_proc.hex_no_prefix)
        self.add_update_operation()
        self.add_delete_operation()

        self._add_create_operation()
        self._add_power_off_operation()
        self._add_restart_operation()

        self.pri_df_labels = [
            PO_NAME, PO_ID, PO_STATE, PO_RMC, PO_REFCODE, PO_MEM, PO_CPU,
            PO_ENT]

    def update(self):
        self.validate_wrps_single_elem(self.ec.wrappers,
                                       self.name, op.UPDATE_NAME)
        p_set_fields = self.ec.parsed_args.set_fields

        if len(p_set_fields) == 0:
            raise e.InvalidSetFieldsError(object_type=self.name,
                                          operation=op.UPDATE_NAME)
        split_sfs = u.split_set_fields(p_set_fields)
        sm_sf = 'PartitionProcessorConfiguration.sharing_mode'
        if sm_sf in split_sfs:
            sf_val = self._determine_sharing_mode(
                split_sfs[sm_sf],
                DED_PROC if self.ec.wrappers[0].proc_config.has_dedicated
                else SHR_PROC)
            self.set_field_val(self.ec.wrappers[0], sm_sf, sf_val)
            split_sfs.pop(sm_sf)
        obj_type.update_retry(self.ec.wrappers[0], split_sfs, self)

    def _add_power_on_operation(self, has_ipl_src=False):
        bootstr_help = _("Specify bootstring parameter for the power-on job.")
        bootmode_help = _("Specify bootmode parameter for the power-on job.")
        keylock_help = _("Specify keylock parameter for the power-on job.")
        ipl_src_help = _("Specify IBMi IPLSource parameter for the power-on "
                         "job.")
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--bootstring', required=False,
                                 arg_help=bootstr_help, arg_type=str),
                     op.Argument('--bootmode', required=False,
                                 arg_help=bootmode_help, arg_type=str,
                                 choices=pwr.BootMode.ALL_VALUES),
                     op.Argument('--keylock', required=False,
                                 arg_help=keylock_help, arg_type=str,
                                 choices=pwr.KeylockPos.ALL_VALUES)])
        if has_ipl_src:
            args.append(op.Argument('--iplsource', required=False,
                                    arg_help=ipl_src_help, arg_type=str,
                                    choices=lp_w.IPLSrc.ALL_VALUES))
        self.add_operation(op.Operation(POWER_ON_OP_NAME, self.power_on,
                                        op_args=args,
                                        object_model=self.model))

    def _add_power_off_operation(self):
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--hard', action="store_true"),
                     op.Argument('--restart', action="store_true"),
                     op.Argument('--no-prompt', action="store_true",
                                 arg_help=help.NOPROMPT_HELP)])
        self.add_operation(op.Operation(POWER_OFF_OP_NAME, self.power_off,
                                        op_args=args,
                                        object_model=self.model))

    def _add_restart_operation(self):
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--no-prompt', action="store_true",
                                 arg_help=help.NOPROMPT_HELP)])
        self.add_operation(op.Operation(RESTART_OP_NAME, self.restart,
                                        op_args=args,
                                        object_model=self.model))

    def _add_create_operation(self):
        # Help text
        name_help = _("Required. The name of the partition.")
        type_help = _("Required. The type of partition. Options: "
                      "AIX/Linux, OS400")
        mem_help = _("Required. The amount of memory for the "
                     "partition in MB.")
        min_mem_help = _("The minimum amount of memory in MB. "
                         "Default: --mem parameter.")
        max_mem_help = _("The maximum amount of memory in MB. "
                         "Default: --mem parameter.")
        max_io_help = _("The number of IO slots to configure for the "
                        "partition. Default: 64.")
        proc_help = _("The number of dedicated/virtual processors, depending "
                      "on configuration.")
        mode_kwargs = {ShrModeArgs.NORMAL: ds_modes.SHARE_IDLE_PROCS,
                       ShrModeArgs.KEEP: ds_modes.KEEP_IDLE_PROCS,
                       ShrModeArgs.ACTIVE: ds_modes.SHARE_IDLE_PROCS_ACTIVE,
                       ShrModeArgs.ALWAYS: ds_modes.SHARE_IDLE_PROCS_ALWAYS,
                       ShrModeArgs.CAPPED: s_modes.CAPPED,
                       ShrModeArgs.UNCAPPED: s_modes.UNCAPPED}
        shr_mode_help = _("The processor sharing mode of the partition.\n"
                          "The valid options for this depend on the value "
                          "in the --proc-type argument. The choices for "
                          "those options are below.\n"
                          "Dedicated Options:\n"
                          "    normal (%(normal)s) (Default)\n"
                          "    keep (%(keep)s)\n"
                          "    active (%(active)s)\n"
                          "    always (%(always)s)\n"
                          "Shared Options:\n"
                          "    %(capped)s\n"
                          "    %(uncapped)s (Default)") % mode_kwargs
        min_proc_help = _("The minimum number of processors. Default: "
                          "--proc parameter.")
        max_proc_help = _("The maximum number of processors. Default: "
                          "--proc parameter.")
        proc_unit_help = _("The number of processing units for a shared "
                           "processor configuration.")
        min_punit_help = _("The minimum number of processors units "
                           "for a shared processor configuration. "
                           "Default: --proc-unit parameter")
        max_punit_help = _("The maximum number of processors units "
                           "for a shared processor configuration. "
                           "Default: --proc-unit parameter")
        unc_wgt_help = _("The uncapped weight of a shared processor "
                         "configuration. Default: 128.")
        proc_type_help = _("The processor configuration type to be assigned "
                           "to this partition. Required if processor config "
                           "XML is not provided.")
        drc_name_help = _("The names of I/O Adapters to be added to this "
                          "partition after creation. Must not be owned at "
                          "time of creation. To remove an IOSlot from an "
                          "existing partition, see the IOSlot detach help.")
        id_help = _("The positive integer ID of the partition. Must be "
                    "unique. Optional.")
        pool_help = _("field=value identifier search field to set the shared "
                      "processor pool on the created partition.\n"
                      "Examples: id=1\n"
                      "          name=spp_name")

        # Argument groups
        part_arg_grp = u.utf8_str(_('Partition arguments'))
        proc_cfg_grp = u.utf8_str(_('Processor Configuration arguments'))

        crt_args = o_g.default_args(self, parent_models=[ms_w.System])
        if self.model is lp_w.LPAR:
            crt_args.append(op.Argument('--type', arg_type=lpar_type,
                                        required=True,
                                        arg_help=help.wrap_help(type_help),
                                        choices=[bp_w.LPARType.OS400,
                                                 bp_w.LPARType.AIXLINUX],
                                        group=part_arg_grp))
        crt_args.extend([op.Argument('--name', arg_type=str, required=True,
                                     arg_help=help.wrap_help(name_help),
                                     group=part_arg_grp),
                         op.Argument('--id', arg_type=int, required=False,
                                     arg_help=help.wrap_help(id_help),
                                     group=part_arg_grp),
                         op.Argument('--mem', arg_type=str, required=True,
                                     arg_help=help.wrap_help(mem_help),
                                     group=part_arg_grp),
                         op.Argument('--min-mem', arg_type=str, required=False,
                                     arg_help=help.wrap_help(min_mem_help),
                                     group=part_arg_grp),
                         op.Argument('--max-mem', arg_type=str, required=False,
                                     arg_help=help.wrap_help(max_mem_help),
                                     group=part_arg_grp),
                         op.Argument('--max-io-slots', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(max_io_help),
                                     group=part_arg_grp),
                         op.Argument('--proc-type', arg_type=str,
                                     choices=[DED_PROC, SHR_PROC],
                                     required=True,
                                     arg_help=help.wrap_help(proc_type_help),
                                     group=proc_cfg_grp),
                         op.Argument('--proc', arg_type=str, required=True,
                                     arg_help=help.wrap_help(proc_help),
                                     group=proc_cfg_grp),
                         op.Argument('--sharing-mode', arg_type=str,
                                     required=True,
                                     arg_help=help.wrap_help(shr_mode_help),
                                     group=proc_cfg_grp),
                         op.Argument('--proc-unit', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(proc_unit_help),
                                     group=proc_cfg_grp),
                         op.Argument('--min-proc', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(min_proc_help),
                                     group=proc_cfg_grp),
                         op.Argument('--max-proc', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(max_proc_help),
                                     group=proc_cfg_grp),
                         op.Argument('--uncapped-weight', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(unc_wgt_help),
                                     group=proc_cfg_grp),
                         op.Argument('--min-proc-unit', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(min_punit_help),
                                     group=proc_cfg_grp),
                         op.Argument('--max-proc-unit', arg_type=str,
                                     required=False,
                                     arg_help=help.wrap_help(max_punit_help),
                                     group=proc_cfg_grp),
                         op.Argument('--drc-names', arg_type=str, nargs='+',
                                     required=False,
                                     arg_help=help.wrap_help(drc_name_help),
                                     group=part_arg_grp),
                         op.Argument('--pool', arg_type=str, required=False,
                                     arg_help=help.wrap_help(pool_help),
                                     group=proc_cfg_grp)])
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def power_on(self):
        ''' Power on Partition job processing function '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       POWER_ON_OP_NAME)

        name = self.ec.wrappers[0].name

        u.print_(_(
            "Powering on partition %s, this may take a few minutes.") % name)

        add_parms = {}

        bootstring = self.ec.parsed_args.bootstring
        bootmode = self.ec.parsed_args.bootmode
        keylock = self.ec.parsed_args.keylock
        iplsource = None
        if "iplsource" in dir(self.ec.parsed_args):
            iplsource = self.ec.parsed_args.iplsource

        if bootstring:
            add_parms['bootstring'] = bootstring
        if bootmode:
            add_parms[pwr.BootMode.KEY] = bootmode
        if iplsource:
            add_parms[lp_w.IPLSrc.KEY] = iplsource
        if keylock:
            add_parms[pwr.KeylockPos.KEY] = keylock

        pwr.power_on(self.ec.wrappers[0], self.ec.msys.uuid,
                     add_parms=add_parms)
        # If we don't get an exception back from pypowervm job monitoring,
        # we currently assume it was powered on properly.

        u.print_(_("Partition %s power-on successful.") % name)

    def power_off(self):
        ''' Power off Partition job processing function '''
        self.power_off_restart(POWER_OFF_OP_NAME)

    def restart(self):
        ''' Restart Partition job processing function '''
        self.power_off_restart(RESTART_OP_NAME)

    def _pwr_off_restart_op_valid(self, vios_op_query, op_query):
        # Power off and restart is valid if:
        #     1) No prompt argument is specified
        #     2) We're a non-management logical partition
        #     3) We're a VIOS and the user responds with Y to the prompt
        #     4) We're the management lpar and the user responds with Y
        return (self.ec.parsed_args.no_prompt
                or (not self.ec.wrappers[0].is_mgmt_partition
                    and isinstance(self.ec.wrappers[0], lp_w.LPAR))
                or (isinstance(self.ec.wrappers[0], vios_w.VIOS)
                    and u.input(vios_op_query).upper() == "Y")
                or (self.ec.wrappers[0].is_mgmt_partition
                    and u.input(op_query).upper() == "Y"))

    def power_off_restart(self, op_name):
        '''Perform a power off/restart job on a LPAR or VIOS.

        :param op_name: One of 'dumprestart', 'restart', or 'power-off'.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name, op_name)

        name = self.ec.wrappers[0].name

        op_query = _("Are you sure you wish to restart the management "
                     "partition (Y/N)?")
        vios_op_query = _("Restarting the VIOS will leave running "
                          "partitions without access to their storage "
                          "and/or network unless you have a redundant "
                          "VIOS that is up and running. Are you sure you "
                          "wish to restart the VIOS (Y/N)?")
        if op_name == POWER_OFF_OP_NAME:
            op_query = _("Powering off the management partition will leave "
                         "you without control of your system. You will need "
                         "to restart the entire system to get back control. "
                         "Are you sure you wish to power off the management "
                         "partition (Y/N)?")
            vios_op_query = _("Powering off the VIOS will leave running "
                              "partitions without access to their storage "
                              "and/or network unless you have a redundant "
                              "VIOS that is up and running. Are you sure you "
                              "wish to power off the VIOS (Y/N)?")

        op_valid = self._pwr_off_restart_op_valid(vios_op_query, op_query)

        if op_valid:
            if op_name == POWER_OFF_OP_NAME:
                u.print_(
                    _("Powering off partition %s, this may take a few "
                      "minutes.") % name)
            else:
                u.print_(
                    _("Restarting partition %s, this may take a few "
                      "minutes.") % name)

            power_off_opts = pwr_opts.PowerOffOpts().soft_detect(
                self.ec.wrappers[0])
            prs = op_name == POWER_OFF_OP_NAME and self.ec.parsed_args.restart
            if op_name == RESTART_OP_NAME or prs:
                power_off_opts = power_off_opts.restart()
            if op_name == POWER_OFF_OP_NAME and self.ec.parsed_args.hard:
                power_off_opts = power_off_opts.vsp_hard()
            if op_name == DUMPRESTART_OP_NAME:
                power_off_opts = power_off_opts.operation(
                    pwr_opts.PowerOffOperation.DUMPRESTART)

            pwr.PowerOp.stop(self.ec.wrappers[0], power_off_opts)
            # If we don't get an exception back from pypowervm job monitoring,
            # we currently assume our op was successful.
            u.print_(_(
                "Partition %(name)s %(op)s successful.") % {'name': name,
                                                            'op': op_name})
        else:
            u.print_(_("User cancelled %s operation.") % op_name)

    def _set_pool(self, l_attrs):
        if self.ec.parsed_args.pool:
            pool_arg = u.split_eq_arg(self.ec.parsed_args.pool, POOL_IDS)
            if 'name' in pool_arg:
                found_pool = spp.SharedProcPool.search(
                    self.ec.sm.adapter, parent_type=ms_w.System.schema_type,
                    parent_uuid=self.ec.msys.uuid, name=pool_arg['name'])
                if found_pool:
                    l_attrs[lb.SPP] = found_pool[0].id
                else:
                    raise e.InvalidIDArgumentError(
                        id_arg="name", valid_fields=", ".join(POOL_IDS))
            else:
                l_attrs[lb.SPP] = pool_arg['id']

    def create(self):
        ''' Create operation processing function '''
        l_attrs = {}

        # Build proc configuration based on specified args.
        l_attrs[lb.VCPU] = self.ec.parsed_args.proc

        l_attrs[lb.SHARING_MODE] = self._determine_sharing_mode(
            self.ec.parsed_args.sharing_mode, self.ec.parsed_args.proc_type)

        if self.ec.parsed_args.min_proc:
            l_attrs[lb.MIN_VCPU] = self.ec.parsed_args.min_proc

        if self.ec.parsed_args.max_proc:
            l_attrs[lb.MAX_VCPU] = self.ec.parsed_args.max_proc

        if self.ec.parsed_args.proc_type == SHR_PROC:
            if self.ec.parsed_args.proc_unit is None:
                raise e.MissingArgumentError(arg="--proc-unit")
            l_attrs[lb.PROC_UNITS] = self.ec.parsed_args.proc_unit

            if self.ec.parsed_args.min_proc_unit:
                l_attrs[lb.MIN_PROC_U] = self.ec.parsed_args.min_proc_unit

            if self.ec.parsed_args.max_proc_unit:
                l_attrs[lb.MAX_PROC_U] = self.ec.parsed_args.max_proc_unit

            p_u_weight = self.ec.parsed_args.uncapped_weight
            if p_u_weight:
                l_attrs[lb.UNCAPPED_WEIGHT] = p_u_weight

        if self.ec.parsed_args.id:
            l_attrs[lb.ID] = self.ec.parsed_args.id

        l_attrs[lb.NAME] = self.ec.parsed_args.name

        if self.model is lp_w.LPAR:
            if self.ec.parsed_args.type is None:
                raise e.MissingArgumentError(arg="--type")
            l_attrs[lb.ENV] = self.ec.parsed_args.type
        else:
            l_attrs[lb.ENV] = bp_w.LPARType.VIOS

        l_attrs[lb.MEM] = self.ec.parsed_args.mem

        if self.ec.parsed_args.min_mem:
            l_attrs[lb.MIN_MEM] = self.ec.parsed_args.min_mem

        if self.ec.parsed_args.max_mem:
            l_attrs[lb.MAX_MEM] = self.ec.parsed_args.max_mem

        if self.ec.parsed_args.max_io_slots:
            l_attrs[lb.MAX_IO_SLOTS] = self.ec.parsed_args.max_io_slots

        self._set_pool(l_attrs)

        stdz = lb.DefaultStandardize(self.ec.msys)
        part_bldr = lb.LPARBuilder(self.ec.sm.adapter, l_attrs, stdz)
        new_part = part_bldr.build().create(parent=self.ec.msys)

        if self.ec.parsed_args.drc_names:
            drc_names = self.ec.parsed_args.drc_names
            io_slot.add_io_slots(self, new_part,
                                 [self.ec.get_msys_io_slot(drc_name)
                                  for drc_name in drc_names])

        self.ec.wrappers = [new_part]
        # Only display if a user specifies specific display fields
        if (self.ec.parsed_args.display_fields
                or self.ec.parsed_args.all_out):
            self.display()

    def _validate_shr_mode(self, shr_mode, proc_type):
        # Fail if user specifies a dedicated proc and an invalid share mode
        not_in_ded = shr_mode not in ShrModeArgs.DED_VALUES
        if proc_type == DED_PROC and not_in_ded:
            shr_modes = ', '.join(ShrModeArgs.DED_VALUES)
            raise e.InvalidSharingModeError(
                proc_type=DED_PROC, shr_modes=shr_modes)

        # Fail if user specifies a shared proc and an invalid share mode
        not_in_shr = shr_mode not in ShrModeArgs.SHR_VALUES
        if proc_type == SHR_PROC and not_in_shr:
            shr_modes = ', '.join(ShrModeArgs.SHR_VALUES)
            raise e.InvalidSharingModeError(
                proc_type=SHR_PROC, shr_modes=shr_modes)

    def _determine_sharing_mode(self, shr_mode, proc_type):
        if shr_mode is None:
            if proc_type == DED_PROC:
                return ds_modes.SHARE_IDLE_PROCS
            else:
                return s_modes.UNCAPPED
        self._validate_shr_mode(shr_mode, proc_type)
        return SHR_MODE_MAP[shr_mode] if shr_mode in SHR_MODE_MAP else shr_mode

    def get_out_rows(self):
        out_rows = []
        for wrp in sorted(self.ec.wrappers, key=lambda w: w.id):
            rmc_st = NOT_APP if wrp.is_mgmt_partition else six.text_type(
                wrp.rmc_state)
            out_row = [wrp.name if wrp.name else '',
                       six.text_type(wrp.id), six.text_type(wrp.state), rmc_st]
            if self.has_env:
                out_row.append(six.text_type(wrp.env))
            out_row.extend([six.text_type(
                wrp.ref_code) if wrp.ref_code else '', six.text_type(
                    wrp.mem_config.desired)])
            if wrp.proc_config.has_dedicated:
                out_row.extend([
                    six.text_type(
                        wrp.proc_config.dedicated_proc_cfg.desired), ''])
            else:
                des_virt = wrp.proc_config.shared_proc_cfg.desired_virtual
                des_unit = wrp.proc_config.shared_proc_cfg.desired_units
                out_row.extend([six.text_type(des_virt),
                                six.text_type(des_unit)])
            out_rows.append(out_row)

        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(self.pri_df_labels,
                                             self.get_out_rows())
        return "%(title)s\n%(tbl)s" % {'title': _(self.title),
                                       'tbl': po_table.get_string()}
