# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import logging

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_ioslot as help_ioslot
import pvmcli.util.exceptions as e
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.output_processor as out_proc

import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.logical_partition as lpar_w
import pypowervm.wrappers.managed_system as msys_w
import pypowervm.wrappers.virtual_io_server as vios_w


PRI_DISP_FIELD_LABELS = [_("Loc Code (DRC Name)"), _("LPAR"), _("Description")]


OP_ATTACH = 'attach'
OP_DETACH = 'detach'


def add_io_slots(obj_type, part, new_ios):
    ''' Add the provided I/O Slots to a given partition.

    :param obj_type: The Object Type facilitating the session management.
    :param part: The partition to receive the I/O Slots.
    :param new_ios: List of I/O slots to add to the given partition.
    '''
    adp = obj_type.ec.sm.adapter

    if new_ios:
        new_slot_ids = []

        for new_io in new_ios:
            new_slot = bp_w.IOSlot.bld(adp, new_io.bus_grp_required,
                                       new_io.drc_index)
            part.io_config.io_slots.append(new_slot)
            new_slot_ids.append(new_io.drc_index)

        upd_lpar = part.update()
        out_slots = []
        for io_slot in upd_lpar.io_config.io_slots:
            if io_slot.drc_index in new_slot_ids:
                out_slots.append(io_slot)
        return out_slots
    else:
        raise e.NoEntriesFoundError(object_type=msys_w.IOSlot.schema_type)


class IOSlot(ObjectType):
    def __init__(self):
        super(IOSlot, self).__init__()

        self.field_models = [msys_w.IOSlot]

        self.exc_disp_fields.extend(
            ['IOSlot.dyn_reconfig_conn_index', 'IOSlot.dyn_reconfig_conn_name',
             'IOSlot.pci_revision_id', 'IOSlot.pci_sub_dev_id',
             'IOSlot.pci_sub_vendor_id'])

        list_elem_wrp_funcs = [msys_w.System.asio_config,
                               msys_w.ASIOConfig.io_slots]
        d_op = self.add_display_operation(model=msys_w.IOSlot,
                                          parent_models=[msys_w.System],
                                          elem_wrp_funcs=list_elem_wrp_funcs)

        # Set appropriate fields to output hex values
        d_op.set_d_field_out_func(msys_w.IOSlot, 'drc_index',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_class',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_rev_id',
                                  out_func=out_proc.two_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_subsys_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_subsys_vendor_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(msys_w.IOSlot, 'pci_vendor_id',
                                  out_func=out_proc.four_char_hex)

        elem_wrp_funcs = [lpar_w.LPAR.io_config,
                          bp_w.PartitionIOConfiguration.io_slots]

        # Add detach operation
        det_name_help = _("The names of I/O Adapters to be removed from the "
                          "specified partition.")
        det_args = o_g.default_args(self,
                                    parent_models=[lpar_w.LPAR, vios_w.VIOS],
                                    include_output=False)
        det_args.extend([op.Argument('--drc-names', arg_type=str,
                                     required=True, nargs='+',
                                     arg_help=help.wrap_help(det_name_help))])
        self.add_operation(op.Operation('detach', self.detach,
                                        op_args=det_args,
                                        object_model=self.model,
                                        parent_models=[lpar_w.LPAR,
                                                       vios_w.VIOS],
                                        elem_wrp_funcs=elem_wrp_funcs))

        # Add attach operation
        att_name_help = _("The names of I/O Adapters to be attached to the "
                          "specified parent partition. ")
        att_args = o_g.default_args(self,
                                    parent_models=[lpar_w.LPAR, vios_w.VIOS],
                                    include_output=False)
        att_args.extend([op.Argument('--drc-names', arg_type=str,
                                     required=True, nargs='+',
                                     arg_help=help.wrap_help(att_name_help))])
        self.add_operation(op.Operation(OP_ATTACH, self.attach,
                                        op_args=att_args,
                                        object_model=self.model,
                                        parent_models=[lpar_w.LPAR,
                                                       vios_w.VIOS],
                                        elem_wrp_funcs=elem_wrp_funcs))

        self.setup_arguments()

    def ioslot_out_rows(self):
        drc_name_to_lpar = {}
        all_lpars = lpar_w.LPAR.get(self.ec.sm.adapter)
        all_lpars.extend(vios_w.VIOS.get(self.ec.sm.adapter))
        for lpar in all_lpars:
            for io_slot in lpar.io_config.io_slots:
                drc_name_to_lpar[io_slot.drc_name] = lpar

        rows = []
        name_to_ioslot = {}
        for wrp in self.ec.wrappers:
            name_to_ioslot[wrp.drc_name] = wrp
        for drc_name in sorted(name_to_ioslot):
            wrp = name_to_ioslot[drc_name]
            lpar_name = (drc_name_to_lpar[drc_name].name
                         if drc_name in drc_name_to_lpar else "")
            rows.append([drc_name, lpar_name, wrp.description])
        return rows

    def attach(self):
        self.validate_wrps_single_elem(self.ec.elem_p_wrappers,
                                       lpar_w.LPAR.schema_type, OP_ATTACH)

        slots_to_add = []
        for drc_name in self.ec.parsed_args.drc_names:
            slot_to_add = self.ec.get_msys_io_slot(drc_name)
            if not slot_to_add:
                raise e.InvalidDrcNameError(name=drc_name)
            else:
                slots_to_add.append(slot_to_add)
        add_io_slots(self, self.ec.elem_p_wrappers[0], slots_to_add)

    def detach(self):
        ''' Custom detach function for IOSlot ObjectType. '''
        self.validate_wrps_single_elem(self.ec.elem_p_wrappers,
                                       lpar_w.LPAR.schema_type, OP_DETACH)

        sys_io_slots = []
        for drc_name in self.ec.parsed_args.drc_names:
            sys_io_slot = self.ec.get_msys_io_slot(drc_name)
            if sys_io_slot:
                sys_io_slots.append(sys_io_slot.drc_index)
            else:
                logging.getLogger('pvmctl').error(
                    "Could not find IO Slot with DRC name %s.", drc_name)

        if len(sys_io_slots) == 0:
            raise e.NoEntriesFoundError(object_type=self.name)

        part_wrp = self.ec.elem_p_wrappers[0]
        new_io_slot_list = []
        for io_slot in part_wrp.io_config.io_slots:
            if io_slot.drc_index not in sys_io_slots:
                new_io_slot_list.append(io_slot)

        if len(new_io_slot_list) == len(part_wrp.io_config.io_slots):
            raise e.NoEntriesFoundError(object_type=self.name)

        part_wrp.io_config.io_slots = new_io_slot_list
        part_wrp.update()

    @property
    def model(self):
        return bp_w.IOSlot

    @property
    def name(self):
        return msys_w.IOSlot.schema_type

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.ioslot_out_rows())

        # Left align description column
        po_table.align["Description"] = "l"
        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("I/O Slots"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_ioslot.help_text
