# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

from oslo_config import cfg

import argparse
import pvmcli.command.base_partition as bp_c
import pvmcli.command.operation as op
import pvmcli.help.help_lpar as help_lpar
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g

import pypowervm.tasks.ibmi as ibmi_task
import pypowervm.tasks.migration as mig
from pypowervm.wrappers import logical_partition as lpar_w


TIMEOUT_HELP = _("Maximum number of seconds for job to complete. If not "
                 "specified, will not timeout.")
PANEL_HELP = _("The IBMi panel function to execute.")
DUMPRESTART_HELP = _("Force a system dump before restarting partition.")
OP_MIGRATE = 'migrate'
OP_MIGRATE_STOP = 'migrate-stop'
OP_MIGRATE_RECOVER = 'migrate-recover'
OP_IBMI_PANEL = 'panel'


class LogicalPartition(bp_c.BasePartition):
    type_name = "LogicalPartition"
    title = "Logical Partitions"

    def __init__(self):
        super(LogicalPartition, self).__init__()
        self.pri_df_labels.insert(4, bp_c.PO_ENV)

        self._add_migrate_argument()
        self._add_recover_argument()
        self._add_stop_argument()
        self._add_power_on_operation(has_ipl_src=True)
        self._add_ibmi_panel_op()
        self._add_dumprestart_operation()

        self.setup_arguments()

    def _add_migrate_argument(self):
        mgd_sys_help = _("The name of the managed system to migrate to.")
        validate_help = _("Indication of whether to just validate the "
                          "migration or actually perform it. Default: False.")
        tgt_mgmt_srv_help = _("Required. The IP of the PowerVM management "
                              "platform managing the target host.")
        tgt_mgmt_usr_help = _("The user ID to use on the target "
                              "PowerVM management platform.")

        args = o_g.default_args(self, include_output=False,
                                object_id_required=True)
        args.extend([op.Argument('--sys', arg_type=str, required=False,
                                 arg_help=help.wrap_help(mgd_sys_help)),
                     op.Argument('--validate', action="store_true",
                                 default=False,
                                 arg_help=help.wrap_help(validate_help)),
                     op.Argument('--mgmt-svr', arg_type=str, required=True,
                                 arg_help=help.wrap_help(tgt_mgmt_srv_help)),
                     op.Argument('--mgmt-usr', arg_type=str, required=False,
                                 arg_help=help.wrap_help(tgt_mgmt_usr_help)),
                     op.Argument('--timeout', arg_type=int, required=False,
                                 arg_help=help.wrap_help(TIMEOUT_HELP)),
                     # --allow-ovs is for internal use only and is undocumented
                     op.Argument('--allow-ovs', action="store_true",
                                 default=False,
                                 arg_help=argparse.SUPPRESS)])
        self.add_operation(op.Operation(OP_MIGRATE, self.migrate,
                                        op_args=args,
                                        object_model=self.model))

    def _add_recover_argument(self):
        force_help = _("Force the migration to recover when errors are "
                       "encountered.")

        args = o_g.default_args(self, include_output=False,
                                object_id_required=True)
        args.extend([op.Argument('--force', action="store_true",
                                 default=False,
                                 arg_help=help.wrap_help(force_help)),
                     op.Argument('--timeout', arg_type=int, required=False,
                                 arg_help=help.wrap_help(TIMEOUT_HELP))])
        self.add_operation(op.Operation(OP_MIGRATE_RECOVER, self.recover,
                                        op_args=args,
                                        object_model=self.model))

    def _add_stop_argument(self):
        args = o_g.default_args(self, include_output=False,
                                object_id_required=True)
        args.extend([op.Argument('--timeout', arg_type=int, required=False,
                                 arg_help=help.wrap_help(TIMEOUT_HELP))])
        self.add_operation(op.Operation(OP_MIGRATE_STOP, self.stop,
                                        op_args=args,
                                        object_model=self.model))

    def _add_ibmi_panel_op(self):
        args = o_g.default_args(self, include_output=False,
                                object_id_required=True)
        args.extend([
            op.Argument('--timeout', arg_type=int, required=False,
                        arg_help=help.wrap_help(TIMEOUT_HELP)),
            op.Argument('--op', arg_type=str, required=True,
                        arg_help=help.wrap_help(PANEL_HELP),
                        choices=ibmi_task.IBMiPanelOperations.ALL_VALUES)])
        self.add_operation(op.Operation(
            OP_IBMI_PANEL, self.ibmi_panel, op_args=args,
            object_model=self.model))

    def _add_dumprestart_operation(self):
        args = o_g.default_args(self, include_output=False)
        args.extend([op.Argument('--no-prompt', action="store_true",
                                 arg_help=help.NOPROMPT_HELP)])
        self.add_operation(op.Operation(bp_c.DUMPRESTART_OP_NAME,
                                        self.dumprestart,
                                        op_args=args,
                                        object_model=self.model))

    def dumprestart(self):
        self.power_off_restart(bp_c.DUMPRESTART_OP_NAME)

    def ibmi_panel(self):
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       OP_IBMI_PANEL)
        timeout = cfg.CONF.pypowervm_job_request_timeout
        if self.ec.parsed_args.timeout:
            timeout = self.ec.parsed_args.timeout
        ibmi_task.start_panel_job(self.ec.wrappers[0], self.ec.parsed_args.op,
                                  timeout=timeout)

    def migrate(self):
        ''' migrate operation processing function '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       OP_MIGRATE)

        success_msg = (_("Migration of logical partition %(name)s succeeded.")
                       % {'name': self.ec.wrappers[0].name})
        err_msg = (_("Migration of logical partition %(name)s failed.")
                   % {'name': self.ec.wrappers[0].name})
        if self.ec.parsed_args.validate:
            success_msg = (_("Validation of logical partition %(name)s "
                             "succeeded.")
                           % {'name': self.ec.wrappers[0].name})
            err_msg = (_("Validation of logical partition %(name)s failed.")
                       % {'name': self.ec.wrappers[0].name})
            self.print_out(
                _("Validating logical partition %(name)s migration...")
                % {'name': self.ec.wrappers[0].name})
        else:
            self.print_out(_("Migration of logical partition %(name)s "
                             "starting. This may take a while.")
                           % {'name': self.ec.wrappers[0].name})

        try:
            arg_kwds = dict(tgt_mgd_sys=self.ec.parsed_args.sys,
                            validate_only=self.ec.parsed_args.validate,
                            tgt_mgmt_svr=self.ec.parsed_args.mgmt_svr,
                            tgt_mgmt_usr=self.ec.parsed_args.mgmt_usr,
                            sdn_override=self.ec.parsed_args.allow_ovs)
            if self.ec.parsed_args.timeout:
                arg_kwds['timeout'] = self.ec.parsed_args.timeout
            else:
                arg_kwds['timeout'] = 0

            mig.migrate_lpar(self.ec.wrappers[0], **arg_kwds)
        except Exception as mig_exc:
            self.print_err(err_msg)
            raise mig_exc
        self.print_out(success_msg)

    def recover(self):
        ''' migrate-recover operation processing function '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       OP_MIGRATE_RECOVER)

        self.print_out(_("Recovery of logical partition %(name)s starting. "
                         "This may take a while.")
                       % {'name': self.ec.wrappers[0].name})
        try:
            arg_kwds = dict()
            if self.ec.parsed_args.force:
                arg_kwds['force'] = self.ec.parsed_args.force
            if self.ec.parsed_args.timeout:
                arg_kwds['timeout'] = self.ec.parsed_args.timeout
            else:
                arg_kwds['timeout'] = 0
            mig.migrate_recover(self.ec.wrappers[0], **arg_kwds)
        except Exception as mig_exc:
            self.print_err(_("Recovery of logical partition %(name)s failed.")
                           % {'name': self.ec.wrappers[0].name})
            raise mig_exc
        self.print_out(_("Recovery of logical partition %(name)s succeeded.")
                       % {'name': self.ec.wrappers[0].name})

    def stop(self):
        ''' migrate-stop operation processing function '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       OP_MIGRATE_STOP)

        self.print_out(_("Stopping migration of logical partition %(name)s...")
                       % {'name': self.ec.wrappers[0].name})
        try:
            arg_kwds = dict()
            if self.ec.parsed_args.timeout:
                arg_kwds['timeout'] = self.ec.parsed_args.timeout
            else:
                arg_kwds['timeout'] = 0
            mig.migrate_abort(self.ec.wrappers[0], **arg_kwds)
        except Exception as mig_exc:
            self.print_err(_("Stopping the migration of logical partition "
                             "%(name)s failed.")
                           % {'name': self.ec.wrappers[0].name})
            raise mig_exc
        self.print_out(_("Stopping the migration of logical partition "
                         "%(name)s was successfully initiated. Check the "
                         "state of the partition to verify whether the stop "
                         "was successful.")
                       % {'name': self.ec.wrappers[0].name})

    @property
    def model(self):
        return lpar_w.LPAR

    @property
    def help(self):
        return help_lpar.help_text
