# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import os
import six
from six.moves import html_parser

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.command.sriov_adapter as sriov_c
import pvmcli.help.help_managed_system as help_msys
import pvmcli.util.exceptions as exc
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.output_processor as out_proc
import pvmcli.util.util as util

import pypowervm.tasks.management_console as m_con
import pypowervm.tasks.master_mode as m_mode
import pypowervm.wrappers.iocard as iocard
import pypowervm.wrappers.managed_system as ms_w
import pypowervm.wrappers.mtms as mtms


REQUEST_NAME = "request-master"
LIST_KEY_NAME = "list-public-key"
ADD_KEY_NAME = "add-auth-key"
LIST_KEYS_NAME = "list-auth-keys"
REST_LOG_LEVELS_NAME = "rest-log-levels"

REST_LOG_TYPES = ('audit', 'developer', 'ffdc', 'job', 'security')
REST_LOG_LEVELS = ('TRACE', 'ALL', 'ERROR', 'INFO', 'FATAL', 'DEBUG',
                   'OFF', 'WARN')

PO_NAME = _("Name")
PO_MTMS = _("MTMS")
PO_MEM = _("Mem")
PO_FREEMEM = _("Free Mem")
PO_PROC = _("Proc")
PO_FREEPROC = _("Free Proc")

PRI_DISP_FIELD_LABELS = [PO_NAME, PO_MTMS, PO_MEM, PO_FREEMEM, PO_PROC,
                         PO_FREEPROC]


class ManagedSystem(ObjectType):
    def __init__(self):
        super(ManagedSystem, self).__init__()

        # DISPLAY
        self.add_field_models([mtms.MTMS, ms_w.ASIOConfig, ms_w.IOSlot,
                               iocard.SRIOVAdapter, iocard.SRIOVConvPPort,
                               iocard.SRIOVEthPPort])

        self.remap_schema(iocard.SRIOVConvPPort.schema_type,
                          sriov_c.PPORT_SCHEMA)
        self.remap_schema(iocard.SRIOVEthPPort.schema_type,
                          sriov_c.PPORT_SCHEMA)

        # Exclude this field as it is a wrapper reference from PhysicalPort
        # back to its parent SRIOVAdapter wrapper. Including this field
        # causes an infinite recursion in OutputGraph.init(), and
        # is not a useful display field as all the display fields for
        # SRIOVAdapter are already covered by said parent wrapper.
        self.exc_disp_fields.extend(
            ['PhysicalPort.sriov_adap', 'IOSlot.dyn_reconfig_conn_index',
             'IOSlot.dyn_reconfig_conn_name', 'IOSlot.pci_revision_id',
             'IOSlot.pci_sub_dev_id', 'IOSlot.pci_sub_vendor_id'])

        d_op = self.add_display_operation()
        d_op.set_d_field_label(ms_w.System, 'asio_config',
                               ms_w.ASIOConfig.schema_type)
        d_op.set_d_field_label(ms_w.System, 'mtms', mtms.MTMS.schema_type)
        d_op.set_d_field_label(ms_w.ASIOConfig, 'io_slots',
                               ms_w.IOSlot.schema_type)
        d_op.set_d_field_label(ms_w.System, 'memory_configurable',
                               'memory_configurable (MB)')
        d_op.set_d_field_label(ms_w.System, 'memory_free',
                               'memory_free (MB)')
        d_op.set_d_field_label(ms_w.System, 'memory_region_size',
                               'memory_region_size (MB)')
        d_op.set_d_field_label(ms_w.System, 'memory_total',
                               'memory_total (MB)')
        d_op.set_d_field_out_func(ms_w.IOSlot, 'drc_index',
                                  out_func=out_proc.hex_no_prefix)
        d_op.set_d_field_out_func(ms_w.IOSlot, 'pci_class',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(ms_w.IOSlot, 'pci_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(ms_w.IOSlot, 'pci_subsys_dev_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(ms_w.IOSlot, 'pci_subsys_vendor_id',
                                  out_func=out_proc.four_char_hex)
        d_op.set_d_field_out_func(ms_w.IOSlot, 'pci_vendor_id',
                                  out_func=out_proc.four_char_hex)

        self.add_update_operation()

        self._add_request_master_operation()
        self._add_list_public_key_operation()
        self._add_auth_key_operation()
        self._add_list_keys_operation()
        self._add_rest_log_levels_operation()

        # TODO clbush (IBM): Jobs
        # ClearSRIOVPhysicalPortStatistics (Maybe)
        # ModifySrIOVAdapterMode (Definite)
        # QueryReservedMemoryRequiredForPartition (Definite)

        self.setup_arguments()

    def _add_request_master_operation(self):
        args = o_g.default_args(self)
        self.add_operation(op.Operation(REQUEST_NAME, self.request_master,
                                        op_args=args,
                                        object_model=self.model))

    def request_master(self):
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       REQUEST_NAME)

        m_mode.request_master(self.ec.wrappers[0])

    def _add_list_public_key_operation(self):
        args = o_g.default_args(self, include_output=False)
        self.add_operation(op.Operation(LIST_KEY_NAME, self.list_pub_key,
                                        op_args=args,
                                        object_model=self.model))

    def list_pub_key(self):
        ''' List the public key on the management console.

        There is only one managed system for any Novalink instance, so
        we facilitate the listing and management of keys on the management
        console through the ManagedSystem object type.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       LIST_KEY_NAME)

        key = m_con.get_public_key(self.ec.sm.adapter)
        self.print_out(key)

    def _add_list_keys_operation(self):
        args = o_g.default_args(self, include_output=False)
        self.add_operation(op.Operation(LIST_KEYS_NAME, self.list_keys,
                                        op_args=args,
                                        object_model=self.model))

    def list_keys(self):
        ''' List all authorized keys on the management console.

        There is only one managed system for any Novalink instance, so
        we facilitate the listing and management of keys on the management
        console through the ManagedSystem object type.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       LIST_KEY_NAME)

        keys = m_con.get_authorized_keys(self.ec.sm.adapter)

        if not keys:
            raise exc.NoKeysFoundError()
        self.print_out(os.linesep.join(keys))

    def _add_auth_key_operation(self):
        pub_key_help = _("The new authorized public key to be added, if it "
                         "does not already exist.")

        args = o_g.default_args(self, include_output=False)
        args.append(op.Argument('--key', arg_type=str, required=True,
                                arg_help=help.wrap_help(pub_key_help)))
        self.add_operation(op.Operation(ADD_KEY_NAME, self.add_auth_key,
                                        op_args=args,
                                        object_model=self.model))

    def add_auth_key(self):
        ''' Add an authorized key to the management console.

        There is only one managed system for any Novalink instance, so
        we facilitate the listing and management of keys on the management
        console through the ManagedSystem object type.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers, self.name,
                                       ADD_KEY_NAME)

        # Don't allow empty keys
        if not self.ec.parsed_args.key:
            raise exc.EmptyKeyError()

        m_con.add_authorized_key(self.ec.sm.adapter, self.ec.parsed_args.key)

    def _add_rest_log_levels_operation(self):
        set_log_levels_help = _(
            'Set one or more logging levels on the REST server. '
            'Example: --set audit=OFF developer=TRACE')
        args = o_g.default_args(self, include_output=False)
        args.append(op.Argument(
            '--set', arg_type=str, required=False,
            arg_help=help.wrap_help(set_log_levels_help), nargs='+',
            metavar="type=level"))
        self.add_operation(op.Operation(
            REST_LOG_LEVELS_NAME, self.rest_log_levels, args, self.model,
            suppress_help=True))

    def rest_log_levels(self):
        querystring = ''
        sets = self.ec.parsed_args.set
        if sets:
            # Validate
            for split in (arg.split('=', 1) for arg in sets):
                if len(split) != 2:
                    raise exc.InvalidKeyValue(arg='--set')
                logtype, level = split
                if logtype not in REST_LOG_TYPES:
                    raise exc.InvalidRESTLogType(
                        logtype=logtype, valid=', '.join(REST_LOG_TYPES))
                if level not in REST_LOG_LEVELS:
                    raise exc.InvalidRESTLogLevel(
                        level=level, valid=', '.join(REST_LOG_LEVELS))
            querystring = '?' + '&'.join(sets)
        resp = self.ec.sm.adapter.read_by_path(
            '/rest/api/uom/Debug/SetLoggingLevel' + querystring)
        out = html_parser.HTMLParser().unescape(resp.body[5:][:-6])
        if not querystring:
            # Strip off the USAGE statement
            out = out[:out.find('=' * 57)]
        util.print_(out)

    @property
    def model(self):
        return ms_w.System

    def get_out_rows(self):
        wrp = self.ec.wrappers[0]
        return [[wrp.system_name, wrp.mtms.mtms_str,
                 six.text_type(wrp.memory_configurable),
                 six.text_type(wrp.memory_free),
                 six.text_type(wrp.proc_units_configurable),
                 six.text_type(wrp.proc_units_avail)]]

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.get_out_rows())
        return "%(title)s\n%(tbl)s" % {'title': _('Managed System'),
                                       'tbl': po_table.get_string()}

    @property
    def help(self):
        return help_msys.help_text
