# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import logging
import six

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_network_bridge as help_nbr
from pvmcli.util.i18n import _
from pvmcli.util import exceptions as e
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as u

import pypowervm.tasks.network_bridger as n_bridger
from pypowervm.wrappers import network as n
from pypowervm.wrappers import managed_system as ms_w
from pypowervm.wrappers import virtual_io_server as v


PRI_DISP_FIELD_LABELS = [_("PVID"), _("VSwitch"),
                         _("VIOS"), _("Back Dev"), _("SEA"), _("Ld Grp"),
                         _("Addl VLANs")]


# Maximum number of VIOS per NetworkBridge
MAX_VIOS = 2

ID_FIELD_TYPES = {'pvid': int}

OP_ADD_VLAN_NAME = 'add-vlans'
OP_REMOVE_VLAN_NAME = 'remove-vlans'
ADD_VLAN_IDS_HELP = _("Space separated list of VLAN IDs to add to this "
                      "network bridge.")
REMOVE_VLAN_IDS_HELP = _("Space separated list of VLAN IDs to remove from "
                         "this network bridge.")


class NetworkBridge(ObjectType):
    def __init__(self):
        super(NetworkBridge, self).__init__()

        self.add_field_models([n.LoadGroup,
                               n.TrunkAdapter,
                               n.SEA,
                               n.EthernetBackingDevice])
        self.add_display_operation(parent_models=[ms_w.System],
                                   obj_id_field_types=ID_FIELD_TYPES)
        self.add_update_operation(parent_models=[ms_w.System],
                                  obj_id_field_types=ID_FIELD_TYPES)
        self.add_delete_operation(parent_models=[ms_w.System],
                                  obj_id_field_types=ID_FIELD_TYPES)
        self._add_create_operation()
        self._add_add_vlans_operation()
        # TODO (clbush): Ability here and pypowervm to remove VLANS off
        #                primary adapter
        self._add_remove_vlans_operation()

        self.setup_arguments()

        self.vlan_cache = {}

    def _bld_ta_lg_maps(self, wrp):
        ''' Build two dictionaries, one from VIOS to trunk adapters, and
        another from trunk adapter to load group.

        :param wrp: NetBridge wrapper object
        :return: tuple of dicts - VIOS to Trunk adapter and Trunk to load group
        '''
        vios_id_to_lg_tas = {}
        ta_to_lg = {}

        for ld_grp in wrp.load_grps:
            for ta in ld_grp.trunk_adapters:
                if ta.vios_id in vios_id_to_lg_tas:
                    vios_id_to_lg_tas[ta.vios_id].append(ta)
                else:
                    vios_id_to_lg_tas[ta.vios_id] = [ta]
                ta_to_lg[ta.loc_code] = ld_grp
        return vios_id_to_lg_tas, ta_to_lg

    def _nbr_out_rows_wrp(self, wrp):
        ''' Return the priority output rows for a provided NetBridge wrapper.

        :param wrp: NetBridge wrapper object
        :return: A list of list of strings to be provided for priority output.
        '''
        out_rows = []
        # Get a map from VIOS to SEA
        vios_id_to_seas = {}
        vios_id_to_name = {}
        for sea in wrp.seas:
            found_vios = u.vios_by_uri(self.ec.sm.adapter, sea.vio_uri)
            vios_id_to_seas[found_vios.id] = sea
            vios_id_to_name[found_vios.id] = found_vios.name

        # Get a map from VIOS to trunk adapters and ta to load group
        vios_id_to_lg_tas, ta_to_lg = self._bld_ta_lg_maps(wrp)

        out_row = [six.text_type(wrp.pvid), u.vswitch_search(
            self.ec.sm.adapter, wrp.vswitch_id, self.ec.msys.uuid).name]
        for vios_id in sorted(vios_id_to_name):
            sea = vios_id_to_seas[vios_id]

            out_row.extend([vios_id_to_name[vios_id],
                            sea.backing_device.dev_name, sea.dev_name,
                            vios_id_to_lg_tas[vios_id][0].dev_name])

            lg_vlans = u.group_join_list(
                ta_to_lg[vios_id_to_lg_tas[vios_id][0].loc_code].tagged_vlans)
            out_row.append(lg_vlans[0] if lg_vlans else "")
            out_rows.append(out_row)

            for lg_vlan in lg_vlans[1:]:
                out_rows.append(["", "", "", "", "", "", lg_vlan])
            for ta in vios_id_to_lg_tas[vios_id][1:]:
                out_row = ["", "", "", "", "", ta.dev_name]
                lg_vlans = u.group_join_list(
                    ta_to_lg[ta.loc_code].tagged_vlans)
                out_row.append(lg_vlans[0] if lg_vlans else "")
                out_rows.append(out_row)
                for lg_vlan in lg_vlans[1:]:
                    out_rows.append(["", "", "", "", "", "", lg_vlan])
            out_row = ["", ""]
        return out_rows

    def nbr_out_rows(self):
        out_rows = []

        # Wrappers based on PVID, for sorting
        wrp_pvid_map = {}
        for wrp in self.ec.wrappers:
            if wrp.pvid in wrp_pvid_map:
                wrp_pvid_map[wrp.pvid].append(wrp)
            else:
                wrp_pvid_map[wrp.pvid] = [wrp]

        for wrp_pvid in sorted(wrp_pvid_map):
            for nbr_wrp in wrp_pvid_map[wrp_pvid]:
                out_rows.extend(self._nbr_out_rows_wrp(nbr_wrp))

        return out_rows

    @property
    def model(self):
        return n.NetBridge

    def create(self):
        ''' Create operation processing function. '''
        ms_uuid = self.ec.msys.uuid

        nb_pvid = self.ec.parsed_args.pvid

        nb_vswitch_name = self.ec.parsed_args.vswitch_name
        vswitch_list = n.VSwitch.search(self.ec.sm.adapter,
                                        parent_type=ms_w.System.schema_type,
                                        parent_uuid=ms_uuid,
                                        name=nb_vswitch_name)

        vswitch = vswitch_list[0] if vswitch_list else None
        if not vswitch:
            raise e.NoEntriesFoundError(object_type=n.VSwitch.schema_type)

        nb_vios_adpts = self.ec.parsed_args.vios_adapters
        if len(nb_vios_adpts) % 2 != 0:
            # Should be an even number of arguments since we're
            # dealing with pairs.
            raise e.VIOSAdaptersPairError()

        if len(nb_vios_adpts) > MAX_VIOS * 2:
            # Can only have 2 VIOS (so 4 elements in this paired list)
            raise e.VIOSAdaptersQuantityError()

        nb_vios_adpt_map = []
        for i in range(0, len(nb_vios_adpts), 2):
            viosname, adptname = nb_vios_adpts[i:i+2]
            vios_list = v.VIOS.search(self.ec.sm.adapter,
                                      name=viosname)
            vios = vios_list[0] if vios_list else None
            if not vios:
                raise e.InvalidVIOSNameError(name=viosname)

            nb_vios_adpt_map.append((vios.related_href, adptname))

        new_nb = n.NetBridge.bld(self.ec.sm.adapter,
                                 nb_pvid, nb_vios_adpt_map, vswitch)

        logging.getLogger('pvmctl').info(u.utf8_str(
            "Sending to REST with XML: %s" % new_nb.toxmlstring(pretty=True)))
        created_nb = new_nb.create(parent_type=ms_w.System.schema_type,
                                   parent_uuid=ms_uuid)

        nb_vlans = self.ec.parsed_args.vlans
        if nb_vlans:
            n_bridger.ensure_vlans_on_nb(self.ec.sm.adapter, ms_uuid,
                                         created_nb.uuid, nb_vlans)

        self.ec.wrappers = [created_nb]
        # Only display if a user specifies specific display fields
        if (self.ec.parsed_args.display_fields
                or self.ec.parsed_args.all_out):
            self.display()

    def _add_create_operation(self):
        pvid_help = _("Primary VLAN ID for this Network Bridge.")
        vswitch_name_help = _("Name of the backing PowerVM Virtual Switch "
                              "on this Network Bridge's Trunk Adapter.")
        vios_name_adpt_help = _("A list of Virtual I/O Server name to backing "
                                "ethernet device name pairs. Maximum of 2 "
                                "pairs.\nExample: --vios-adapters vios1 "
                                "ent1 vios2 ent2")
        vios_name_adpt_metavar = _("<VIOS 1 Name> <Backing Adapter 1 Name> "
                                   "[<VIOS 2 Name> "
                                   "<Backing Adapter 2 Name>]")
        crt_args = o_g.default_args(self, parent_models=[ms_w.System])
        crt_args.extend([op.Argument('--pvid', arg_help=pvid_help,
                                     required=True, arg_type=str),
                         op.Argument('--vswitch-name',
                                     arg_help=vswitch_name_help,
                                     required=True, arg_type=str),
                         op.Argument('--vios-adapters',
                                     arg_help=vios_name_adpt_help,
                                     required=True, nargs='+',
                                     metavar=vios_name_adpt_metavar),
                         op.Argument('--vlans',
                                     arg_help=ADD_VLAN_IDS_HELP, arg_type=int,
                                     required=False, nargs='+')])
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def _add_remove_vlans_operation(self):
        args = o_g.default_args(self, parent_models=[ms_w.System],
                                include_output=False)
        args.append(op.Argument('--vlans', arg_help=REMOVE_VLAN_IDS_HELP,
                                arg_type=int, required=True, nargs='+'))
        self.add_operation(op.Operation(OP_REMOVE_VLAN_NAME, self.remove_vlans,
                                        op_args=args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def remove_vlans(self):
        ms_uuid = self.ec.msys.uuid
        nb_vlans = self.ec.parsed_args.vlans
        # TODO (IBM): Need to have the ability to batch the removals so
        #             they're all done under a single update
        for nb_vlan in nb_vlans:
            n_bridger.remove_vlan_from_nb(self.ec.sm.adapter, ms_uuid,
                                          self.ec.wrappers[0].uuid, nb_vlan)

    def _add_add_vlans_operation(self):
        args = o_g.default_args(self, parent_models=[ms_w.System],
                                include_output=False)
        args.append(op.Argument('--vlans', arg_help=ADD_VLAN_IDS_HELP,
                                arg_type=int, required=True, nargs='+'))
        self.add_operation(op.Operation(OP_ADD_VLAN_NAME, self.add_vlans,
                                        op_args=args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def add_vlans(self):
        ms_uuid = self.ec.msys.uuid
        nb_vlans = self.ec.parsed_args.vlans
        n_bridger.ensure_vlans_on_nb(self.ec.sm.adapter, ms_uuid,
                                     self.ec.wrappers[0].uuid, nb_vlans)

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.nbr_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Network Bridges"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_nbr.help_text
