# PYTHON_ARGCOMPLETE_OK

# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.

import argparse
from argparse import RawTextHelpFormatter as rthf
from abc import ABCMeta, abstractproperty
import prettytable as pt
import re
import six
import sys

import pvmcli.command.operation as op
import pvmcli.util.exceptions as e
from pvmcli.util import help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as op_gen
import pvmcli.util.output_processor as out_proc
import pvmcli.util.session_manager as sm
import pvmcli.util.util as util

from pypowervm.utils import retry
import pypowervm.wrappers.entry_wrapper as e_w
import pypowervm.wrappers.managed_system as m_sys
import pypowervm.wrappers.storage as stor


@retry.retry(argmod_func=retry.refresh_wrapper)
def update_retry(wrp, set_fields, obj_type):
    # It's possible for set_fields to be empty if update() is overridden.
    if set_fields:
        for s_field in set_fields:
            obj_type.set_field_val(wrp, s_field, set_fields[s_field])

    wrap = obj_type.ec.wrappers[0]
    if isinstance(wrp, stor.VDisk):
        wrap = obj_type.ec.elem_c_wrappers[0]
    elif isinstance(wrp, stor.PV):
        wrap = obj_type.ec.elem_p_wrappers[0]

    wrap = wrap.update()


class ExecutionContext(object):
    def __init__(self, session_manager, parsed_args, exec_op):
        ''' Executing context for this query. '''
        self.sm = session_manager
        self.parsed_args = parsed_args
        self.exec_op = exec_op

        self._msys = None
        self.wrappers = None
        self.elem_p_wrappers = []
        self.elem_c_wrappers = []

        # Map from parent wrappers to all child wrappers
        self.parent_wrp_to_children = {}

    def _set_wrappers(self, wrps):
        self.wrappers = wrps

    def _add_parent_wrappers(self, wrps):
        self.elem_p_wrappers.extend(wrps)

    def _add_child_wrappers(self, wrps):
        self.elem_c_wrappers.extend(wrps)

    def query_wrappers(self):
        ''' Query for wrappers of this object type based on parsed ids. '''
        # We are a root type
        if not self.exec_op.parent_models:
            self._set_wrappers(self._query_wrps(
                self.exec_op.object_model, obj_id=self.parsed_args.object_id,
                id_fields=self.exec_op.id_fields))
        elif not self.exec_op.elem_wrp_funcs:
            # We are an entry, not an element with a parent model
            obj_m = self.exec_op.object_model
            obj_id = self.parsed_args.object_id
            self._set_wrappers(self._query_child_wrps(
                obj_m, obj_id, self.exec_op.id_fields))
        else:
            self._set_wrappers(self._query_element_wrps())

    def _query_wrps(self, model, obj_id=None, id_fields=None,
                    parent_type=None, parent_uuid=None):
        if obj_id and id_fields:
            id_kwargs = util.split_eq_arg(obj_id, id_fields)
            return model.search(self.sm.adapter, xag=[],
                                parent_type=parent_type,
                                parent_uuid=parent_uuid, **id_kwargs)
        if parent_type:
            # TODO (clbush): xag based on pypowervm decorators
            resp = self.sm.adapter.read(root_type=parent_type,
                                        root_id=parent_uuid,
                                        child_type=model.schema_type,
                                        xag=[])
        else:
            # TODO (clbush): xag based on pypowervm decorators
            resp = self.sm.adapter.read(model.schema_type, xag=[])

        wrapped_resp = model.wrap(resp)
        return (wrapped_resp if isinstance(wrapped_resp, list)
                else [wrapped_resp])

    def _query_child_wrps(self, object_model, object_id, id_fields):
        ''' Retrieve all child wrappers of an Object Type.
            Only called if this object type is a child type.
        '''
        # Retrieve relevant parent wrappers
        for parent_model in self.exec_op.parent_models:
            self._add_parent_wrappers(self._query_wrps(
                parent_model, obj_id=self.parsed_args.parent_id,
                id_fields=self.exec_op.parent_id_fields))

        # Based on parent wrappers, retrieve object wrappers
        child_wrps = []
        for p_wrp in self.elem_p_wrappers:
            c_wrps = self._query_wrps(
                object_model, obj_id=object_id, id_fields=id_fields,
                parent_type=p_wrp.schema_type, parent_uuid=p_wrp.uuid)
            self.parent_wrp_to_children[p_wrp] = c_wrps
            child_wrps.extend(c_wrps)
        return child_wrps

    def _query_element_wrps(self):
        # Load wrapper cache with all entries relevant to this query
        wrps = []
        if self.exec_op.child_models:
            for child_model in self.exec_op.child_models:
                wrps.extend(self._query_child_wrps(
                    child_model,
                    self.parsed_args.child_id,
                    self.exec_op.child_id_fields))
                self._add_child_wrappers(wrps)
        elif self.exec_op.parent_models:
            for parent_model in self.exec_op.parent_models:
                p_wrps = self._query_wrps(
                    parent_model, obj_id=self.parsed_args.parent_id,
                    id_fields=self.exec_op.parent_id_fields)
                wrps.extend(p_wrps)
            self._add_parent_wrappers(wrps)

        # Walk the child wrappers, building a new list that will become
        # the child wrappers when complete.
        elem_wrp_list = []
        ew_funcs = self.exec_op.elem_wrp_funcs
        for wrp in wrps:
            new_elem_vals = self._walk_elem_wrp_funcs(wrp, ew_funcs, [])
            if new_elem_vals:
                if self.parsed_args.object_id:
                    id_map = util.split_eq_arg(self.parsed_args.object_id,
                                               self.exec_op.id_fields)
                    for new_elem_val in new_elem_vals:
                        new_attr_val = getattr(new_elem_val,
                                               list(id_map.keys())[0])
                        cast_nav = six.text_type(new_attr_val)
                        if cast_nav == id_map[list(id_map.keys())[0]]:
                            elem_wrp_list.append(new_elem_val)
                else:
                    elem_wrp_list.extend(new_elem_vals)
        return elem_wrp_list

    def _walk_elem_wrp_funcs(self, wrp, elem_wrp_funcs, curr_list):
        attr_val = getattr(wrp, elem_wrp_funcs[0].fget.__name__)
        if len(elem_wrp_funcs) == 1:
            curr_list.extend(attr_val)
            return curr_list
        elif isinstance(attr_val, e_w.Wrapper):
            self._walk_elem_wrp_funcs(attr_val, elem_wrp_funcs[1:],
                                      curr_list)
        elif isinstance(attr_val, e_w.WrapperElemList):
            for elem in attr_val:
                self._walk_elem_wrp_funcs(elem, elem_wrp_funcs[1:], curr_list)

        return curr_list

    @property
    def msys(self):
        if not self._msys:
            self._msys = self.find_managed_system()
        return self._msys

    def get_msys_io_slot(self, drc_name):
        for io_slot in self.msys.asio_config.io_slots:
            if io_slot.dyn_reconfig_conn_name == drc_name:
                return io_slot

        return None

    def find_managed_system(self):
        ''' Query for the only managed system of this instance. '''
        # There is only 1 managed system per instance, return its wrapper.
        # TODO (clbush): xag based on pypowervm decorators
        resp = self.sm.adapter.read(m_sys.System.schema_type, xag=[])
        if resp is None:
            # Something went terribly wrong if we see this.
            raise e.NoEntriesFoundError(object_type=m_sys.System.schema_type)

        return m_sys.System.wrap(resp)[0]

    @property
    def id_arg(self):
        ''' The --object-id Argument object for this operation. '''
        return self.exec_op.op_args[op.ID_NAME]

    @property
    def parent_id_arg(self):
        ''' The --parent-id Argument object for this operation. '''
        return self.exec_op.op_args[op.PARENT_ID_NAME]

    @property
    def child_id_arg(self):
        ''' The --child-id Argument object for this operation. '''
        return self.exec_op.op_args[op.CHILD_ID_NAME]


@six.add_metaclass(ABCMeta)
class ObjectType(object):
    ''' Base Object Type that all object types must extend.

        Lazily builds up all arguments for all object types based on the
        specified object type and operation. Also provides common operation
        processing that can be used by concrete object type implementations.
    '''

    def __init__(self):
        # Parser used to parse and process all provided arguments
        self.parser = argparse.ArgumentParser(add_help=False)

        self.sess_man = None
        self.parsed_args = None

        # Execution context, constructed on call to execute method
        self.ec = None

        # ops is a map from an operation function to its Operation
        # object. This allows for Operation fields to be accessed during
        # execution.
        self.ops = {}
        self.name_to_ops = {}
        self._validators = []
        self._field_models = [self.model]
        self._schema_remap = {}

        self.cli_out = sys.stdout
        self.cli_err = sys.stderr
        # Map from wrapper object to output graph
        self.out_graphs = {}

        self.exc_disp_fields = []
        self.exc_set_fields = []

    @abstractproperty
    def model(self):
        ''' pypowervm entry wrapper model for this ObjectType since we don't
            know what operation will be executed upon ObjectType construction.

            Used for argparse and help text construction. The processing
            itself is handled separately through Operation objects' root_model,
            child_model and elem_wrp_func methods.
        '''
        # TODO (clbush): This has become increasingly redundant with operation
        #                driven processing. Remove the need for this property.
        raise NotImplementedError()

    @property
    def name(self):
        ''' The name of this object type.
            Defaults to the model's schema type.
        '''
        return self.model.schema_type

    @property
    def field_models(self):
        ''' The wrapper models to parse when generating display/set fields. '''
        return self._field_models

    @field_models.setter
    def field_models(self, new_models):
        self._field_models = new_models

    def print_out(self, str_out):
        ''' Prints str_out to cli_out. '''
        util.print_(str_out, outfile=self.cli_out)

    def print_err(self, str_out):
        ''' Prints str_out to cli_err. '''
        util.print_(str_out, outfile=self.cli_err)

    def add_field_models(self, models):
        ''' Add models to the list to be parsed by field_models. '''
        self._field_models.extend(models)

    def remap_schema(self, old_schema, new_schema):
        ''' Remap a schema type to support schemas on fields that are
            different than what is specified in pypowervm.

            Example: ClientNetworkAdapter in pypowervm is
                     VirtualEthernetAdapter here.

        :param old_schema: A String representing the traditional schema type
        :param new_schema: The new String representing that schema type
        '''
        self._schema_remap[old_schema] = new_schema

    def get_schema(self, traditional_schema):
        ''' Return the CLI representation of a field's schema.

        :param traditional_schema: The adapter provided schema type
        :returns: The remapped schema type if it exists, adapter schema if not
        '''
        return (self._schema_remap[traditional_schema]
                if traditional_schema in self._schema_remap
                else traditional_schema)

    def get_disp_fields(self):
        ''' Retrieve all display fields.

        Only populated and available after construction of class with list
        argument.

        :returns: The list of DisplayField objects for this ObjectType.
        '''
        return self.name_to_ops[op.LIST_NAME].op_args[op.DF_NAME].fields

    def add_operation(self, oper, parent_id_field_types={},
                      child_id_field_types={}, obj_id_field_types={}):
        ''' Add a supported operation for this object type.

        :param oper: Operation object to be added
        :param parent_id_field_types: Overridden parent id field types
                                      Example: {'pvid': int, 'id': int}
        :param child_id_field_types: Overridden child id field types
                                     Example: {'pvid': int, 'id': int}
        :param object_id_field_types: Overridden object id field types
                                      Example: {'pvid': int, 'id': int}
        '''
        if type(oper) != op.Operation:
            raise TypeError()
        self.ops[oper.op_func] = oper
        self.name_to_ops[oper.op_name] = oper
        self._set_op_field_types(
            oper, parent_id_field_types=parent_id_field_types,
            child_id_field_types=child_id_field_types,
            obj_id_field_types=obj_id_field_types)

    def _set_op_field_types(self, oper, parent_id_field_types={},
                            child_id_field_types={},
                            obj_id_field_types={}):
        ''' Set the field types of the parent, child and object models
            on a specified operation.

        :param oper: Operation of which to set the field types.
        :param parent_id_field_types: A map from field name to field types for
                                      the parent model. If a field isn't
                                      specified, defaults to str.
                                      Ex: {'field_foo': int,
                                           'field_bar': WrapperElemList}
        :param child_id_field_types: A map from field name to field types for
                                     the child model. If a field isn't
                                     specified, defaults to str.
                                     Ex: {'field_foo': int,
                                          'field_bar': WrapperElemList}
        :param obj_id_field_types: A map from field name to field types for
                                   object model. If a field isn't
                                   specified, defaults to str.
                                   Ex: {'field_foo': int,
                                        'field_bar': WrapperElemList}
        '''
        for f in parent_id_field_types:
            oper.parent_id_fields[f].field_type = parent_id_field_types[f]
        for f in child_id_field_types:
            oper.child_id_fields[f].field_type = child_id_field_types[f]
        for f in obj_id_field_types:
            oper.id_fields[f].field_type = obj_id_field_types[f]

    def add_display_operation(self, model=None, parent_models=None,
                              child_models=None, elem_wrp_funcs=None,
                              parent_id_field_types={},
                              child_id_field_types={},
                              obj_id_field_types={}, op_name=op.LIST_NAME):
        ''' Add the display operation on this object type.

        :param model: The entry wrapper model for this operation. This may
                      be different than what is provided by the model property
                      as update/deletes on element types can be against a
                      different parent wrapper than what is used for list.
                      Defaults to self.model if not specified.
        :param parent_models: The top level models of this object type,
                              if it isn't the model of this object type.
        :param child_models: The 2nd level models of this object type,
                             if it isn't the model of this object type.
        :param elem_wrp_funcs: List of element wrapper functions if
                               this is an element type.
        :param parent_id_field_types: A map from field name to field types for
                                      the parent model. If a field isn't
                                      specified, defaults to str.
                                      Ex: {'field_foo': int,
                                           'field_bar': WrapperElemList}
        :param child_id_field_types: A map from field name to field types for
                                     the child model. If a field isn't
                                     specified, defaults to str.
                                     Ex: {'field_foo': int,
                                          'field_bar': WrapperElemList}
        :param obj_id_field_types: A map from field name to field types for
                                   object model. If a field isn't
                                   specified, defaults to str.
                                   Ex: {'field_foo': int,
                                        'field_bar': WrapperElemList}
        :param op_name: The user specifiable name of the operation.
        :return: The new display operation.
        '''
        if not model:
            model = self.model
        args = op_gen.default_args(self, parent_models=parent_models,
                                   child_models=child_models)

        # Build operation
        d_op = op.Operation(op_name, self.display,
                            args, object_model=model,
                            parent_models=parent_models,
                            child_models=child_models,
                            elem_wrp_funcs=elem_wrp_funcs)

        self.add_operation(d_op, parent_id_field_types=parent_id_field_types,
                           child_id_field_types=child_id_field_types,
                           obj_id_field_types=obj_id_field_types)
        return d_op

    def add_update_operation(self, model=None, parent_models=None,
                             child_models=None, elem_wrp_funcs=None,
                             parent_id_field_types={},
                             child_id_field_types={},
                             obj_id_field_types={}, op_name=op.UPDATE_NAME):
        ''' Add the update operation on this object type.

        :param model: The entry wrapper model for this operation. This may
                      be different than what is provided by the model property
                      as update/deletes on element types can be against a
                      different parent wrapper than what is used for list.
                      Defaults to self.model if not specified.
        :param parent_models: The top level models of this object type,
                              if it isn't the model of this object type.
        :param child_models: The 2nd level models of this object type,
                             if it isn't the model of this object type.
        :param elem_wrp_funcs: List of element wrapper functions if
                               this is an element type.
        :param parent_id_field_types: A map from field name to field types for
                                      the parent model. If a field isn't
                                      specified, defaults to str.
                                      Ex: {'field_foo': int,
                                           'field_bar': WrapperElemList}
        :param child_id_field_types: A map from field name to field types for
                                     the child model. If a field isn't
                                     specified, defaults to str.
                                     Ex: {'field_foo': int,
                                          'field_bar': WrapperElemList}
        :param obj_id_field_types: A map from field name to field types for
                                   object model. If a field isn't
                                   specified, defaults to str.
                                   Ex: {'field_foo': int,
                                        'field_bar': WrapperElemList}
        :param op_name: The user specifiable name of the operation.
        :return: The new update operation.
        '''
        if not model:
            model = self.model
        args = op_gen.default_args(self, parent_models=parent_models,
                                   child_models=child_models,
                                   include_output=False)

        # Build arguments
        args.append(op_gen.set_argument(self))

        # Build operation
        u_op = op.Operation(op_name, self.update,
                            args, object_model=model,
                            parent_models=parent_models,
                            child_models=child_models,
                            elem_wrp_funcs=elem_wrp_funcs)

        self.add_operation(u_op, parent_id_field_types=parent_id_field_types,
                           child_id_field_types=child_id_field_types,
                           obj_id_field_types=obj_id_field_types)

        return u_op

    def add_delete_operation(self, model=None, parent_models=None,
                             child_models=None, elem_wrp_funcs=None,
                             parent_id_field_types={},
                             child_id_field_types={},
                             obj_id_field_types={}, op_name=op.DELETE_NAME,
                             child_id_req=False, parent_id_req=False,
                             obj_id_req=True):
        ''' Add the delete operation on this object type.

        :param model: The entry wrapper model for this operation. This may
                      be different than what is provided by the model property
                      as update/deletes on element types can be against a
                      different parent wrapper than what is used for list.
                      Defaults to self.model if not specified.
        :param parent_models: The top level models of this object type,
                              if it isn't the model of this object type.
        :param child_models: The 2nd level models of this object type,
                             if it isn't the model of this object type.
        :param elem_wrp_funcs: List of element wrapper functions if
                               this is an element type.
        :param parent_id_field_types: A map from field name to field types for
                                      the parent model. If a field isn't
                                      specified, defaults to str.
                                      Ex: {'field_foo': int,
                                           'field_bar': WrapperElemList}
        :param child_id_field_types: A map from field name to field types for
                                     the child model. If a field isn't
                                     specified, defaults to str.
                                     Ex: {'field_foo': int,
                                          'field_bar': WrapperElemList}
        :param obj_id_field_types: A map from field name to field types for
                                   object model. If a field isn't
                                   specified, defaults to str.
                                   Ex: {'field_foo': int,
                                        'field_bar': WrapperElemList}
        :param op_name: The user specifiable name of the operation.
        :return: The new delete operation.
        '''
        if not model:
            model = self.model
        args = op_gen.default_args(
            self, parent_models=parent_models, child_models=child_models,
            include_output=False, object_id_required=obj_id_req,
            child_id_required=child_id_req, parent_id_required=parent_id_req)

        # Build operation
        d_op = op.Operation(op_name, self.delete,
                            args, object_model=model,
                            parent_models=parent_models,
                            child_models=child_models,
                            elem_wrp_funcs=elem_wrp_funcs)

        self.add_operation(d_op, parent_id_field_types=parent_id_field_types,
                           child_id_field_types=child_id_field_types,
                           obj_id_field_types=obj_id_field_types)

        return d_op

    def validate_wrps_single_elem(self, wrp_list, obj_name, op_name):
        if not wrp_list:
            raise e.NoEntriesFoundError(object_type=obj_name)
        if len(wrp_list) > 1:
            raise e.MultipleEntriesError(operation=op_name)

    def setup_arguments(self):
        ''' Setup all arguments for the specified object type and operation.'''
        pvm_name = util.utf8_str(help.PROG_NAME)
        help_arg = util.utf8_str(help.HELP_ARGS)
        op_arg = util.utf8_str(help.OP_USG)

        # Set parser metadata
        self.parser.prog = util.utf8_str(pvm_name + ' ' + self.name)
        self.parser.usage = ' '.join([pvm_name, help_arg, self.name, op_arg])

        # Set parser help arguments and help arg text
        help_group = self.parser.add_argument_group(
            util.utf8_str(_('help arguments')))
        help_group.add_argument('help',
                                help=util.utf8_str(help.HELP_DESC),
                                action="store_true")
        help_group.add_argument('--help',
                                help=argparse.SUPPRESS,
                                action="help")

        self._setup_operation_arguments()

    def _setup_usage(self, id_usg, pvm_name, help_arg, oper):
        # TODO (clbush | IBM): {args} needs to be explicitly defined based on
        # the operation. For instance, vnic delete has no extra {args}
        # beyond what is defined by the usage text beforehand.
        if id_usg:
            return util.utf8_str(' '.join([
                pvm_name, help_arg, self.name, oper.op_name,
                self._identifier_usage(oper), _("{args}")]))
        else:
            return util.utf8_str(' '.join([pvm_name, help_arg, self.name,
                                           oper.op_name, _("{args}")]))

    def _setup_operation_arguments(self):
        ''' Build out the arguments for the valid operations on this
            object type.
        '''

        op_subparsers = self.parser.add_subparsers()

        pvm_name = util.utf8_str(help.PROG_NAME)
        help_arg = util.utf8_str(help.HELP_ARGS)
        # Add custom operations specified by subclass
        for oper in self.ops.values():

            usg = self._setup_usage(self._identifier_usage(oper),
                                    pvm_name, help_arg, oper)

            subparser_prog = util.utf8_str(
                ' '.join((pvm_name, self.name, oper.op_name)))

            # TODO (clbush): Look into enforcing help text arg ordering
            sp = op_subparsers.add_parser(util.utf8_str(oper.op_name),
                                          prog=util.utf8_str(subparser_prog),
                                          usage=util.utf8_str(usg),
                                          formatter_class=rthf,
                                          add_help=False)

            sp.set_defaults(func=oper.op_func)

            grps = {}
            for arg_key in oper.op_args:
                a = oper.op_args[arg_key]
                if a.group not in grps:
                    grps[a.group] = (sp.add_argument_group(a.group)
                                     if a.group else sp)
                grp = grps[a.group]
                arg_names = [a.arg_name]
                if a.short_name:
                    arg_names.append(a.short_name)
                # actions are special fields, need to set them
                # on construction
                action_k = {'action': a.action} if a.action else {}
                added_arg = grp.add_argument(*arg_names, **action_k)
                self._set_fields_on_arg(added_arg, a)

            epilog = ''

            if op.DF_NAME in oper.op_args:
                epilog += self._df_epilog(oper.op_args[op.DF_NAME].fields)

            if op.SF_NAME in oper.op_args:
                epilog += self._sf_epilog(oper.op_args[op.SF_NAME].fields)

            sp.epilog = epilog
            oper.subparser = sp

    def _id_str(self, models, root_str):
        default_str = ''

        for o_m in models or []:
            m_options = []
            if o_m != m_sys.System:
                m_options.append('<%s field>=value' % o_m.schema_type)
            if m_options:
                default_str = root_str % ' | '.join(m_options)

        return util.utf8_str(default_str)

    def _identifier_usage(self, oper):
        parent_str = self._id_str(oper.parent_models,
                                  '{--parent-id | -p} %s ')
        child_str = self._id_str(oper.child_models,
                                 '{--child-id | -c} %s ')
        obj_str = self._id_str([self.model],
                               '{--object-id | -i} %s')

        return util.utf8_str('%s' % (parent_str + child_str + obj_str)
                             if any([parent_str, child_str, obj_str]) else '')

    def _set_fields_on_arg(self, added_arg, arg):
        # argparse add_argument function parameters change
        # their behavior based on what is and is not set.
        # With the custom Argument class, only set something
        # on argparse if it was set in the Argument construction.
        if arg.required:
            added_arg.required = arg.required
        if arg.arg_help:
            added_arg.help = arg.arg_help
        if arg.arg_type:
            added_arg.type = arg.arg_type
        if arg.metavar:
            added_arg.metavar = arg.metavar
        if arg.nargs:
            added_arg.nargs = arg.nargs
        if arg.choices:
            added_arg.choices = arg.choices
        if arg.default:
            added_arg.default = arg.default

    def _df_epilog(self, d_fields):
        ''' Build up a new line delimited list of display options. '''
        epilog = '--display-fields ' + _('options') + ':\n    '
        return util.utf8_str(epilog + '\n    '.join(sorted(d_fields)) + '\n\n')

    def _sf_epilog(self, s_fields):
        ''' Build up a new line delimited list of update options. '''
        epilog = '--set-fields ' + _('options') + ':\n    '
        return util.utf8_str(epilog + '\n    '.join(sorted(s_fields)) + '\n\n')

    def dfield_val(self, wrp, df_name):
        ''' Retrieve the value of the display field from a pypowervm wrapper.

        :param wrp: The Wrapper object containing the value to retrieve.
                    Note that this may be a child wrapper of the given arg.
        :param df_name: The display field name specifying the requested value.
        '''
        if wrp not in self.out_graphs:
            self.out_graphs[wrp] = out_proc.OutputGraph(self, wrp)

        out_graph = self.out_graphs[wrp]
        d_fields = self.get_disp_fields()
        if df_name in d_fields:
            out_lists = out_graph.get_spec_disp_lists([df_name], None)
            return ','.join(
                [d_fields[df_name].out_func(o_v[0].attr_vals[df_name])
                 for o_v in out_lists
                 if o_v[0] and df_name in o_v[0].attr_vals])
        else:
            return ''

    def resize_pri_out_table(self, col_labels, out_rows):
        ''' Build and resize a pretty table for the given output rows.

        :param col_labels: A list of table column header labels
        :param out_rows: The values for each cell in each row.
        :param max_tbl_len: The maximum length of the table, defaults to 80.
        :returns: A prettytable object for the specified out_rows and columns.
        '''
        max_tbl_len = util.get_term_width()
        if self.ec.parsed_args.max_width:
            max_tbl_len = self.ec.parsed_args.max_width

        if max_tbl_len < 80:
            raise e.PriOutputMinLengthError()

        pri_out_table = pt.PrettyTable(col_labels)

        num_cols = len(col_labels)

        # Determine max length of each column
        max_cell_lens = [len(each_field) for each_field in col_labels]
        for out_row in out_rows:
            max_cell_lens = [max(prev_max_cell,
                                 len(row_cell) if row_cell else 0)
                             for (prev_max_cell, row_cell)
                             in zip(max_cell_lens, out_row)]

        # The total length of the expanded table is the sum of the length
        # of each cell value, 2 for space padding and 1 for | padding on each
        # cell, and 1 for the final | padding at the end of the table.
        total_len = sum(max_cell_lens) + (3 * num_cols + 1)

        # If we need to cut down to less than max_tbl_len chars...
        if total_len > max_tbl_len:
            # Average available length available for each cell is the specified
            # max_tbl_len minus 2 characters for padding on each cell, 1
            # character for the border on each cell, then 1 character for the
            # end of the table, all divided by the number of cells in the
            # table. NOTE: With the division, we will be losing a few
            # characters, but this will guarantee we are smaller than the max
            # length.
            max_avg_len_per_cell = ((max_tbl_len - (3 * num_cols + 1))
                                    / num_cols)

            # Determine space "saved" per cell
            saved_cell = [0 for each_field in col_labels]
            cell_ind = 0
            num_over_max = 0
            for max_cell_len in max_cell_lens:
                saved_cell[cell_ind] = max(
                    max_avg_len_per_cell - max_cell_len, 0)
                if max_cell_len >= max_avg_len_per_cell:
                    num_over_max += 1
                cell_ind += 1

            # The extra padding each "over max length" cell gets is the
            # average of all the extra space per cell that is over the max
            extra_padding = sum(saved_cell) / num_over_max

            # Cut cells over max length evenly based on max_tbl_len limit
            for out_row in out_rows:
                cell_ind = 0
                for cell_entry in out_row:
                    if len(cell_entry) > max_avg_len_per_cell:
                        if len(out_row[cell_ind]) > (max_avg_len_per_cell
                                                     + extra_padding):
                            # If we have to cut a cell, end it with the >
                            # character to indicate to a user the value has
                            # been truncated.
                            out_row[cell_ind] = (
                                six.text_type(out_row[cell_ind][0:(int(
                                    max_avg_len_per_cell +
                                    extra_padding - 1))]) + ">")
                    cell_ind += 1

        for out_row in out_rows:
            pri_out_table.add_row(out_row)

        return pri_out_table

    def _chk_and_print_help(self, op_to_exec,
                            is_help_text_requested, help_out):
        if not op_to_exec and is_help_text_requested:
            help_str = self.help
            util.print_(help_str, outfile=help_out)
            return True
        elif op_to_exec and is_help_text_requested:
            op_info = self.name_to_ops[op_to_exec]
            help_str = op_info.op_help
            if not help_str:
                help_str = op_info.subparser.format_help()
            util.print_(help_str, outfile=help_out)
            return True
        return False

    @property
    def help(self):
        return self.parser.format_help()

    def build(self, obj_type_args, user_arg=None, pass_arg=None,
              host_arg=None, port_arg=None, help_out=sys.stdout):
        ''' Function called by the parser as an entry point into object type
            operation processing.

            :param obj_type_args: A list of argument strings for argparse.
            :param user_arg: The outer-parser user auth argument, if any.
            :param pass_arg: The outer-parser password auth argument, if any.
            :param host_arg: The outer-parser host auth argument, if any.
            :param port_arg: The outer-parser port auth argument, if any.
            :param help_out: The output file for help text output.
            :returns: Boolean representing if provided args call for execution.
                      This will be false with a help argument, for instance.
        '''
        is_help_text_requested = False
        op_to_exec = None
        args_to_parse = []
        for arg_to_check in obj_type_args:
            if arg_to_check in self.name_to_ops:
                op_to_exec = arg_to_check
            if arg_to_check in ('--help', 'help'):
                is_help_text_requested = True
            else:
                args_to_parse.append(arg_to_check)

        help_printed = self._chk_and_print_help(
            op_to_exec, is_help_text_requested, help_out)

        if help_printed:
            return False

        try:
            self.parsed_args = self.parser.parse_args(args_to_parse)
        except UnicodeError:
            # This is a workaround if parse_args fails and the locale
            # needs unicode characters. Argparse's built in message printing
            # will not encode properly. We will only hit this block if
            # we attempt to print help text in the case of invalid CLI args.
            if not op_to_exec:
                help_str = self.parser.format_help()
            else:
                help_str = self.name_to_ops[op_to_exec].subparser.format_help()
            util.print_(help_str, outfile=help_out)
            return False

        self.sess_man = sm.SessionManager(user_arg=user_arg, pass_arg=pass_arg,
                                          host_arg=host_arg, port_arg=port_arg)

        exec_op = self.ops[self.parsed_args.func]
        self.ec = ExecutionContext(self.sess_man, self.parsed_args, exec_op)

        return True

    def execute(self):
        ''' Create the session adapter, query wrappers and execute the user
            specified operation.
        '''
        self.ec.sm.create_session_adapter()
        if self.ec.exec_op.query_wrps:
            self.ec.query_wrappers()
        self.ec.parsed_args.func()

    def display(self):
        ''' Common display operation function. Needs to be here so execution
            context exists when calling into output processor.

            This method is the default Operation.op_func value for the
            pvmctl <ObjectType> list operation. That op_func value is then
            used as the default func argument of argparse when that package is
            determining which function to call to process the specified
            operation.
        '''
        # If we have no wrappers to show, we should be returning an error.
        # TODO(clbush): Don't error here, need return code of 0, no stdout
        if len(self.ec.wrappers) < 1:
            raise e.NoEntriesFoundError(object_type=self.name)

        elif self.ec.parsed_args.xml:
            out_proc.raw_xml_display(self)

        elif self.ec.parsed_args.all_out:
            out_proc.raw_all_display(self)

        elif self.ec.parsed_args.display_fields:
            out_proc.specified_display(self)

        else:
            out_proc.default_display(self)

    def update(self):
        ''' Common update operation function.

            This method is the default Operation.op_func value for the
            pvmctl <ObjectType> update operation. That op_func value is then
            used as the default func argument of argparse when that package is
            determining which function to call to process the specified
            operation.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers,
                                       self.name, op.UPDATE_NAME)
        p_set_fields = self.ec.parsed_args.set_fields

        if len(p_set_fields) == 0:
            raise e.InvalidSetFieldsError(object_type=self.name,
                                          operation=op.UPDATE_NAME)

        update_retry(
            self.ec.wrappers[0], util.split_set_fields(p_set_fields), self)

    def set_field_val(self, wrp, sf_key, sf_val):
        ''' Set the value on a field of a wrapper by recursively finding it.

        :param wrp: The wrapper to set.
        :param sf_key: The key into the pregenerated set fields.
                       Generally of the form Schema.Field_Name.
        :param sf_val: The value to set for the field within the wrapper.
        :returns: True if this call sets the value, False if it has not.
        '''
        s_field_schema, s_field_name = sf_key.split('.', 1)
        schema_cmp = (six.text_type(s_field_schema) == six.text_type(
            self.get_schema(wrp.schema_type)))
        if schema_cmp:
            if not wrp.element:
                raise e.InvalidSetFieldElemNotFoundError(
                    schema=s_field_schema, obj_type=self.name)
            set_arg = self.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME]
            if sf_key not in set_arg.fields:
                raise e.InvalidSetFieldsError(operation=op.UPDATE_NAME,
                                              object_type=self.name)
            try:
                if set_arg.fields[sf_key].field_type == list:
                    setattr(wrp, s_field_name, re.split('[, ]+', sf_val))
                else:
                    setattr(wrp, s_field_name,
                            set_arg.fields[sf_key].field_type(sf_val))
            except AttributeError:
                raise e.InvalidSetFieldValueError(field=sf_key)
            return True
        else:
            for attr in (a for a in dir(wrp)
                         if a not in op_gen.NON_USER_WRAPPER_FIELD_NAMES
                         and not a.startswith('_')):
                if (isinstance(getattr(wrp, attr), e_w.Wrapper) and
                        self.set_field_val(
                            getattr(wrp, attr), sf_key, sf_val)):
                    return False

        return False

    def delete(self):
        ''' Common delete operation function.

            This method is the default Operation.op_func value for the
            pvmctl <ObjectType> delete operation. That op_func value is then
            used as the default func argument of argparse when that package is
            determining which function to call to process the specified
            operation.
        '''
        self.validate_wrps_single_elem(self.ec.wrappers,
                                       self.name, op.DELETE_NAME)
        self.ec.wrappers[0].delete()
