# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import six

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_shared_proc_pool as help_spp
from pvmcli.util.i18n import _
import pvmcli.util.exceptions as exc
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as util

from pypowervm.wrappers import logical_partition as lpar_w
from pypowervm.wrappers import managed_system as ms_w
from pypowervm.wrappers import virtual_io_server as vios_w
from pypowervm.wrappers import shared_proc_pool


PRI_DISP_FIELD_LABELS = [_("ID"), _("Name"), _("Available"), _("Max"),
                         _("Reserved"), _("Partitions")]


NA_FIELD_STR = '----'


class SharedProcessorPool(ObjectType):
    def __init__(self):
        super(SharedProcessorPool, self).__init__()

        self.add_display_operation(parent_models=[ms_w.System])
        self.add_update_operation(parent_models=[ms_w.System])
        self.add_delete_operation(parent_models=[ms_w.System])
        self._add_create_operation()

        self.setup_arguments()

    @property
    def model(self):
        return shared_proc_pool.SharedProcPool

    def _spp_in_use(self, spp_to_update, create_id):
        if not spp_to_update:
            if create_id:
                raise exc.SpecifiedSPPNotFreeError(spp_id=create_id)
            else:
                raise exc.NoFreeSPPError()

    def create(self):
        # Get all SPPs
        all_spps = self.model.get(
            self.ec.sm.adapter, parent_type=ms_w.System.schema_type,
            parent_uuid=self.ec.msys.uuid)
        create_id = self.ec.parsed_args.spp_id
        spp_to_update = None
        for spp in all_spps:
            if (spp.max_proc_units == 0 and not spp.is_default and
                    ((not create_id) or spp.id == create_id)):
                spp_to_update = spp
                break

        self._spp_in_use(spp_to_update, create_id)

        spp_to_update.name = self.ec.parsed_args.name
        spp_to_update.max_proc_units = self.ec.parsed_args.max
        if self.ec.parsed_args.reserved:
            spp_to_update.pend_rsrv_proc_units = self.ec.parsed_args.reserved

        spp_to_update.update()

    def _add_create_operation(self):
        spp_id_help = _("ID of the shared processor pool. Will automatically "
                        "choose next available if not specified.")
        spp_name_help = ("Name of the shared processor pool.")
        max_help = ("Floating point number of max processor units.")
        reserved_help = ("Floating point number of pending reserved processor "
                         "units.")
        crt_args = o_g.default_args(self, parent_models=[ms_w.System])
        crt_args.extend([op.Argument('--spp-id', arg_help=spp_id_help,
                                     required=False, arg_type=int),
                         op.Argument('--name', arg_help=spp_name_help,
                                     required=True, arg_type=str),
                         op.Argument('--max', arg_help=max_help,
                                     required=True, arg_type=float),
                         op.Argument('--reserved', arg_help=reserved_help,
                                     arg_type=float, required=False)])
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def filter_invalid_wrps(self):
        ''' Filter out all non-default pools that have max_proc_units == 0 '''
        self.ec.wrappers = [wrp for wrp in self.ec.wrappers
                            if wrp.max_proc_units != 0 or wrp.is_default]

    def display(self):
        ''' Overridden display method to filter out non-user wrappers. '''
        self.filter_invalid_wrps()
        super(SharedProcessorPool, self).display()

    def delete(self):
        self.validate_wrps_single_elem(self.ec.wrappers,
                                       self.name, op.DELETE_NAME)

        # Will attempt to 'delete' even with assigned partitions. The REST
        # layer will populate the proper HTTP error up to pvmctl.
        self.ec.wrappers[0].pend_rsrv_proc_units = 0
        self.ec.wrappers[0].max_proc_units = 0
        self.ec.wrappers[0].update()

    def spp_out_rows(self):
        out_rows = []
        all_parts = lpar_w.LPAR.get(self.ec.sm.adapter)

        all_parts.extend(vios_w.VIOS.get(self.ec.sm.adapter))
        pool_to_part_map = {}

        for shr_part in [part for part in all_parts
                         if not part.proc_config.has_dedicated]:
            part_pool_id = shr_part.proc_config.shared_proc_cfg.pool_id
            if part_pool_id in pool_to_part_map:
                pool_to_part_map[part_pool_id].append(shr_part.name)
            else:
                pool_to_part_map[part_pool_id] = [shr_part.name]

        for wrp in self.ec.wrappers:
            out_row = [six.text_type(wrp.id), wrp.name]
            if wrp.is_default:
                out_row.extend([NA_FIELD_STR, NA_FIELD_STR, NA_FIELD_STR])
            else:
                out_row.extend([six.text_type(wrp.avail_proc_units),
                                six.text_type(wrp.max_proc_units),
                                six.text_type(wrp.pend_rsrv_proc_units)])

            if wrp.id in pool_to_part_map:
                for spp_part in util.group_join_list(pool_to_part_map[wrp.id],
                                                     num_cols=1):
                    out_row.append(spp_part)
                    out_rows.append(out_row)
                    out_row = ["", "", "", "", ""]
            else:
                out_row.append("")
                out_rows.append(out_row)
        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.spp_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Shared Processor Pools"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_spp.help_text
