# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import six

from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_sriov_adapter as help_sriov
import pvmcli.util.exceptions as e
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as op_gen
import pvmcli.util.output_processor as out_proc
import pvmcli.util.util as cli_util

import pypowervm.tasks.sriov as sriov_task
from pypowervm.wrappers import iocard as iocard
from pypowervm.wrappers import managed_system as msys
import pypowervm.util as util


PPORT_SCHEMA = "PhysicalPort"
PRI_DISP_FIELD_LABELS = [_("ID"), _("State"), _("Mode"), _("Loc Code"),
                         _("Port Stat"), _("Port Lbl"), _("Av Cap"),
                         _("Av LPs")]
LINK_STATUS_UP = _("Up")
LINK_STATUS_DOWN = _("Down")


class SRIOVAdapter(ObjectType):
    def __init__(self):
        super(SRIOVAdapter, self).__init__()

        self.add_field_models([iocard.SRIOVConvPPort, iocard.SRIOVEthPPort])

        self.remap_schema(iocard.SRIOVConvPPort.schema_type, PPORT_SCHEMA)
        self.remap_schema(iocard.SRIOVEthPPort.schema_type, PPORT_SCHEMA)

        # Exclude this field as it is a wrapper reference from PhysicalPort
        # back to its parent SRIOVAdapter wrapper. Including this field
        # causes an infinite recursion in OutputGraph.init(), and
        # is not a useful display field as all the display fields for
        # SRIOVAdapter are already covered by said parent wrapper.
        self.exc_disp_fields.append('PhysicalPort.sriov_adap')

        d_op = self.add_display_operation(
            parent_models=[msys.System],
            elem_wrp_funcs=[msys.System.asio_config,
                            msys.ASIOConfig.sriov_adapters])

        self.add_update_operation()

        d_op.op_args[op.DF_NAME].fields[
            'PhysicalPort.link_status'].out_func = self._link_status_out
        all_cap_field = d_op.op_args[op.DF_NAME].fields[
            'PhysicalPort.allocated_capacity']
        all_cap_field.out_func = util.sanitize_percent_for_api
        d_op.op_args[op.DF_NAME].fields[
            'SRIOVAdapter.id'].out_func = out_proc.hex_no_prefix

        self.setup_arguments()

    @property
    def model(self):
        return iocard.SRIOVAdapter

    def add_update_operation(self):
        force_help = _("If specified, physical ports that are in use by "
                       "vNICs will update. If not specified, the update "
                       "operation will fail if those ports are in use.")
        loc_help = _("The physical location code of a physical port to be "
                     "updated.")
        parent_models = [msys.System]
        elem_wrp_funcs = [msys.System.asio_config,
                          msys.ASIOConfig.sriov_adapters]
        args = op_gen.default_args(self, parent_models=parent_models)

        # Build arguments
        args.append(op_gen.set_argument(self))
        args.extend([op.Argument('--port-loc', arg_help=loc_help,
                                 required=False, arg_type=str),
                     op.Argument('--force', arg_help=force_help,
                                 required=False, action="store_true")])
        # Build operation
        u_op = op.Operation(
            op.UPDATE_NAME, self.update, args, object_model=self.model,
            parent_models=parent_models, elem_wrp_funcs=elem_wrp_funcs)

        self.add_operation(u_op)

    def _filter_wrps_loc(self, loc_code):
        for s_adpt in self.ec.wrappers:
            for p_port in s_adpt.phys_ports:
                if p_port.loc_code == loc_code:
                    # There should only be one phys port with a given loc
                    # code, so if we find it just return that adapter.
                    return [s_adpt]
        return []

    def update(self):
        sfs = cli_util.split_set_fields(self.ec.parsed_args.set_fields)

        if len(sfs) == 0:
            raise e.InvalidSetFieldsError(object_type=self.name,
                                          operation=op.UPDATE_NAME)

        def safe_update(sys_w):
            # If loc is specified, filter out unrelated wrappers
            if self.ec.parsed_args.port_loc:
                self.ec.wrappers = self._filter_wrps_loc(
                    self.ec.parsed_args.port_loc)

            self.validate_wrps_single_elem(
                self.ec.wrappers, self.model.schema_type, op.UPDATE_NAME)
            ret_val = False

            # Raise error if we're setting a physical port and loc is None
            # Or find the physical port we're trying to update
            pport_upd = None

            for set_field in sfs:
                if set_field.startswith(PPORT_SCHEMA):
                    if not self.ec.parsed_args.port_loc:
                        raise e.PPortUpdateRequiresLocError()
                    else:
                        pport_upd = sriov_task.find_pport(
                            sys_w, self.ec.parsed_args.port_loc)
                        break

            for set_field in sfs:
                if set_field.startswith(PPORT_SCHEMA):
                    self.set_field_val(pport_upd, set_field, sfs[set_field])
                    ret_val = True
                else:
                    self.set_field_val(
                        self.ec.wrappers[0], set_field, sfs[set_field])
                    ret_val = True
            return ret_val

        sriov_task.safe_update_pports(
            self.ec.elem_p_wrappers[0], safe_update,
            force=self.ec.parsed_args.force)

    def sriov_out_rows(self):
        out_rows = []
        for wrp in self.ec.wrappers:
            sriov_id = self.dfield_val(wrp, "SRIOVAdapter.sriov_adap_id")
            if sriov_id == "":
                sriov_id = "N/A"
            out_row = [sriov_id]
            adp_mode = self.dfield_val(wrp, "SRIOVAdapter.mode")
            adp_state = self.dfield_val(wrp, "SRIOVAdapter.state")
            if adp_mode == iocard.SRIOVAdapterMode.DEDICATED:
                out_row.append("")
            else:
                out_row.append(adp_state)
            out_row.append(adp_mode)
            if (adp_mode == iocard.SRIOVAdapterMode.DEDICATED or
                    adp_state == iocard.SRIOVAdapterState.MISSING or
                    adp_state == iocard.SRIOVAdapterState.MISMATCH):
                out_row.extend([cli_util.utf8_str(wrp.phys_loc_code),
                                "", "", "", ""])
                out_rows.append(out_row)
            elif wrp.phys_ports:
                for p_port in wrp.phys_ports:
                    avail_cap = (1 - p_port.allocated_capacity) * 100
                    avail_lps = p_port.cfg_max_lps - p_port.cfg_lps
                    out_row.extend([six.text_type(p_port.loc_code),
                                    self._link_status_out(p_port.link_status),
                                    six.text_type(p_port.label),
                                    "%s%%" % six.text_type(avail_cap),
                                    six.text_type(avail_lps)])
                    out_rows.append(out_row)
                    out_row = ["", "", ""]
            else:
                out_row.extend(["", "", "", "", ""])
                out_rows.append(out_row)
        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.sriov_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("SR-IOV Adapters"), 'n_tbl': po_table.get_string()}

    def _link_status_out(self, link_status):
        return LINK_STATUS_UP if link_status else LINK_STATUS_DOWN

    @property
    def help(self):
        return help_sriov.help_text
