# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import six

from pvmcli.command.object_type import ObjectType
import pvmcli.command.base_partition as bp_c
import pvmcli.command.operation as op
import pvmcli.command.sriov_adapter as sriov_c
import pvmcli.help.help_sriov_ethlp as help_ethlp
import pvmcli.util.exceptions as e
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.output_processor as out_proc
import pvmcli.util.util as cli_util

from pypowervm.tasks import sriov as sriov_task
import pypowervm.util as util
from pypowervm.wrappers import logical_partition as lpar_w
from pypowervm.wrappers import iocard as iocard
from pypowervm.wrappers import virtual_io_server as vios_w


NON_V_PRI_OUT_LABELS = [_("Loc. Code"), _("PVID"),
                        _("Partition"), _("Cap."), _("MAC")]
V_PRI_OUT_LABELS = [_("LPAR"), _("VIOS"), _("Loc. Code"), _("Active"),
                    _("Fail. Pri."), _("Cap.")]
M_PRI_OUT_LABELS = [_("Loc. Code")]
STAR_CHR = six.text_type('*')
INV_STR = _("MISCONFIGURED")


class SRIOVLogicalPort(ObjectType):
    def __init__(self):
        super(SRIOVLogicalPort, self).__init__()

        d_op = self.add_display_operation(parent_models=[vios_w.VIOS,
                                                         lpar_w.LPAR])
        d_op.set_d_field_out_func(iocard.SRIOVEthLPort, 'lport_id',
                                  out_func=out_proc.hex_no_prefix)
        cap_field = d_op.op_args[op.DF_NAME].fields[
            'SRIOVEthernetLogicalPort.cfg_capacity']
        cap_field.out_func = util.sanitize_percent_for_api
        max_cap = d_op.op_args[op.DF_NAME].fields[
            'SRIOVEthernetLogicalPort.cfg_max_capacity']
        max_cap.out_func = util.sanitize_percent_for_api
        vnic_help = ("If specified, only show the logical ports with a "
                     "VNIC port usage.")
        d_op.add_arg(op.Argument('--vnic', action='store_true',
                                 default=False, arg_help=vnic_help))
        self.add_delete_operation(parent_models=[vios_w.VIOS, lpar_w.LPAR])

        self._add_create_operation()

        self.setup_arguments()

    def _add_create_operation(self):
        loc_help = _("The location code of the physical port backing this "
                     "logical port. For example: "
                     "`U78C7.001.RCH0004-P1-C8-T1`, available via `pvmctl "
                     "sriov list -d PhysicalPort.loc_code`.")
        pvid_help = _("The port VLAN identifier for this logical port. Any "
                      "untagged traffic passing through this port will have "
                      "this VLAN tag added.")
        promisc_help = _("Specify this argument if using the logical port for "
                         "bridging (e.g. SEA, OVS, etc.); do not if assigning "
                         "directly to an LPAR.  Only one logical port per "
                         "physical port may be promiscuous.")
        cap_help = _("The configured capacity of the logical port as a value "
                     "between 0.00-100.00 representing the configured "
                     "capacity percentage. This represents the fraction of "
                     "the physical port's bandwidth this logical port will "
                     "receive. The provided value must also be a multiple "
                     "of the physical port's capacity granularity.")
        max_help = _("The maximum capacity of the logical port as a value "
                     "between 0.00-100.00 representing the maximum "
                     "capacity percentage. This represents the maximum of "
                     "the physical port's bandwidth this logical port will "
                     "receive.")
        crt_args = o_g.default_args(
            self, parent_models=[vios_w.VIOS, lpar_w.LPAR],
            parent_id_required=True)
        crt_args.extend([op.Argument('--loc', arg_help=loc_help,
                                     required=True, arg_type=str),
                         op.Argument('--pvid', arg_help=pvid_help,
                                     required=False, arg_type=int),
                         op.Argument('--promisc', arg_help=promisc_help,
                                     action="store_true", required=False),
                         op.Argument('--capacity', arg_help=cap_help,
                                     arg_type=str, required=False),
                         op.Argument('--max-capacity', arg_help=max_help,
                                     arg_type=str, required=False)])
        self.add_operation(op.Operation(
            op.CREATE_NAME, self.create, op_args=crt_args,
            object_model=self.model, parent_models=[vios_w.VIOS, lpar_w.LPAR]))

    def _set_cap(self, p_cap, arg):
        cap = cli_util.parse_float_as_pct(p_cap) if p_cap else None
        if cap and (cap < 0 or cap > 1):
            raise e.CapacityOutOfBoundsError(arg=arg)
        return cap

    def create(self):
        self.validate_wrps_single_elem(self.ec.elem_p_wrappers,
                                       bp_c.PARTITION_NAME, op.CREATE_NAME)

        pport = sriov_task.find_pport(
            self.ec.msys, self.ec.parsed_args.loc)
        if not pport:
            raise e.NoEntriesFoundError(object_type=sriov_c.PPORT_SCHEMA)

        pvid = self.ec.parsed_args.pvid
        promisc = self.ec.parsed_args.promisc
        cap = self._set_cap(self.ec.parsed_args.capacity, "capacity")
        max_cap = self._set_cap(self.ec.parsed_args.max_capacity,
                                "max_capacity")

        new_port = iocard.SRIOVEthLPort.bld(
            self.ec.sm.adapter, sriov_adap_id=pport.sriov_adap_id,
            pport_id=pport.port_id, pvid=pvid, is_promisc=promisc,
            cfg_capacity=cap, max_capacity=max_cap)

        new_port.create(parent=self.ec.elem_p_wrappers[0])

    def _proc_vnic_row(self, lp_h, lp_href_to_vbd, lp_w_map, wrp):
        if lp_h in lp_href_to_vbd:
            for vbd, lp_w in lp_href_to_vbd[lp_h]:
                if lp_w.name not in lp_w_map:
                    lp_w_map[lp_w.name] = []
                lp_w_map[lp_w.name].append([
                    cli_util.vios_by_uri(self.ec.sm.adapter,
                                         vbd.vios_href).name,
                    wrp.loc_code,
                    STAR_CHR if vbd.is_active else '',
                    six.text_type(vbd.failover_pri),
                    util.sanitize_percent_for_api(
                        wrp.cfg_capacity)])
        else:
            if INV_STR not in lp_w_map:
                lp_w_map[INV_STR] = []
            lp_w_map[INV_STR].append([wrp.loc_code])

    def vnic_out_rows(self, lp_href_to_vbd):
        vnic_out_rows = []
        misc_out_rows = []
        lp_w_map = {}
        for part_wrp in sorted(self.ec.parent_wrp_to_children,
                               key=lambda w: w.name):
            for wrp in self.ec.parent_wrp_to_children[part_wrp]:
                if wrp.vnic_port_usage != iocard.VNICPortUsage.NOT_VNIC:
                    self._proc_vnic_row(util.dice_href(wrp.href),
                                        lp_href_to_vbd, lp_w_map, wrp)
        for lp_w_name in sorted(lp_w_map):
            out_row = [lp_w_name]
            for lp_data in lp_w_map[lp_w_name]:
                if lp_w_name == INV_STR:
                    misc_out_rows.append(lp_data)
                else:
                    out_row.extend(lp_data)
                    vnic_out_rows.append(out_row)
                out_row = ['']
        return vnic_out_rows, misc_out_rows

    def lp_out_rows(self):
        out_rows = []
        for part_wrp in self.ec.parent_wrp_to_children:
            for wrp in self.ec.parent_wrp_to_children[part_wrp]:
                if wrp.vnic_port_usage == iocard.VNICPortUsage.NOT_VNIC:
                    out_row = [wrp.loc_code, six.text_type(wrp.pvid),
                               part_wrp.name,
                               util.sanitize_percent_for_api(wrp.cfg_capacity),
                               wrp.cur_mac]
                    out_rows.append(out_row)
        return out_rows

    def pri_output(self):
        lp_href_to_vbd = {}
        all_lpars = lpar_w.LPAR.get(self.ec.sm.adapter)
        for a_lp in all_lpars:
            for vnic_wrp in iocard.VNIC.get(self.ec.sm.adapter, parent=a_lp):
                for v_bdev in vnic_wrp.back_devs:
                    lp_href = util.dice_href(v_bdev.lport_href)
                    if lp_href not in lp_href_to_vbd:
                        lp_href_to_vbd[lp_href] = []
                    lp_href_to_vbd[lp_href].append((v_bdev, a_lp))

        vnic_out_rows, misc_out_rows = self.vnic_out_rows(lp_href_to_vbd)
        v_tbl = self.resize_pri_out_table(
            V_PRI_OUT_LABELS, vnic_out_rows) if vnic_out_rows else ''
        v_ttl = _("SR-IOV Ethernet Logical Ports (VNIC)")
        v_out = "%(ttl)s\n%(tbl)s\n" % {
            'ttl': v_ttl, 'tbl': v_tbl.get_string()} if v_tbl else ''

        lp_out = ''
        if not self.ec.parsed_args.vnic:
            lp_out_rows = self.lp_out_rows()
            lp_tbl = self.resize_pri_out_table(
                NON_V_PRI_OUT_LABELS, lp_out_rows) if lp_out_rows else ''
            lp_ttl = _("SR-IOV Ethernet Logical Ports (Non-VNIC)")
            lp_out = "%(ttl)s\n%(tbl)s\n" % {
                'ttl': lp_ttl, 'tbl': lp_tbl.get_string()} if lp_tbl else ''

        m_tbl = self.resize_pri_out_table(
            M_PRI_OUT_LABELS, misc_out_rows) if misc_out_rows else ''
        m_ttl = _("Misconfigured Ethernet Logical Ports")
        m_out = "%(ttl)s\n%(tbl)s\n" % {
            'ttl': m_ttl, 'tbl': m_tbl.get_string()} if m_tbl else ''

        return "%(v_out)s%(lp_out)s%(m_out)s" % {'lp_out': lp_out,
                                                 'v_out': v_out,
                                                 'm_out': m_out}

    @property
    def model(self):
        return iocard.SRIOVEthLPort

    @property
    def help(self):
        return help_ethlp.help_text
