# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.base_partition as bp
import pvmcli.command.operation as op
import pvmcli.help.help_virtual_ethernet_adapter as help_vea
import pvmcli.util.exceptions as e
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as u

import pypowervm.wrappers.logical_partition as l
import pypowervm.wrappers.managed_system as ms_w
import pypowervm.wrappers.network as n
import pypowervm.wrappers.virtual_io_server as v


PRI_DISP_FIELD_LABELS = [_("PVID"), _("VSwitch"), _("LPAR"), _("MAC"),
                         _("Slot"), _("Trunk"), _("Tagged VLANs")]


VEA_SCHEMA = "VirtualEthernetAdapter"


class VirtualEthernetAdapter(ObjectType):
    def __init__(self):
        super(VirtualEthernetAdapter, self).__init__()

        self.remap_schema(n.CNA.schema_type, VEA_SCHEMA)

        self.exc_disp_fields.extend(
            ['VirtualEthernetAdapter.dev_name',
             'VirtualEthernetAdapter.vswitch_uri'])
        self.exc_set_fields.extend(
            ['VirtualEthernetAdapter.mac',
             'VirtualEthernetAdapter.uuid',
             'VirtualEthernetAdapter.vswitch_uri'])
        self.add_display_operation(parent_models=[l.LPAR, v.VIOS])
        self.add_update_operation(parent_models=[l.LPAR, v.VIOS])
        self.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME].fields[
            'VirtualEthernetAdapter.tagged_vlans'].field_type = list
        self.add_delete_operation(parent_models=[l.LPAR, v.VIOS])
        self.add_create_operation()

        self.setup_arguments()

    def _bld_vea_out_rows(self, wrp, vsw):
        out_rows = []
        vlan_3_list = u.group_join_list(wrp.tagged_vlans)
        out_row = [self.dfield_val(wrp,
                                   "VirtualEthernetAdapter.pvid"), vsw,
                   self.dfield_val(wrp,
                                   "VirtualEthernetAdapter.lpar_id"),
                   self.dfield_val(wrp,
                                   "VirtualEthernetAdapter.mac"),
                   self.dfield_val(wrp,
                                   "VirtualEthernetAdapter.slot"),
                   self.dfield_val(wrp,
                                   "VirtualEthernetAdapter.is_trunk"),
                   vlan_3_list[0]]
        out_rows.append(out_row)
        # Add empty rows for each tagged VLAN
        for vlan_3 in vlan_3_list[1:]:
            out_rows.append(['', '', '', '', '', '', vlan_3])
        return out_rows

    def vea_out_rows(self):
        vsw_to_wrp = {}
        for wrp in self.ec.wrappers:
            if u.vswitch_by_uri(self.ec.sm.adapter,
                                wrp.vswitch_uri).name in vsw_to_wrp:
                vsw_to_wrp[
                    u.vswitch_by_uri(self.ec.sm.adapter,
                                     wrp.vswitch_uri).name].append(wrp)
            else:
                vsw_to_wrp[u.vswitch_by_uri(
                    self.ec.sm.adapter, wrp.vswitch_uri).name] = [wrp]

        out_rows = []
        for vsw in sorted(vsw_to_wrp):
            wrps = vsw_to_wrp[vsw]
            pvid_to_wrp = {}
            for wrp in wrps:
                if wrp.pvid in pvid_to_wrp:
                    pvid_to_wrp[wrp.pvid].append(wrp)
                else:
                    pvid_to_wrp[wrp.pvid] = [wrp]

            for pvid in sorted(pvid_to_wrp):
                for wrp in pvid_to_wrp[pvid]:
                    out_rows.extend(self._bld_vea_out_rows(wrp, vsw))

        return out_rows

    @property
    def model(self):
        return n.CNA

    def add_create_operation(self):
        pvid_help = _("Required. Primary VLAN ID for this adapter.")
        vswitch_help = _("Required. Name of the Virtual Switch that supports "
                         "this adapter.")
        slot_help = _("The slot on the client LPAR that should be used. If "
                      "not specified, the next available high slot will be "
                      "used.")
        mac_help = _("Optional user provided MAC address. If not provided "
                     "the system will generate one.")
        vlans_help = _("Set of additional VLANs for this adapter. The limit "
                       "is 20 additional VLANs per adapter. "
                       "Example: --vlans 1 2 3 4")
        trunk_help = _("USED IN VERY RARE CIRCUMSTANCES ONLY. Integer trunk "
                       "priority value to set if this adapter is to be "
                       "created as a trunk adapter on a Virtual I/O Server. "
                       "If this is set, this is not created as a standard "
                       "client network adapter.")
        crt_args = o_g.default_args(self, parent_models=[l.LPAR, v.VIOS],
                                    parent_id_required=True)
        crt_args.extend([op.Argument('--pvid', arg_type=int, required=True,
                                     arg_help=help.wrap_help(pvid_help)),
                         op.Argument('--vswitch', arg_type=str, required=True,
                                     arg_help=help.wrap_help(vswitch_help)),
                         op.Argument('--slot', arg_type=int,
                                     required=False,
                                     arg_help=help.wrap_help(slot_help)),
                         op.Argument('--mac', arg_type=str, required=False,
                                     arg_help=help.wrap_help(mac_help)),
                         op.Argument('--vlans', arg_type=int, nargs='+',
                                     required=False,
                                     arg_help=help.wrap_help(vlans_help)),
                         op.Argument('--trunk-pri', arg_type=int,
                                     required=False,
                                     arg_help=help.wrap_help(trunk_help))])
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[l.LPAR, v.VIOS]))

    def create(self):
        self.validate_wrps_single_elem(self.ec.elem_p_wrappers,
                                       bp.PARTITION_NAME, op.CREATE_NAME)

        vea_pvid = self.ec.parsed_args.pvid

        vea_vswitch_name = self.ec.parsed_args.vswitch
        vswitch_list = n.VSwitch.search(self.ec.sm.adapter,
                                        parent_type=ms_w.System.schema_type,
                                        parent_uuid=self.ec.msys.uuid,
                                        name=vea_vswitch_name)

        vea_vswitch = vswitch_list[0] if vswitch_list else None
        if not vea_vswitch:
            raise e.NoEntriesFoundError(object_type=n.VSwitch.schema_type)

        vea_slot = self.ec.parsed_args.slot
        vea_mac = self.ec.parsed_args.mac
        vea_vlans = self.ec.parsed_args.vlans
        vea_trunk_pri = self.ec.parsed_args.trunk_pri

        new_vea = n.CNA.bld(adapter=self.ec.sm.adapter, pvid=vea_pvid,
                            vswitch_href=vea_vswitch.related_href,
                            slot_num=vea_slot, mac_addr=vea_mac,
                            addl_tagged_vlans=vea_vlans,
                            trunk_pri=vea_trunk_pri)

        parent_schema_type = self.ec.elem_p_wrappers[0].schema_type
        res = new_vea.create(parent_type=parent_schema_type,
                             parent_uuid=self.ec.elem_p_wrappers[0].uuid)

        self.ec.wrappers = [res]
        # Only display if a user specifies specific display fields
        if (self.ec.parsed_args.display_fields
                or self.ec.parsed_args.all_out):
            self.display()

    def update(self):
        # If tagged vlans are set, make sure tagged vlan supported is True

        # Gather all user provided fields to set
        set_fields = {}
        for psf in self.ec.parsed_args.set_fields:
            key, val = psf.split('=', 1)
            set_fields[key] = val

        # If tagged vlans are specified, turn on tagged vlan support
        if ('VirtualEthernetAdapter.tagged_vlans' in set_fields
                and 'VirtualEthernetAdapter.is_tagged_vlan_supported'
                not in set_fields):
                    self.ec.parsed_args.set_fields.append(
                        'VirtualEthernetAdapter.is_tagged_vlan_supported=True')

        # Call generic update method
        super(VirtualEthernetAdapter, self).update()

    @property
    def name(self):
        return VEA_SCHEMA

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.vea_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual Ethernet Adapters"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_vea.help_text
