# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_virtual_fc_mapping as help_vfc
from pvmcli.util.i18n import _
import pvmcli.util.util as util
import pvmcli.util.exceptions as e
import pvmcli.util.op_generator as o_g

import pypowervm.tasks.vfc_mapper as vfc_map
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.storage as s
import pypowervm.wrappers.virtual_io_server as vios_w
import pypowervm.wrappers.logical_partition as l_w

PRI_DISP_FIELD_LABELS = [_("LPAR"), _("LPAR Slot"), _("VIOS"), _("VIOS Slot"),
                         _("Port Name"), _("Adpt. Name")]

search_args = [op.Argument('--vios-slot', required=False, arg_type=int),
               op.Argument('--adpt', required=False, arg_type=str),
               op.Argument('--lpar', required=False, arg_type=str),
               op.Argument('--port', required=False, arg_type=str)]


class VFCStorageMapping(ObjectType):
    def __init__(self):
        super(VFCStorageMapping, self).__init__()

        self.add_field_models([s.VFCClientAdapterElement,
                               s.VFCServerAdapterElement,
                               bp.PhysFCPort])

        list_elem_wrp_funcs = [vios_w.VIOS.vfc_mappings]
        self.add_display_operation(model=self.model,
                                   parent_models=[vios_w.VIOS],
                                   elem_wrp_funcs=list_elem_wrp_funcs)
        self._add_create_operation()
        self._add_delete_operation()
        self._add_remap_operation()
        self.setup_arguments()

    @property
    def model(self):
        return vios_w.VFCMapping

    def fc_out_rows(self):
        out_rows = []
        for wrp in self.ec.wrappers:
            lpar_id = self.dfield_val(wrp, "ServerAdapter.lpar_id")
            vios_id = self.dfield_val(wrp, "ServerAdapter.vios_id")
            out_rows.append([
                util.lpar_name_by_id(self.ec.sm.adapter,
                                     lpar_id) if lpar_id else "",
                self.dfield_val(wrp, "ServerAdapter.lpar_slot_num"),
                util.vios_name_by_id(self.ec.sm.adapter,
                                     vios_id) if vios_id else "",
                self.dfield_val(wrp, "ServerAdapter.vios_slot_num"),
                self.dfield_val(wrp, "PhysicalFibreChannelPort.name"),
                self.dfield_val(wrp, "ServerAdapter.name")])
        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.fc_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual FC Mappings"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_vfc.help_text

    def create(self):
        ''' Create operation processing function. '''
        self.validate_wrps_single_elem(
            self.ec.elem_p_wrappers, vios_w.VIOS.schema_type, op.CREATE_NAME)

        # Find LPAR
        found_lpars = self.ec.sm.split_arg_search(
            self.ec.parsed_args.lpar, l_w.LPAR)
        self.validate_wrps_single_elem(
            found_lpars, l_w.LPAR.schema_type, op.CREATE_NAME)

        pfc_port_name = self.ec.parsed_args.pfcport
        vios = self.ec.elem_p_wrappers[0]
        pfc_wwpn = vfc_map.find_pfc_wwpn_by_name(vios, pfc_port_name)
        if pfc_wwpn is None:
            raise e.NoPFCPortFoundError(arg_type='name',
                                        port_arg=pfc_port_name,
                                        vios_name=vios.name)
        port_map = vfc_map.derive_base_npiv_map([vios], [pfc_wwpn], 1)
        vfc_map.add_map(vios, None, found_lpars[0].uuid, port_map[0])
        vios.update()

    def _add_create_operation(self):
        crt_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        crt_args.extend([op.Argument('--lpar', required=True, arg_type=str),
                         op.Argument('--pfcport', required=True,
                                     arg_type=str)])

        list_elem_wrp_funcs = [vios_w.VIOS.vfc_mappings]
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=list_elem_wrp_funcs,
                                        op_help=help_vfc.help_create))

    def delete(self):
        ''' Delete operation processing function. '''
        self.validate_wrps_single_elem(
            self.ec.elem_p_wrappers, vios_w.VIOS.schema_type, op.DELETE_NAME)

        vios = self.ec.elem_p_wrappers[0]
        port_dict = dict()
        serv_dict = dict()

        # Find LPAR ID
        if self.ec.parsed_args.lpar:
            found_lpars = self.ec.sm.split_arg_search(
                self.ec.parsed_args.lpar, l_w.LPAR)
            self.validate_wrps_single_elem(
                found_lpars, l_w.LPAR.schema_type, op.DELETE_NAME)
            serv_dict['lpar_id'] = str(found_lpars[0].id)

        if self.ec.parsed_args.vios_slot:
            serv_dict['vios_slot_num'] = str(self.ec.parsed_args.vios_slot)

        if self.ec.parsed_args.adpt:
            id_type, id_val = self.ec.parsed_args.adpt.split('=', 1)
            serv_dict[id_type] = id_val

        if self.ec.parsed_args.port:
            id_type, id_val = self.ec.parsed_args.port.split('=', 1)
            port_dict[id_type] = id_val

        found_maps = self._find_maps(
            vios.vfc_mappings, port=port_dict, s_adpt=serv_dict)
        if len(found_maps) != 1:
            raise e.SingleMappingEntryDeleteError(
                type='VFC', num_mappings=len(found_maps))

        vios.vfc_mappings.remove(found_maps[0])
        vios.update()

    def _add_delete_operation(self):
        del_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        del_args.extend(search_args)

        list_elem_wrp_funcs = [vios_w.VIOS.vfc_mappings]
        self.add_operation(op.Operation(op.DELETE_NAME, self.delete,
                                        op_args=del_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=list_elem_wrp_funcs,
                                        op_help=help_vfc.help_delete))

    def remap(self):
        ''' Remap operation processing function. '''
        self.validate_wrps_single_elem(
            self.ec.elem_p_wrappers, vios_w.VIOS.schema_type, op.REMAP_NAME)

        vios = self.ec.elem_p_wrappers[0]
        port_dict = dict()
        serv_dict = dict()

        # Find LPAR ID
        if self.ec.parsed_args.lpar:
            found_lpars = self.ec.sm.split_arg_search(
                self.ec.parsed_args.lpar, l_w.LPAR)
            self.validate_wrps_single_elem(
                found_lpars, l_w.LPAR.schema_type, op.REMAP_NAME)
            serv_dict['lpar_id'] = str(found_lpars[0].id)

        if self.ec.parsed_args.vios_slot:
            serv_dict['vios_slot_num'] = str(self.ec.parsed_args.vios_slot)

        if self.ec.parsed_args.adpt:
            id_type, id_val = self.ec.parsed_args.adpt.split('=', 1)
            serv_dict[id_type] = id_val

        if self.ec.parsed_args.port:
            id_type, id_val = self.ec.parsed_args.port.split('=', 1)
            port_dict[id_type] = id_val

        found_maps = self._find_maps(
            vios.vfc_mappings, port=port_dict, s_adpt=serv_dict)
        if len(found_maps) != 1:
            raise e.SingleMappingEntryDeleteError(
                type='VFC', num_mappings=len(found_maps))

        # Make sure new port for remap actually exists
        new_port_atr, new_port_val = self.ec.parsed_args.new_port.split('=', 1)
        for port in vios.pfc_ports:
            if getattr(port, new_port_atr) == new_port_val:
                new_port_name = port.name
                break
        else:
            raise e.NoPFCPortFoundError(arg_type=new_port_atr,
                                        port_arg=new_port_val,
                                        vios_name=vios.name)

        mod_map_idx = vios.vfc_mappings.index(found_maps[0])
        new_bp = bp.PhysFCPort.bld_ref(
            vios.adapter, new_port_name, ref_tag='Port')
        vios.vfc_mappings[mod_map_idx].backing_port = new_bp
        vios.update()

    def _add_remap_operation(self):
        rmp_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        rmp_args.extend(search_args)
        rmp_args.append(op.Argument('--new-port', required=True, arg_type=str))

        list_elem_wrp_funcs = [vios_w.VIOS.vfc_mappings]
        self.add_operation(op.Operation(op.REMAP_NAME, self.remap,
                                        op_args=rmp_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=list_elem_wrp_funcs,
                                        op_help=help_vfc.help_remap))

    def _find_maps(self, map_list, port=None, s_adpt=None):
        """Filter a list of VFC mappings based on args passed.

        :param map_list: The mappings to filter.  Iterable of VFCMapping.
        :param port: (Optional, Default=None) Dict of port attributes to match
                     each mapping to.
                     Example: {'name':'fcs0'}
        :param s_adpt: (Optional, Default=None) Dict of server adapter
                       attributes to match each mapping to.
                       Example: {'name':'vfchost3', 'vios_slot_num':'11'}
        :return: A list comprising the subset of vfc mappings whose attributes
                 match the ones specified in the params.
        """
        def is_match(vmap, prop, pdict):
            map_prop = getattr(vmap, prop)

            if not map_prop:
                return False

            for key, val in pdict.items():
                if str(getattr(map_prop, key)) != val:
                    return False

            return True

        matching_maps = []

        for vmap in map_list:
            if port and not is_match(vmap, 'backing_port', port):
                continue

            if s_adpt and not is_match(vmap, 'server_adapter', s_adpt):
                continue

            # If we made it here...passed all matching criteria specified
            matching_maps.append(vmap)

        return matching_maps
