# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import pvmcli.command.operation as op
import pvmcli.help.help_virtual_scsi_mapping as help_scsi
import pvmcli.util.exceptions as e
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as util
import pypowervm.tasks.scsi_mapper as scsi_map
import pypowervm.wrappers.logical_partition as l_w
import pypowervm.wrappers.storage as s
import pypowervm.wrappers.virtual_io_server as vios_w
from pvmcli.command.object_type import ObjectType
from pvmcli.util.i18n import _

PRI_DISP_FIELD_LABELS = [_("LPAR"), _("LPAR Slot"), _("VIOS"), _("VIOS Slot"),
                         _("Storage")]

LPAR_HELP_TEXT = ("name=value identifier search field to determine the "
                  "LPAR for the storage connection.\n"
                  "Examples: id=1\n"
                  "          name=lpar_name")
VG_HELP_TEXT = ("name=value identifier search field to determine the "
                "volume group owner of the specified storage.\n"
                "Examples: serial_id=1\n"
                "          name=vg_name")
SSP_HELP_TEXT = ("name=value identifier search field to determine the "
                 "shared storage pool owner of the specified logical "
                 "unit.\nExamples: udid=1234abcd\n"
                 "          name=ssp_name")


class SCSIStorageMapping(ObjectType):
    def __init__(self):
        super(SCSIStorageMapping, self).__init__()

        self.add_field_models([
            s.VDisk, s.VOptMedia, s.LU, s.PV, s.VSCSIClientAdapterElement,
            s.VSCSIServerAdapterElement, s.LUTargetDev, s.PVTargetDev,
            s.VDiskTargetDev, s.VOptTargetDev])

        list_elem_wrp_funcs = [vios_w.VIOS.scsi_mappings]
        self.add_display_operation(model=self.model,
                                   parent_models=[vios_w.VIOS],
                                   elem_wrp_funcs=list_elem_wrp_funcs)
        self._add_create_operation()
        self._add_remap_operation()
        self._add_delete_operation()

        self.setup_arguments()

    def scsi_out_rows(self):
        out_rows = []

        for wrp in self.ec.wrappers:
            lpar_wrp_id = self.dfield_val(wrp, "ServerAdapter.lpar_id")
            lpar_name = util.part_name_by_id(self.ec.sm.adapter, lpar_wrp_id)
            vios_wrp_id = self.dfield_val(wrp, "ServerAdapter.vios_id")
            out_rows.append([
                lpar_name if lpar_name else _("<None>"),
                self.dfield_val(wrp, "ServerAdapter.lpar_slot_num"),
                util.part_name_by_id(self.ec.sm.adapter, vios_wrp_id)
                if vios_wrp_id else _("<None>"),
                self.dfield_val(wrp, "ServerAdapter.vios_slot_num"),
                self.dfield_val(wrp, "ServerAdapter.backing_dev_name")])
        return sorted(out_rows, key=lambda x: x[0])

    def _find_file(self, id_arg):
        id_s = id_arg.split('=', 1)
        self._check_prefixes(id_s, ('name', 'path'))
        return id_s[0] if len(id_s) == 1 else id_s[1]

    def _find_ssp(self):
        ssp_arg = self.ec.parsed_args.ssp
        if ssp_arg:
            p_stors = self.ec.sm.split_arg_search(ssp_arg, s.SSP)
        else:
            p_stors = s.SSP.get(self.ec.sm.adapter)
        if not p_stors:
            raise e.NoEntriesFoundError(object_type=s.SSP.schema_type)
        return p_stors

    def _find_vg(self):
        vio_uuid = self.ec.elem_p_wrappers[0].uuid
        vg_arg = getattr(self, 'ec.parsed_args.vg', None)
        if vg_arg:
            p_stors = self.ec.sm.split_arg_search(
                vg_arg, s.VG, parent_type=vios_w.VIOS,
                parent_uuid=vio_uuid)
        else:
            p_stors = s.VG.get(
                self.ec.sm.adapter, parent_type=vios_w.VIOS,
                parent_uuid=vio_uuid)
        if not p_stors:
            raise e.NoEntriesFoundError(object_type=s.VG.schema_type)
        return p_stors

    def _find_disk(self, id_arg):
        def _match(obj, field, value):
            return field in ('name', 'udid') and getattr(obj, field) == value

        type_arg = self.ec.parsed_args.type
        if not id_arg:
            raise e.SCSITypeDeleteRequiresStorIdError()

        if type_arg == 'lu':
            parent_stors = self._find_ssp()
        elif type_arg == 'pv':
            parent_stors = self.ec.elem_p_wrappers
            if not parent_stors:
                raise e.NoEntriesFoundError(object_type=s.PV.schema_type)
        elif type_arg == 'file':
            return self._find_file(id_arg)
        else:
            parent_stors = self._find_vg()

        if not isinstance(parent_stors, list):
            parent_stors = [parent_stors]

        split_id = id_arg.split('=', 1)

        # --stor-id takes one of the following formats:
        #    --stor-id <name>
        #    --stor-id name=<name>
        #    --stor-id udid=<udid>
        id_type = 'udid' if split_id[0] == 'udid' else 'name'
        id_val = split_id[-1]

        found_disks = []
        for parent_stor in parent_stors:
            if type_arg == 'pv':
                p_stor_func = parent_stor.phys_vols
            elif type_arg == 'lv':
                p_stor_func = parent_stor.virtual_disks
            elif type_arg == 'lu':
                p_stor_func = parent_stor.logical_units

            if type_arg == 'vopt':
                for vmr in parent_stor.vmedia_repos:
                    for vmr_vopt in vmr.optical_media:
                        if _match(vmr_vopt, id_type, id_val):
                            found_disks.append(vmr_vopt)
            else:
                for vg_s in p_stor_func:
                    if _match(vg_s, id_type, id_val):
                        found_disks.append(vg_s)

        type_name_map = {'vopt': "VirtualOpticalMedia",
                         'pv': "PhysicalVolume",
                         'lv': "LogicalVolume",
                         'lu': "LogicalUnit",
                         'file': "FileIO"}

        if not found_disks:
            raise e.NoFoundDisksError(storage_type=type_name_map[type_arg],
                                      stor_id_val=id_val,
                                      stor_id_type=id_type)

        if len(found_disks) > 1:
            raise e.MultipleFoundDisksError(
                storage_type=type_name_map[type_arg],
                stor_id_val=id_val,
                stor_id_type=id_type)

        return found_disks[0]

    def create(self):
        ''' Create operation processing function. '''
        self.validate_wrps_single_elem(
            self.ec.elem_p_wrappers, vios_w.VIOS.schema_type, op.CREATE_NAME)

        # Find LPAR
        found_lpars = self.ec.sm.split_arg_search(
            self.ec.parsed_args.lpar, l_w.LPAR)
        self.validate_wrps_single_elem(
            found_lpars, l_w.LPAR.schema_type, op.CREATE_NAME)

        type_arg = self.ec.parsed_args.type

        disk_info = self._find_disk(self.ec.parsed_args.stor_id)
        if type_arg.lower() == 'file':
            if self.ec.parsed_args.bstype:
                mapped_stor = s.FileIO.bld(
                    self.ec.sm.adapter, disk_info,
                    backstore_type=self.ec.parsed_args.bstype)
            else:
                mapped_stor = s.FileIO.bld(self.ec.sm.adapter, disk_info)
        else:
            mapped_stor = disk_info
        scsi_map.add_vscsi_mapping(
            self.ec.msys.uuid, self.ec.elem_p_wrappers[0], found_lpars[0].uuid,
            mapped_stor)

    def _add_create_operation(self):
        crt_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        crt_args.extend([op.Argument('--lpar', required=True, arg_type=str),
                         op.Argument('--vg', required=False, arg_type=str),
                         op.Argument('--ssp', required=False, arg_type=str),
                         op.Argument('--type', required=True, arg_type=str,
                                     choices=['lv', 'vopt', 'lu', 'pv',
                                              'file']),
                         op.Argument('--stor-id', required=True, arg_type=str),
                         op.Argument('--bstype', required=False,
                                     arg_type=str.lower,
                                     choices=[s.BackStoreType.FILE_IO,
                                              s.BackStoreType.USER_QCOW])])
        list_elem_wrp_funcs = [vios_w.VIOS.scsi_mappings]
        self.add_operation(op.Operation('create', self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=list_elem_wrp_funcs,
                                        op_help=help_scsi.help_create))

    def remap(self):
        ''' Remap operation processing function. '''
        type_arg = self.ec.parsed_args.type
        stor_id_arg = self.ec.parsed_args.stor_id
        new_stor_id_arg = self.ec.parsed_args.new_stor_id

        # Find LPAR
        found_lpars = self.ec.sm.split_arg_search(
            self.ec.parsed_args.lpar, l_w.LPAR)
        self.validate_wrps_single_elem(
            found_lpars, l_w.LPAR.schema_type, op.REMAP_NAME)

        mod_map = None
        if type_arg.lower() == 'vopt':
            mod_map = scsi_map.modify_vopt_mapping(
                self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                found_lpars[0].uuid,
                new_media=self._find_disk(new_stor_id_arg),
                media_name=self._find_disk(stor_id_arg).name)[1]

        if not mod_map:
            raise e.MappingNoEntriesModifiedError(type='scsi')

    def _add_remap_operation(self):
        stor_id_help = _("Limited name=value identifier search field to "
                         "determine the storage being modified. Valid "
                         "Identifiers: name, udid. "
                         "If not specified, default to name.")
        new_stor_id_help = _("Identifier to determine storage element "
                             "replacing existing one. Valid Identifiers: "
                             "name, udid. If not specified, default to name.")
        rmp_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        rmp_args.extend([op.Argument('--lpar', arg_help=LPAR_HELP_TEXT,
                                     required=True, arg_type=str),
                         op.Argument('--type',
                                     required=True, arg_type=str,
                                     choices=['vopt']),
                         op.Argument('--stor-id',
                                     arg_help=stor_id_help,
                                     required=True, arg_type=str),
                         op.Argument('--new-stor-id',
                                     arg_help=new_stor_id_help,
                                     required=True, arg_type=str)])
        list_elem_wrp_funcs = [vios_w.VIOS.scsi_mappings]
        self.add_operation(op.Operation(op.REMAP_NAME, self.remap,
                                        op_args=rmp_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=list_elem_wrp_funcs))

    def _slot_match(self, scsi_wrp, vs):
        w_vs = scsi_wrp.server_adapter.vios_slot_num
        return vs and w_vs == vs

    def _should_delete(self, scsi_wrp):
        # If we're filtering on UDID, they're unique, so check accordingly
        wrp_udid = self.ec.parsed_args.udid
        if wrp_udid and scsi_wrp.server_adapter.udid == wrp_udid:
            return True
        elif not wrp_udid:
            # If no UDID, or no UDID match, check against VIOS/LPAR ID args
            vios_slot = self.ec.parsed_args.vios_slot
            # If there is a VIOS slot, and it matches
            if self._slot_match(scsi_wrp, vios_slot):
                # If the user ID'd a LPAR as well
                if self.ec.parsed_args.lpar:
                    # Verify the LPAR is unique and return whether we have
                    # a match
                    lpar_wrps = self.ec.sm.split_arg_search(
                        self.ec.parsed_args.lpar, l_w.LPAR)
                    self.validate_wrps_single_elem(
                        lpar_wrps, l_w.LPAR.schema_type, op.DELETE_NAME)
                    return lpar_wrps[0].id == scsi_wrp.server_adapter.lpar_id
                # If there was no LPAR argument, we have a match
                return True
            # If no matches, we don't delete.
        return False

    def _no_stor_delete(self, vios, stor_id_arg):
        id_val = stor_id_arg.split("=", 1)[-1]
        for smap in vios.scsi_mappings:
            if smap.server_adapter.backing_dev_name == id_val:
                vios.scsi_mappings.remove(smap)
                vios.update()
                return smap

        # If we get this far, reraise NoFoundDisksError.
        raise

    def _no_type_delete(self):
        vios_to_upd = None
        for vios_wrp in self.ec.elem_p_wrappers:
            num_maps = len(vios_wrp.scsi_mappings)
            vios_wrp.scsi_mappings = [sw for sw in vios_wrp.scsi_mappings
                                      if not self._should_delete(sw)]
            if num_maps > len(vios_wrp.scsi_mappings):
                if vios_to_upd or num_maps - len(vios_wrp.scsi_mappings) > 1:
                    raise e.SingleMappingEntryDeleteError(
                        type='vSCSI',
                        num_mappings=(num_maps - len(vios_wrp.scsi_mappings)))
                vios_to_upd = vios_wrp
        if not vios_to_upd:
            raise e.MappingNoEntriesDeletedError(type='scsi')
        vios_to_upd.update()

    def _check_prefixes(self, split_str, expected):
        if len(split_str) == 2 and split_str[0].lower() not in expected:
            raise e.SCSIInvalidStorIDPrefixError(exp_pre=', '.join(expected))

    def delete(self):
        ''' Delete operation processing function. '''
        type_arg = self.ec.parsed_args.type
        stor_id_arg = self.ec.parsed_args.stor_id
        vios = self.ec.elem_p_wrappers[0]
        if type_arg:
            type_arg = type_arg.lower()
            # Find LPAR
            if not self.ec.parsed_args.lpar:
                raise e.MappingTypedDeleteRequiresLPAR()
            found_lpars = self.ec.sm.split_arg_search(
                self.ec.parsed_args.lpar, l_w.LPAR)
            self.validate_wrps_single_elem(
                found_lpars, l_w.LPAR.schema_type, op.DELETE_NAME)
            rm_maps = None
            if type_arg.lower() == 'vopt':
                try:
                    media = self._find_disk(stor_id_arg)
                except e.NoFoundDisksError:
                    # Backing stor elem not found, but map could still exist
                    rm_maps = self._no_stor_delete(vios, stor_id_arg)
                else:
                    rm_maps = scsi_map.remove_vopt_mapping(
                        self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                        found_lpars[0].uuid,
                        media_name=media.name)[1]
            elif type_arg.lower() == 'pv':
                rm_maps = scsi_map.remove_pv_mapping(
                    self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                    found_lpars[0].uuid,
                    self._find_disk(stor_id_arg).name)[1]
            elif type_arg.lower() == 'lv':
                rm_maps = scsi_map.remove_vdisk_mapping(
                    self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                    found_lpars[0].uuid,
                    disk_names=[self._find_disk(stor_id_arg).name])[1]
            elif type_arg.lower() == 'file':
                rm_maps = scsi_map.remove_vdisk_mapping(
                    self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                    found_lpars[0].uuid,
                    disk_names=[self._find_disk(stor_id_arg)])[1]
            elif type_arg.lower() == 'lu':
                rm_maps = scsi_map.remove_lu_mapping(
                    self.ec.sm.adapter, self.ec.elem_p_wrappers[0],
                    found_lpars[0].uuid,
                    disk_names=[self._find_disk(stor_id_arg).name])[1]
            if not rm_maps:
                raise e.MappingNoEntriesDeletedError(type='scsi')
        else:
            self._no_type_delete()

    def _add_delete_operation(self):
        del_args = o_g.default_args(self, parent_models=[vios_w.VIOS],
                                    include_output=False)
        del_args.extend([op.Argument('--lpar', required=False, arg_type=str),
                         op.Argument('--vg', required=False, arg_type=str),
                         op.Argument('--ssp', required=False, arg_type=str),
                         op.Argument('--type', required=False, arg_type=str,
                                     choices=['lv', 'vopt', 'lu', 'pv',
                                              'file']),
                         op.Argument('--stor-id', required=False,
                                     arg_type=str),
                         op.Argument('--udid', required=False, arg_type=str),
                         op.Argument('--vios-slot', required=False,
                                     arg_type=int)])
        elem_wrp_funcs = [vios_w.VIOS.scsi_mappings]
        self.add_operation(op.Operation(op.DELETE_NAME, self.delete,
                                        op_args=del_args,
                                        object_model=self.model,
                                        parent_models=[vios_w.VIOS],
                                        elem_wrp_funcs=elem_wrp_funcs,
                                        op_help=help_scsi.help_delete))

    @property
    def model(self):
        return vios_w.VSCSIMapping

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.scsi_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual SCSI Mappings"),
            'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_scsi.help_op
