# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pvmcli.command.object_type import ObjectType
import pvmcli.command.operation as op
import pvmcli.help.help_virtual_switch as help_vswitch
import pvmcli.util.help as help
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g

from pypowervm.wrappers import network as n
from pypowervm.wrappers import managed_system as ms_w


PRI_DISP_FIELD_LABELS = [_("Name"), _("ID"), _("Mode"), _("VNets")]


class VirtualSwitch(ObjectType):
    def __init__(self):
        super(VirtualSwitch, self).__init__()

        self.add_display_operation(parent_models=[ms_w.System])
        self.add_update_operation(parent_models=[ms_w.System])
        self.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME].fields[
            'VirtualSwitch.mode'].field_type = self._get_vs_mode
        self.add_delete_operation(parent_models=[ms_w.System])
        self.add_create_operation()

        self.setup_arguments()
        self.vnet_cache = {}

    def vsw_out_rows(self):
        out_rows = []
        for wrp in self.ec.wrappers:
            vnet_names = self._vnet_names(wrp)
            out_rows.append([self.dfield_val(wrp, "VirtualSwitch.name"),
                             self.dfield_val(wrp, "VirtualSwitch.switch_id"),
                             self.dfield_val(wrp, "VirtualSwitch.mode"),
                             vnet_names[0]])
            for vnet_name in vnet_names[1:]:
                out_rows.append(['', '', '', vnet_name])
        return out_rows

    def _vnet_names(self, wrp):
        vnet_names = []
        for vnet_uri in wrp.vnet_uri_list:
            if vnet_uri not in self.vnet_cache:
                self.vnet_cache[vnet_uri] = n.VNet.wrap(
                    self.ec.sm.adapter.read_by_href(vnet_uri))
            vnet_names.append(self.vnet_cache[vnet_uri].name)

        return vnet_names if vnet_names else ['']

    @property
    def model(self):
        return n.VSwitch

    def add_create_operation(self):
        name_help = _("Required. The name of the switch.")
        mode_help = _("The mode of the switch. Default: Veb")
        crt_args = o_g.default_args(self, parent_models=[ms_w.System])
        crt_args.extend([op.Argument('--name', arg_type=str,
                                     required=True,
                                     arg_help=help.wrap_help(name_help)),
                         op.Argument('--mode', arg_type=str.lower,
                                     required=False,
                                     choices=['veb', 'vepa'],
                                     arg_help=help.wrap_help(mode_help))])
        self.add_operation(op.Operation(op.CREATE_NAME, self.create,
                                        op_args=crt_args,
                                        object_model=self.model,
                                        parent_models=[ms_w.System]))

    def create(self):
        vs_name = self.ec.parsed_args.name
        vs_mode = self._get_vs_mode(self.ec.parsed_args.mode)
        new_vs = n.VSwitch.bld(adapter=self.ec.sm.adapter,
                               name=vs_name, switch_mode=vs_mode)

        c_vs = new_vs.create(parent_type=self.ec.exec_op.parent_schemas[0],
                             parent_uuid=self.ec.msys.uuid)

        self.ec.wrappers = [c_vs]
        # Only display if a user specifies specific display fields
        if (self.ec.parsed_args.display_fields
                or self.ec.parsed_args.all_out):
            self.display()

    def _get_vs_mode(self, vs_mode_arg):
        if vs_mode_arg:
            vs_mode_arg = vs_mode_arg.lower()
        return (n.VSwitchMode.VEPA if vs_mode_arg == 'vepa'
                else n.VSwitchMode.VEB)

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.vsw_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("Virtual Switches"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_vswitch.help_text
