# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import re
import six

import pvmcli.command.object_type as ot
import pvmcli.command.operation as op
import pvmcli.help.help_vnic as help_vnic
import pvmcli.util.exceptions as exc
from pvmcli.util.i18n import _
import pvmcli.util.op_generator as o_g
import pvmcli.util.util as pvmcli_util

import pypowervm.tasks.sriov as sriov_task
from pypowervm.wrappers import iocard as iocard
from pypowervm.wrappers import logical_partition as lpar_w
import pypowervm.util as util


BACK_DEV_ARG_NAME = '--back-dev'
ADD_BACK_DEVS_NAME = "add-back-devs"
REM_BACK_DEVS_NAME = "remove-back-devs"
BACK_DEV_HELP = _("Required: A backing device to assign to this VNIC. "
                  "This argument may be specified multiple times. These "
                  "--back-dev strings take the following form:\n"
                  "vios=<VIOS name> loc=<SR-IOV Phys Port Loc Code>"
                  "[ cap=<capacity percentage value> ]"
                  "[ pri=<failover priority integer value> ]"
                  "[ maxcap=<max capacity percentage value> ]\n")
BACK_DEV_SCHEMA = 'VirtualNICSRIOVBackingDevice'

PRI_DISP_FIELD_LABELS = [_("LPAR"), _("MAC"), _("Slot"), _("PVID"),
                         _("BDev PPort Loc"), _("VIOS")]


class VNIC(ot.ObjectType):
    def __init__(self):
        super(VNIC, self).__init__()

        self.add_field_models([iocard.VNICBackDev])

        d_op = self.add_display_operation(
            parent_models=[lpar_w.LPAR])

        vnic_cap_fld = d_op.op_args[op.DF_NAME].fields[
            'VirtualNICDedicated.capacity']
        vnic_cap_fld.out_func = util.sanitize_percent_for_api
        back_cap = d_op.op_args[op.DF_NAME].fields[
            'VirtualNICSRIOVBackingDevice.capacity']
        back_cap.out_func = util.sanitize_percent_for_api
        des_max_cap = d_op.op_args[op.DF_NAME].fields[
            'VirtualNICSRIOVBackingDevice.desired_max_capacity']
        des_max_cap.out_func = util.sanitize_percent_for_api
        max_cap = d_op.op_args[op.DF_NAME].fields[
            'VirtualNICSRIOVBackingDevice.max_capacity']
        max_cap.out_func = util.sanitize_percent_for_api

        self.add_delete_operation(parent_models=[lpar_w.LPAR])
        self._add_create_operation()

        # Set back_devs with add-back-devs/remove-back-devs operations
        self.exc_set_fields.append('VirtualNICDedicated.back_devs')
        upd_op = self.add_update_operation(parent_models=[lpar_w.LPAR])
        self._add_update_back_dev_args(upd_op)
        self.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME].fields[
            'VirtualNICDedicated.allowed_macs'].field_type = list
        self.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME].fields[
            'VirtualNICDedicated.allowed_vlans'].field_type = list
        self._add_add_back_devs()
        self._add_remove_back_devs()

        self.setup_arguments()

    def _set_back_dev_vals(self, bd_sfs):
        if not self.ec.parsed_args.loc:
            raise exc.VNICBackDevLocRequiredError()
        pport = sriov_task.find_pport(self.ec.msys, self.ec.parsed_args.loc)
        if not pport:
            raise exc.VNICBackDevUpdateLocError(loc=self.ec.parsed_args.loc)

        for bd in self.ec.wrappers[0].back_devs:
            if (pport.port_id == bd.pport_id and
                    pport.sriov_adap_id == bd.sriov_adap_id):
                for bd_sf in bd_sfs:
                    self.set_field_val(bd, bd_sf, bd_sfs[bd_sf])
                break

    def update(self):
        self.validate_wrps_single_elem(
            self.ec.wrappers, self.name, op.UPDATE_NAME)
        sfs = pvmcli_util.split_set_fields(self.ec.parsed_args.set_fields)

        if len(sfs) == 0:
            raise exc.InvalidSetFieldsError(object_type=self.name,
                                            operation=op.UPDATE_NAME)

        bd_sfs = {}
        for sf in sfs:
            if sf.lower().startswith(BACK_DEV_SCHEMA.lower()):
                bd_sfs[sf] = sfs[sf]
        if bd_sfs:
            for bd_sf in bd_sfs:
                del sfs[bd_sf]
            self._set_back_dev_vals(bd_sfs)

        ot.update_retry(self.ec.wrappers[0], sfs, self)

    def _add_update_back_dev_args(self, upd_op):
        loc_help = ("A physical port location code for a backing device to "
                    "update for this VNIC. Required if updating a backing "
                    "device.")
        upd_op.add_arg(op.Argument('--loc', arg_type=str, arg_help=loc_help))

    def _add_remove_back_devs(self):
        rem_args = o_g.default_args(
            self, parent_models=[lpar_w.LPAR], object_id_required=True,
            include_output=False)
        pp_loc_help = ("Required: A physical port location code for a backing "
                       "device to remove from this VNIC. This argument may be "
                       "specified multiple times.")
        rem_args.append(op.Argument(
            '--loc', arg_type=str, arg_help=pp_loc_help, action='append',
            required=True))
        self.add_operation(op.Operation(
            REM_BACK_DEVS_NAME, self._rem_back_devs, op_args=rem_args,
            object_model=self.model, parent_models=[lpar_w.LPAR]))

    def _rem_back_devs(self):
        self.validate_wrps_single_elem(
            self.ec.wrappers, iocard.VNIC.schema_type, REM_BACK_DEVS_NAME)
        fnd_pports = []
        for loc_str in self.ec.parsed_args.loc:
            fnd_pport = sriov_task.find_pport(self.ec.msys, loc_str)
            if not fnd_pport:
                raise exc.NoPPortFoundFromLocError(loc=loc_str)
            fnd_pports.append(fnd_pport)
        pport_ids = [(a_pp.sriov_adap_id, a_pp.port_id) for a_pp in fnd_pports]
        new_b_devs = []
        for b_d in self.ec.wrappers[0].back_devs:
            if (b_d.sriov_adap_id, b_d.pport_id) not in pport_ids:
                new_b_devs.append(b_d)
        if len(self.ec.wrappers[0].back_devs) == len(new_b_devs):
            raise exc.NoEntriesFoundError(
                object_type=iocard.VNICBackDev.schema_type)
        self.ec.wrappers[0].back_devs = new_b_devs
        self.ec.wrappers[0].update()

    def _add_add_back_devs(self):
        add_args = o_g.default_args(
            self, parent_models=[lpar_w.LPAR], object_id_required=True,
            include_output=False)
        back_dev_add_help = BACK_DEV_HELP + _(
            "Example:\n"
            "pvmctl vnic add-back-devs -i mac=11:11:11:11:11:11 "
            "--back-dev vios=vios1 loc=U78CB.001.WZS0485-P1-C5-T3 --back-dev "
            "vios=vios2 loc=U78CB.001.WZS0485-P1-C5-T3 cap=42.42%% pri=10 "
            "maxcap=43%%.")
        add_args.append(op.Argument(BACK_DEV_ARG_NAME, arg_type=str,
                                    arg_help=back_dev_add_help, nargs='+',
                                    action='append', required=True))
        self.add_operation(op.Operation(
            ADD_BACK_DEVS_NAME, self._add_back_devs, op_args=add_args,
            object_model=self.model, parent_models=[lpar_w.LPAR]))

    def _add_back_devs(self):
        self.validate_wrps_single_elem(
            self.ec.wrappers, iocard.VNIC.schema_type, ADD_BACK_DEVS_NAME)

        back_devs = self.ec.wrappers[0].back_devs
        for back_dev in self.ec.parsed_args.back_dev:
            back_devs.append(pvmcli_util.build_back_dev(
                self.ec.sm.adapter, back_dev, self.ec.msys))
        self.ec.wrappers[0].back_devs = back_devs
        self.ec.wrappers[0].update()

    def create(self):
        self.validate_wrps_single_elem(
            self.ec.elem_p_wrappers, lpar_w.LPAR.schema_type, op.CREATE_NAME)

        pvid = self.ec.parsed_args.pvid
        slot = self.ec.parsed_args.slot
        mac = self.ec.parsed_args.mac

        a_vlans_arg = self.ec.parsed_args.allowed_vlans
        if not a_vlans_arg:
            p_vlans = util.VLANList.ALL
        else:
            a_vlans_split = a_vlans_arg
            if len(a_vlans_arg) == 1:
                a_vlans_split = re.split('[, ]+', a_vlans_arg[0])
            p_vlans = util.VLANList.const_or_list(a_vlans_split)

        a_macs_arg = self.ec.parsed_args.allowed_macs
        if not a_macs_arg:
            p_macs = util.MACList.ALL
        else:
            a_macs_split = a_macs_arg
            if len(a_macs_arg) == 1:
                a_macs_split = re.split('[, ]+', a_macs_arg[0])
            p_macs = util.MACList.const_or_list(a_macs_split)

        back_devs = []
        for back_dev in self.ec.parsed_args.back_dev:
            back_devs.append(pvmcli_util.build_back_dev(
                self.ec.sm.adapter, back_dev, self.ec.msys))

        iocard.VNIC.bld(
            self.ec.sm.adapter, pvid=pvid, slot_num=slot, mac_addr=mac,
            allowed_vlans=p_vlans, allowed_macs=p_macs,
            back_devs=back_devs).create(parent=self.ec.elem_p_wrappers[0])

    def _add_create_operation(self):
        pvid_help = _("Port VLAN ID for this vNIC.  If not specified, the "
                      "vNIC's traffic is untagged.")
        slot_help = _("Desired virtual slot number on the owning LPAR.")
        mac_help = _("MAC address for the vNIC.")
        av_help = _("Space separated integer list of allowed VLANs on this "
                    "VNIC. If not specified, all VLANs are allowed. "
                    "Can also specify `all` or `none` for all allowed or "
                    "none allowed.\n"
                    "Examples:\n    --allowed-vlans 1 2 3\n"
                    "    --allowed-vlans none"
                    "    --allowed-vlans all")
        am_help = _("Space separated string list of allowed MAC addresses on "
                    "this VNIC. If not specified, all MAC addresses are "
                    "allowed. Can also specify `all` or `none` for all "
                    "allowed or none allowed.\n"
                    "Example:\n"
                    "    --allowed-macs FF:FF:FF:FF:FF:FF 11:11:11:11:11:11\n"
                    "    --allowed-macs none"
                    "    --allowed-macs all")
        back_dev_crt_help = BACK_DEV_HELP + _(
            "Example:\n"
            "pvmctl vnic create -p name=lpar1 --back-dev vios=vios1 "
            "loc=U78CB.001.WZS0485-P1-C5-T3 --back-dev vios=vios2 "
            "loc=U78CB.001.WZS0485-P1-C5-T3 cap=42.42%% pri=10 maxcap=43%%.")

        crt_args = o_g.default_args(
            self, parent_models=[lpar_w.LPAR], parent_id_required=True)
        crt_args.extend([op.Argument('--pvid', arg_help=pvid_help,
                                     required=False, arg_type=int),
                         op.Argument('--slot', arg_help=slot_help,
                                     required=False, arg_type=int),
                         op.Argument('--mac', arg_help=mac_help,
                                     required=False, arg_type=str),
                         op.Argument('--allowed-vlans', arg_type=str,
                                     arg_help=av_help, nargs='+',
                                     required=False),
                         op.Argument('--allowed-macs', arg_type=str,
                                     arg_help=am_help, nargs='+',
                                     required=False),
                         op.Argument(BACK_DEV_ARG_NAME, arg_type=str,
                                     arg_help=back_dev_crt_help, nargs='+',
                                     action='append', required=True)])
        self.add_operation(op.Operation(
            op.CREATE_NAME, self.create, op_args=crt_args,
            object_model=self.model, parent_models=[lpar_w.LPAR]))

    @property
    def model(self):
        return iocard.VNIC

    def _vios_name_by_href(self, vios_href):
        return six.text_type(pvmcli_util.vios_by_uri(
            self.ec.sm.adapter, vios_href, xag=[]).name)

    def vnic_out_rows(self):
        out_rows = []
        pp_to_loc = {}
        for sr_adp in self.ec.msys.asio_config.sriov_adapters:
            for pport in sr_adp.phys_ports:
                pp_to_loc[(
                    pport.port_id, pport.sriov_adap_id)] = six.text_type(
                        pport.loc_code)

        for lpar_wrp in sorted(self.ec.parent_wrp_to_children,
                               key=lambda w: w.name):
            out_row = [six.text_type(lpar_wrp.name)]
            for s_vnic in self.ec.parent_wrp_to_children[lpar_wrp]:
                pvid_str = six.text_type(s_vnic.pvid)
                out_row.extend([six.text_type(s_vnic.mac),
                                six.text_type(s_vnic.slot), pvid_str])
                for bdev in s_vnic.back_devs:
                    out_row.extend([pp_to_loc[(bdev.pport_id,
                                               bdev.sriov_adap_id)],
                                    self._vios_name_by_href(bdev.vios_href)])
                    out_rows.append(out_row)
                    out_row = ['', '', '', '']
                out_row = ['']

        return out_rows

    def pri_output(self):
        po_table = self.resize_pri_out_table(PRI_DISP_FIELD_LABELS,
                                             self.vnic_out_rows())

        return "%(n_ttl)s\n%(n_tbl)s" % {
            'n_ttl': _("VNICs"), 'n_tbl': po_table.get_string()}

    @property
    def help(self):
        return help_vnic.help_text
