# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.cli_parser as cli_parser
import pvmcli.command.io_slot as io_c
import pvmcli.help.help_ioslot as help_ioslot
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc
import pvmcli.util.output_processor as out_proc

import pypowervm.tests.test_utils.pvmhttp as pvmhttp
import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.logical_partition as lpar_w
import pypowervm.wrappers.virtual_io_server as vios_w
import pypowervm.wrappers.managed_system as msys_w


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] IOSlot list {--object-id "
                    "| -i} <ProfileIOSlot field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    IOSlot.bus_grp_required\n"
     "    IOSlot.description\n"
     "    IOSlot.drc_index\n"
     "    IOSlot.drc_name\n"
     "    IOSlot.feat_codes\n"
     "    IOSlot.part_id\n"
     "    IOSlot.part_name\n"
     "    IOSlot.part_type\n"
     "    IOSlot.part_uuid\n"
     "    IOSlot.pci_class\n"
     "    IOSlot.pci_dev_id\n"
     "    IOSlot.pci_rev_id\n"
     "    IOSlot.pci_subsys_dev_id\n"
     "    IOSlot.pci_subsys_vendor_id\n"
     "    IOSlot.pci_vendor_id\n"])


class IOSlotTest(twa.TestWrapper):
    @property
    def file(self):
        return 'managedsystem.txt'

    @property
    def wrapper_class_to_test(self):
        return msys_w.System

    @mock.patch("pvmcli.util.session_manager")
    @mock.patch("pvmcli.command.object_type.ExecutionContext.query_wrappers")
    @mock.patch(
        "pvmcli.command.object_type.ExecutionContext.find_managed_system")
    def test_io_slot(self, find_msys_mock, query_wrp_mock, sess_man_mock):
        io_slot = io_c.IOSlot()

        test_util.test_obj_type(self, io_slot, 'IOSlot', bp_w.IOSlot,
                                field_models=[msys_w.IOSlot],
                                name_to_ops={'attach', 'list', 'detach'},
                                exp_help=help_ioslot.help_text)

        # Test display field output function overrides
        drc_df = io_slot.get_disp_fields()['IOSlot.drc_index']
        self.assertTrue(drc_df.out_func == out_proc.four_char_hex)
        drc_df = io_slot.get_disp_fields()['IOSlot.pci_class']
        self.assertTrue(drc_df.out_func == out_proc.four_char_hex)
        drc_df = io_slot.get_disp_fields()['IOSlot.pci_subsys_dev_id']
        self.assertTrue(drc_df.out_func == out_proc.four_char_hex)
        drc_df = io_slot.get_disp_fields()['IOSlot.pci_subsys_vendor_id']
        self.assertTrue(drc_df.out_func == out_proc.four_char_hex)
        drc_df = io_slot.get_disp_fields()['IOSlot.pci_vendor_id']
        self.assertTrue(drc_df.out_func == out_proc.four_char_hex)

        # Test Operations
        lw_io_config = lpar_w.LPAR.io_config
        bw_ioslots = bp_w.PartitionIOConfiguration.io_slots
        test_util.test_ops(self,
                           parent_model_ops={io_slot.name_to_ops['list']:
                                             [msys_w.System],
                                             io_slot.name_to_ops['attach']:
                                             [lpar_w.LPAR, vios_w.VIOS],
                                             io_slot.name_to_ops['detach']:
                                             [lpar_w.LPAR, vios_w.VIOS]},
                           child_model_ops={},
                           elem_wrp_func_ops={io_slot.name_to_ops['list']:
                                              [msys_w.System.asio_config,
                                               msys_w.ASIOConfig.io_slots],
                                              io_slot.name_to_ops['attach']:
                                              [lw_io_config, bw_ioslots],
                                              io_slot.name_to_ops['detach']:
                                              [lw_io_config, bw_ioslots]})

        cp = cli_parser.CLIParser(['ioslot', 'list'])
        cp.process(auto_exec=False)

        if self.dwrap:
            cp.obj_type.ec.wrappers = [self.dwrap.asio_config.io_slots[0]]
        drc_df = 'IOSlot.drc_index'
        cp.obj_type.ec.parsed_args.display_fields = [drc_df]
        self.assertEqual(cp.obj_type.ec.wrappers[0].drc_index, 553713681)
        cp.obj_type.cli_out = six.StringIO()
        out_proc.specified_display(cp.obj_type)
        # Verify hex output: hex(553713681) == 21010011 and label override
        self.assertEqual(cp.obj_type.cli_out.getvalue(),
                         'drc_index=21010011\n')

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'IOSlot')

    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.wrap")
    @mock.patch("pypowervm.wrappers.logical_partition.LPAR.wrap")
    def test_list(self, lpar_wrap_mock, vios_wrap_mock):
        io_cfg = bp_w.PartitionIOConfiguration.bld(mock.Mock(), 5)
        bp_ioslot = bp_w.IOSlot.bld(mock.Mock(), False, 1)
        aioslot = bp_w.IOSlot.AssociatedIOSlot.wrap(
            bp_ioslot._find(bp_w.ASSOC_IO_SLOT_ROOT))
        aioslot.set_parm_value(
            bp_w._ASSOC_IO_SLOT_DRC_NAME, "U5294.001.CEC1234-P01-C011")
        bp_ioslot.set_parm_value(bp_w.ASSOC_IO_SLOT_ROOT, aioslot)
        io_cfg.io_slots = [bp_ioslot]

        ret_lpar = lpar_w.LPAR.bld(
            mock.Mock(), "test_lpar",
            bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp_w.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1),
            io_cfg=io_cfg)

        vio_cfg = bp_w.PartitionIOConfiguration.bld(mock.Mock(), 5)
        v_bp_ioslot = bp_w.IOSlot.bld(mock.Mock(), False, 1)
        v_aioslot = bp_w.IOSlot.AssociatedIOSlot.wrap(
            v_bp_ioslot._find(bp_w.ASSOC_IO_SLOT_ROOT))
        v_aioslot.set_parm_value(
            bp_w._ASSOC_IO_SLOT_DRC_NAME, "U5294.001.CEC1234-P01-C013")
        v_bp_ioslot.set_parm_value(bp_w.ASSOC_IO_SLOT_ROOT, v_aioslot)
        vio_cfg.io_slots = [v_bp_ioslot]
        ret_vios = vios_w.VIOS.bld(
            mock.Mock(), "test_vios",
            bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp_w.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1),
            io_cfg=vio_cfg)
        lpar_wrap_mock.return_value = [ret_lpar]
        vios_wrap_mock.return_value = [ret_vios]

        test_util.test_list(
            self, 'IOSlot',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["I/O Slots\n"
             "+----------------------------+-----------+"
             "---------------------------------+\n"
             "|    Loc Code (DRC Name)     |    LPAR   |"
             " Description                     |\n"
             "+----------------------------+-----------+"
             "---------------------------------+\n"
             "| U5294.001.CEC1234-P01-C011 | test_lpar |"
             " I/O Processor                   |\n"
             "| U5294.001.CEC1234-P01-C012 |           |"
             " I/O Processor                   |\n"
             "| U5294.001.CEC1234-P01-C013 | test_vios |"
             " PCI Ultra2 RAID Disk Controller |\n"
             "+----------------------------+-----------+"
             "---------------------------------+\n"],
            ["----------\n"
             "| IOSlot |\n"
             "----------\n"
             "  bus_grp_required=False\n"
             "  description=I/O Processor\n"
             "  drc_index=21010011\n"
             "  drc_name=U5294.001.CEC1234-P01-C011\n"
             "  feat_codes=0\n"
             "  pci_class=0200\n"
             "  pci_dev_id=1234\n"
             "  pci_rev_id=00\n"
             "  pci_subsys_dev_id=04B2\n"
             "  pci_subsys_vendor_id=1014\n"
             "  pci_vendor_id=1014\n"
             "----------\n"
             "| IOSlot |\n"
             "----------\n"
             "  bus_grp_required=True\n"
             "  description=I/O Processor\n"
             "  drc_index=21020011\n"
             "  drc_name=U5294.001.CEC1234-P01-C012\n"
             "  part_id=2\n"
             "  part_name=vios1\n"
             "  part_type=Virtual IO Server\n"
             "  part_uuid=02899D96-9D20-490F-8B1B-4D3DEE1210ED\n"
             "  pci_class=0200\n"
             "  pci_dev_id=1234\n"
             "  pci_rev_id=00\n"
             "  pci_vendor_id=1014\n"
             "----------\n"
             "| IOSlot |\n"
             "----------\n"
             "  bus_grp_required=False\n"
             "  description=PCI Ultra2 RAID Disk Controller\n"
             "  drc_index=21030011\n"
             "  drc_name=U5294.001.CEC1234-P01-C013\n"
             "  pci_class=0100\n"
             "  pci_dev_id=0049\n"
             "  pci_rev_id=03\n"
             "  pci_vendor_id=1014\n"],
            ['<IOSlot xmlns="http://www.ibm.com/xmlns/'],
            ['IOSlot.pci_class', 'IOSlot.description'],
            ["pci_class=0200,description=I/O Processor\n"],
            ["0200,I/O Processor\n"],
            ["pci_class___0200,description___I/O Processor\n"],
            ["pci_class=0200___description=I/O Processor\n"],
            ["pci_class=0200,description=I/O Processor___"
             "pci_class=0200,description=I/O Processor___"
             "pci_class=0100,description=PCI Ultra2 RAID Disk Controller\n"],
            ["pci_class___0200___description___I/O Processor___"
             "pci_class___0200___description___I/O Processor___"
             "pci_class___0100___description___"
             "PCI Ultra2 RAID Disk Controller\n"],
            'ManagedSystem',
            wrps=self.dwrap.asio_config.io_slots[0:3])

    def test_attach(self):
        ''' Test attach operation on IOSlot ObjectType. '''
        # cp = self._setup_and_exec(args=['IOSlot', 'attach', '--drc-names',
        #                                 'ioname', 'ioname2', '-p',
        #                                 'name=testlpar'])
        # func() on parsed_args is the callback used for each operation
        # Verifying we attempt to call into this op func through that func
        # cp.obj_type.ec.parsed_args.func.assert_called_once_with()
        pass

    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.search')
    @mock.patch('pypowervm.wrappers.logical_partition.LPAR.update')
    def test_detach(self, lpar_update_mock, lpar_search_mock):
        ''' Test detach operation on IOSlot ObjectType. '''
        ret_lpar = self._setup_mock_lpar()[0]
        bp_ioslot = bp_w.IOSlot.bld(mock.Mock(), False, 553713681)
        aioslot = bp_w.IOSlot.AssociatedIOSlot.wrap(
            bp_ioslot._find(bp_w.ASSOC_IO_SLOT_ROOT))
        aioslot.set_parm_value(
            bp_w._ASSOC_IO_SLOT_DRC_NAME, "U5294.001.CEC1234-P01-C011")
        bp_ioslot.set_parm_value(bp_w.ASSOC_IO_SLOT_ROOT, aioslot)
        ret_lpar.io_config.io_slots = [bp_ioslot]

        # Set parent wrapper search return results to empty list
        # Will pass down ret_lpar as part of p_wrps arg to setup_mock
        lpar_search_mock.return_value = []

        cp = test_util.setup_mock(self, ['IOSlot', 'detach', '-p',
                                         'name=testlpar', '--drc-names',
                                         'U5294.001.CEC1234-P01-C011'],
                                  self.adpt, wrps=[ret_lpar],
                                  p_wrps=[ret_lpar])
        cp.obj_type.execute()
        # Make sure parent wrappers are searched for
        lpar_search_mock.assert_called_once_with(
            cp.obj_type.ec.sm.adapter, name='testlpar', xag=[],
            parent_uuid=None, parent_type=None)
        ret_lpar.update.assert_called_once_with()
        ret_lpar.update.reset_mock()

        cp = test_util.setup_mock(self, ['IOSlot', 'detach', '-p',
                                         'name=testlpar', '--drc-names',
                                         'IDONTEXIST'],
                                  self.adpt, wrps=[ret_lpar],
                                  p_wrps=[ret_lpar])

        self.assertRaises(exc.NoEntriesFoundError, cp.obj_type.execute)

        ret_lpar.update.assert_not_called()
        ret_lpar.update.reset_mock()

        ret_lpar.io_config.io_slots = []
        cp = test_util.setup_mock(self, ['IOSlot', 'detach', '-p',
                                         'name=testlpar', '--drc-names',
                                         'U5294.001.CEC1234-P01-C011'],
                                  self.adpt, wrps=[ret_lpar],
                                  p_wrps=[ret_lpar])

        self.assertRaises(exc.NoEntriesFoundError, cp.obj_type.execute)

        ret_lpar.update.assert_not_called()

    def _setup_mock_lpar(self):
        resp = pvmhttp.PVMResp(
            pvmfile=pvmhttp.PVMFile('lpar.txt'),
            adapter=self.adpt).get_response()
        return lpar_w.LPAR.wrap(resp)
