# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock

import pvmcli.command.logical_unit as lu_c
import pvmcli.help.help_logical_unit as help_lu
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc
import pvmcli.util.util as util

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.storage as stor
import pypowervm.wrappers.virtual_io_server as vios_w


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] LogicalUnit list "
                    "{--parent-id | -p} <SharedStoragePool field>=value "
                    "{--object-id | -i} <LogicalUnit field>=value "
                    "{args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    LogicalUnit.capacity\n"
     "    LogicalUnit.cloned_from_udid\n"
     "    LogicalUnit.emulate_model\n"
     "    LogicalUnit.in_use\n"
     "    LogicalUnit.is_thin\n"
     "    LogicalUnit.lu_type\n"
     "    LogicalUnit.name\n"
     "    LogicalUnit.tag\n"
     "    LogicalUnit.udid\n"])


class LogicalUnitTest(twa.TestWrapper):
    ''' Test framework for LogicalUnit Object Type. '''
    @property
    def file(self):
        return 'ssp.txt'

    @property
    def wrapper_class_to_test(self):
        return stor.SSP

    def test_logical_unit(self):
        lu = lu_c.LogicalUnit()

        test_util.test_obj_type(self, lu, 'LogicalUnit', stor.LU,
                                field_models=[stor.LU],
                                name_to_ops={'list', 'upload', 'delete'},
                                exp_help=help_lu.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={lu.name_to_ops['list']: [stor.SSP],
                                    lu.name_to_ops['upload']: None,
                                    lu.name_to_ops['delete']: [stor.SSP]},
            child_model_ops={lu.name_to_ops['list']: None,
                             lu.name_to_ops['upload']: None,
                             lu.name_to_ops['delete']: None},
            elem_wrp_func_ops={
                lu.name_to_ops['list']: [stor.SSP.logical_units],
                lu.name_to_ops['upload']: None,
                lu.name_to_ops['delete']: [stor.SSP.logical_units]})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'LogicalUnit')

    def test_list(self):
        ssp_wrap = stor.SSP.wrap(self.resp)
        test_util.test_list(
            self, 'LogicalUnit',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Logical Units\n"
             "+---------+--------+----------+------+------+--------+--------+"
             "\n"
             "|   SSP   |  Name  | Cap (GB) | Type | Thin | Clone  | In use |"
             "\n"
             "+---------+--------+----------+------+------+--------+--------+"
             "\n"
             "| neossp1 | neolu1 |   1.0    | Disk | True | neolu1 |  True  |"
             "\n"
             "+---------+--------+----------+------+------+--------+--------+"
             "\n"],
            ["---------------\n| LogicalUnit |\n---------------\n"
             "  capacity=1.0\n"
             "  emulate_model=True\n"
             "  in_use=True\n"
             "  is_thin=True\n"
             "  lu_type=VirtualIO_Disk\n"
             "  name=neolu1\n"
             "  udid=27cfc907d2abf511e4b2d540f2e95daf301a02b"
             "0904778d755df5a46fe25e500d8\n"],
            ['<LogicalUnit xmlns="http://www.ibm.com/xmlns/'],
            ['LogicalUnit.name', 'LogicalUnit.capacity'],
            ["name=neolu1,capacity=1.0\n"],
            ["neolu1,1.0\n"],
            ["name___neolu1,capacity___1.0\n"],
            ["name=neolu1___capacity=1.0\n"],
            ["name=neolu1,capacity=1.0\n"],
            ["name___neolu1___capacity___1.0\n"],
            'SharedStoragePool',
            wrps=[ssp_wrap.logical_units[0]], p_wrps=[ssp_wrap])

    def test_pri_out(self):
        lu_1 = stor.SSP.wrap(self.resp).logical_units[0]
        lu_2 = stor.LU.bld(mock.Mock(), 'neo_ssp2', 2.0,
                           False, typ=stor.LUType.HIBERNATION)
        lu_2._udid('12345678')
        lu_3 = stor.LU.bld(mock.Mock(), 'neo_ssp345', 4.85,
                           True, typ=stor.LUType.AMS)
        lu_3._udid('90123456')

        ssp_parent = stor.SSP.wrap(self.resp)
        lus = ssp_parent.logical_units
        lus.extend([lu_2, lu_3])
        ssp_parent.logical_units = lus
        cp = test_util.setup_mock(
            self, args=['LogicalUnit', 'list', '--max-width', '120'],
            adpt_fx=self.adpt, wrps=[lu_1, lu_2, lu_3], p_wrps=[ssp_parent])
        cp.obj_type.execute()

        self.assertIn(
            "Logical Units\n"
            "+---------+------------+----------+-----------------------+"
            "-------+------------+--------+\n"
            "|   SSP   |    Name    | Cap (GB) |          Type         |"
            "  Thin |   Clone    | In use |\n"
            "+---------+------------+----------+-----------------------+"
            "-------+------------+--------+\n"
            "| neossp1 |  neo_ssp2  |   2.0    |      Hibernation      |"
            " False |  neo_ssp2  |        |\n"
            "|         | neo_ssp345 |   4.85   | Active_Memory_Sharing |"
            "  True | neo_ssp345 |        |\n"
            "|         |   neolu1   |   1.0    |          Disk         |"
            "  True |   neolu1   |  True  |\n"
            "+---------+------------+----------+-----------------------+"
            "-------+------------+--------+\n",
            cp.obj_type.cli_out.getvalue())

    @mock.patch('pvmcli.command.object_type.ExecutionContext.query_wrappers')
    @mock.patch('os.path.getsize')
    @mock.patch('pypowervm.wrappers.storage.SSP.get')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.get')
    @mock.patch('pypowervm.wrappers.storage.SSP.search')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.tasks.storage.upload_new_lu')
    def test_upload(self, lu_mock, vios_search_mock, ssp_search_mock,
                    vios_get_mock, ssp_get_mock, mock_getsize, q_wrps_mock):
        ret_vios = vios_w.VIOS.bld(
            mock.Mock(), "test_vios",
            bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp_w.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        ret_vios.set_parm_value(bp_w._BP_STATE, bp_w.LPARState.RUNNING)
        ret_vios.set_parm_value(bp_w._BP_RMC_STATE, bp_w.RMCState.ACTIVE)
        vios_search_mock.return_value = [ret_vios]
        vios_get_mock.return_value = [ret_vios]

        ssp_wrp = stor.SSP.wrap(self.resp)
        ssp_search_mock.return_value = [ssp_wrp]
        ssp_get_mock.return_value = [ssp_wrp]

        # Minimum args
        cp = test_util.setup_mock(self, args=['lu', 'upload', '--name',
                                              'lu_name', '--data', 'lu_data'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])

        with mock.patch(test_util.mock_open_str()) as mock_open:
            cp.obj_type.execute()
            lu_mock.assert_called_once_with(
                None, util.name_cache[stor.SSP.schema_type]['neossp1'],
                mock_open().__enter__(), 'lu_name', mock_getsize(),
                sha_chksum=None, d_size=None)
            ssp_search_mock.assert_not_called()
            q_wrps_mock.asssert_not_called()
        lu_mock.reset_mock()
        ssp_search_mock.reset_mock()
        # Reset SSP cache to imitate fresh execution
        util.name_cache = {}

        # All args
        cp = test_util.setup_mock(self, args=['lu', 'upload', '--vios',
                                              'test_vios', '--ssp', 'ssp1',
                                              '--name', 'lu_name',
                                              '--data', 'lu_data',
                                              '--checksum', 'abc',
                                              '--size', '1'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])

        with mock.patch(test_util.mock_open_str()) as mock_open:
            cp.obj_type.execute()
            lu_mock.assert_called_once_with(
                None, ssp_wrp, mock_open().__enter__(), 'lu_name',
                mock_getsize(), sha_chksum='abc', d_size=1)
            ssp_search_mock.assert_called_once_with(
                cp.obj_type.ec.sm.adapter, name='ssp1', parent_uuid=None,
                parent_type=None)
        lu_mock.reset_mock()
        ssp_search_mock.reset_mock()

        # No args
        self.assertRaises(SystemExit, test_util.setup_mock, self,
                          args=['lu', 'upload'], adpt_fx=self.adpt,
                          wrps=[self.dwrap])
        lu_mock.assert_not_called()

        # No data specified and no data size specified
        cp = test_util.setup_mock(
            self, args=['lu', 'upload', '--vios', 'test_vios', '--name',
                        'lu_name'], adpt_fx=self.adpt, wrps=[self.dwrap])
        self.assertRaises(
            exc.UploadStdInRequiresSizeError, cp.obj_type.execute)
        lu_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.storage.SSP.update')
    @mock.patch('pypowervm.wrappers.storage.SSP.search')
    def test_delete(self, ssp_search_mock, ssp_upd_mock):
        ssp_wrp = stor.SSP.wrap(self.resp)
        ssp_search_mock.return_value = [ssp_wrp]

        self.assertRaises(SystemExit, test_util.setup_mock, self,
                          args=['lu', 'delete', '-p', 'name=neossp1'],
                          adpt_fx=self.adpt,
                          wrps=[self.dwrap.logical_units[0]])
        ssp_upd_mock.assert_not_called()
        ssp_upd_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['lu', 'delete', '-i',
                                              'name=neolu1', '-p',
                                              'name=neossp1'],
                                  adpt_fx=self.adpt,
                                  wrps=[self.dwrap.logical_units[0]],
                                  p_wrps=[ssp_wrp])
        cp.obj_type.execute()
        ssp_upd_mock.assert_called_once_with()
        self.assertEqual(len(self.dwrap.logical_units), 0)
