# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import copy
import mock
import six

import pvmcli.command.logical_volume as lv_c
import pvmcli.command.operation as op
import pvmcli.command.volume_group as vg_c
import pvmcli.help.help_logical_volume as help_lv
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc
import pvmcli.util.util as util

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.storage as stor
import pypowervm.wrappers.virtual_io_server as vios_w


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] LogicalVolume list "
                    "{--parent-id | -p} <VirtualIOServer field>=value "
                    "{--child-id | -c} <VolumeGroup field>=value "
                    "{--object-id | -i} <VirtualDisk field>=value "
                    "{args}"])

DISPLAY_FIELD_HELP = (
    ["    LogicalVolume.capacity\n"
     "    LogicalVolume.emulate_model\n"
     "    LogicalVolume.file_format\n"
     "    LogicalVolume.label\n"
     "    LogicalVolume.name\n"
     "    LogicalVolume.read_iops_limit\n"
     "    LogicalVolume.tag\n"
     "    LogicalVolume.udid\n"
     "    LogicalVolume.vdtype\n"
     "    LogicalVolume.vg_uri\n"
     "    LogicalVolume.write_iops_limit\n"])


class LogicalVolumeTest(twa.TestWrapper):
    ''' Test framework for LogicalVolume Object Type. '''
    @property
    def file(self):
        return 'fake_volume_group.txt'

    @property
    def wrapper_class_to_test(self):
        return stor.VG

    def test_logical_volume(self):
        lv = lv_c.LogicalVolume()

        test_util.test_obj_type(self, lv, 'LogicalVolume', stor.VDisk,
                                field_models=[stor.VDisk],
                                name_to_ops={op.CREATE_NAME, op.LIST_NAME,
                                             'upload', op.DELETE_NAME,
                                             op.UPDATE_NAME},
                                exp_help=help_lv.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={lv.name_to_ops['list']: [vios_w.VIOS],
                                    lv.name_to_ops['create']: [vios_w.VIOS],
                                    lv.name_to_ops['upload']: None,
                                    lv.name_to_ops['delete']: [vios_w.VIOS],
                                    lv.name_to_ops['update']: [vios_w.VIOS]},
            child_model_ops={lv.name_to_ops['list']: [stor.VG],
                             lv.name_to_ops['create']: [stor.VG],
                             lv.name_to_ops['upload']: None,
                             lv.name_to_ops['delete']: [stor.VG],
                             lv.name_to_ops['update']: [stor.VG]},
            elem_wrp_func_ops={lv.name_to_ops['list']:
                               [stor.VG.virtual_disks],
                               lv.name_to_ops['create']:
                               [stor.VG.virtual_disks],
                               lv.name_to_ops['upload']: None,
                               lv.name_to_ops['delete']:
                               [stor.VG.virtual_disks],
                               lv.name_to_ops['update']:
                               [stor.VG.virtual_disks]})

    def test_overrides(self):
        ''' Test the overridden Field objects for LV ObjectType '''
        lv = lv_c.LogicalVolume()

        self.assertEqual(lv.get_schema(stor.VDisk.schema_type),
                         vg_c.LV_SCHEMA)

    @mock.patch("pypowervm.adapter.Adapter.read_by_href")
    @mock.patch("pypowervm.util.get_req_path_uuid")
    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.search")
    def test_list(self, vios_mock, uuid_util_mock, read_href_mock):
        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_mock.return_value = [sample_vios_w]
        vg_wrap = stor.VG.wrap(self.resp)

        vg_wrap2 = copy.deepcopy(vg_wrap)
        vg_wrap2.name = 'longer_vg_name'
        v_disks = vg_wrap2.virtual_disks
        v_disks.append(copy.deepcopy(vg_wrap2.virtual_disks[0]))
        v_disks[1].name = "2nd_lv"
        vg_wrap2.virtual_disks = v_disks
        vg_wrap2.virtual_disks[0].name = 'longer_lv_name'
        test_util.test_list(
            self, 'LogicalVolume',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Logical Volumes\n"
             "+-------------+----------------+--------------+-----------+"
             "----------------+--------------+\n"
             "|     VIOS    |    VG Name     | VG Size (GB) | Free (GB) |"
             "    LV Name     | LV Size (GB) |\n"
             "+-------------+----------------+--------------+-----------+"
             "----------------+--------------+\n"
             "| vios_1_Drew |   image_pool   |    1063.3    |   1051.2  |"
             "      asdf      |     1.0      |\n"
             "| vios_1_Drew | longer_vg_name |    1063.3    |   1051.2  |"
             " longer_lv_name |     1.0      |\n"
             "|             |                |              |           |"
             "     2nd_lv     |     1.0      |\n"
             "+-------------+----------------+--------------+-----------+"
             "----------------+--------------+\n"],
            ["-----------------\n| LogicalVolume |\n-----------------\n  capa"
             "city=1.0\n  emulate_model=True\n  label=None\n  name=asdf\n  "
             "read_iops_limit=44\n  "
             "udid=0300f8d6de00004b000000014a54555cd9.1\n  vg_uri=https://9.1."
             "2.3:12443/rest/api/uom/VirtualIOServer/14B854F7-42CE-4FF0-BD57-1"
             "D117054E701/VolumeGroup/b6bdbf1f-eddf-3c81-8801-9859eb6fedcb\n  "
             "write_iops_limit=55\n"],
            ['<VirtualDisk xmlns="http://www.ibm.com/xmlns/systems/power/fir'],
            ['capacity', 'name'],
            ["capacity=1.0,name=asdf\n"],
            ["1.0,asdf\n"],
            ["capacity___1.0,name___asdf\n"],
            ["capacity=1.0___name=asdf\n"],
            ["capacity=1.0,name=asdf\n"],
            ["capacity___1.0___name___asdf\n"],
            {'child_type': 'VolumeGroup',
             'root_id': '014C1F78-B210-40B2-A12E-C82316B878CA',
             'root_type': 'VirtualIOServer'},
            wrps=[vg_wrap.virtual_disks[0]], c_wrps=[vg_wrap, vg_wrap2],
            p_wrps=[sample_vios_w])

    @mock.patch("pypowervm.wrappers.entry_wrapper.EntryWrapper.search")
    @mock.patch("pypowervm.wrappers.storage.VG.update")
    def test_create(self, vg_update_mock, search_mock):
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['lv', 'create', '--help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] LogicalVolume create "
                      "{--parent-id | -p} <VirtualIOServer field>=value "
                      "{--child-id | -c} <VolumeGroup field>=value "
                      "{--object-id | -i} <VirtualDisk field>=value {args}",
                      help_out.getvalue())

        vg_wrap = stor.VG.wrap(self.resp)

        cp = test_util.setup_mock(self, args=['lv', 'create', '-c',
                                              'name=rootvg', '--name',
                                              'lvname', '--size', '10'],
                                  adpt_fx=self.adpt,
                                  c_wrps=[vg_wrap])
        cp.obj_type.execute()
        vg_update_mock.assert_called_once_with()
        vg_update_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['lv', 'create', '-c',
                                              'name=rootvg', '--name',
                                              'lvname', '--size', '10'],
                                  adpt_fx=self.adpt,
                                  c_wrps=[vg_wrap])
        cp.obj_type.execute()
        vg_update_mock.assert_called_once_with()
        vg_update_mock.reset_mock()

        cp = test_util.setup_mock(self, args=['lv', 'create', '-c',
                                              'name=rootvg', '--name',
                                              'lvname', '--size', '10'],
                                  adpt_fx=self.adpt,
                                  c_wrps=[vg_wrap])
        cp.obj_type.execute()
        vg_update_mock.assert_called_once_with()
        vg_update_mock.reset_mock()

        self.assertRaises(
            SystemExit, test_util.setup_mock, self,
            args=['lv', 'create', '-c', 'name=rootvg', '--size', '10',
                  '--label', 'imalabel'], adpt_fx=self.adpt,
            wrps=[vg_wrap.virtual_disks[0]])
        vg_update_mock.assert_not_called()

        self.assertRaises(
            SystemExit, test_util.setup_mock, self,
            args=['lv', 'create', '-c', 'name=rootvg', '--name', 'lvname',
                  '--label', 'imalabel'], adpt_fx=self.adpt,
            wrps=[vg_wrap.virtual_disks[0]])
        vg_update_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.storage.VG.search')
    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.tasks.storage.upload_new_vdisk')
    def test_upload(self, vdisk_mock, vios_search_mock, vg_search_mock):
        ret_vios = vios_w.VIOS.bld(
            mock.Mock(), "test_vios",
            bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp_w.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        ret_vios.set_parm_value(bp_w._BP_STATE, bp_w.LPARState.RUNNING)
        ret_vios.set_parm_value(bp_w._BP_RMC_STATE, bp_w.RMCState.ACTIVE)
        vios_search_mock.return_value = [ret_vios]

        vg_wrp = stor.VG.wrap(self.resp)
        vg_search_mock.return_value = [vg_wrp]

        # Minimum args
        cp = test_util.setup_mock(self, args=['lv', 'upload', '--vios',
                                              'test_vios', '--vg', 'vg1',
                                              '--name', 'lv_name',
                                              '--data', 'lv_data'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])

        mopen = test_util.mock_open_str()
        mosz = 'os.path.getsize'
        with mock.patch(mopen) as mock_open, mock.patch(mosz) as mock_getsize:
            cp.obj_type.execute()
            vdisk_mock.assert_called_once_with(
                cp.obj_type.ec.sm.adapter, None,
                'b6bdbf1f-eddf-3c81-8801-9859eb6fedcb',
                mock_open().__enter__(), 'lv_name',
                mock_getsize(), sha_chksum=None, d_size=None)
            vg_search_mock.assert_called_once_with(
                cp.obj_type.ec.sm.adapter, name='vg1',
                parent_uuid=None, parent_type=vios_w.VIOS)
        vdisk_mock.reset_mock()
        vg_search_mock.reset_mock()
        # Reset VG cache to imitate fresh execution
        util.name_cache = {}

        # All args
        cp = test_util.setup_mock(self, args=['lv', 'upload', '--vios',
                                              'test_vios', '--vg', 'vg1',
                                              '--name', 'lv_name',
                                              '--data', 'lv_data',
                                              '--checksum', 'abc',
                                              '--size', '1'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])

        with mock.patch(mopen) as mock_open, mock.patch(mosz) as mock_getsize:
            cp.obj_type.execute()
            vdisk_mock.assert_called_once_with(
                cp.obj_type.ec.sm.adapter, None,
                'b6bdbf1f-eddf-3c81-8801-9859eb6fedcb',
                mock_open().__enter__(), 'lv_name',
                mock_getsize(), sha_chksum='abc', d_size=1)
            vg_search_mock.assert_called_once_with(
                cp.obj_type.ec.sm.adapter, name='vg1',
                parent_uuid=None, parent_type=vios_w.VIOS)
        vdisk_mock.reset_mock()
        vg_search_mock.reset_mock()

        # No args
        self.assertRaises(SystemExit, test_util.setup_mock, self,
                          args=['lv', 'upload'], adpt_fx=self.adpt,
                          wrps=[self.dwrap])
        vdisk_mock.assert_not_called()

        # No data specified and no data size specified
        cp = test_util.setup_mock(
            self, args=['lv', 'upload', '--vios', 'test_vios', '--name',
                        'lv_name', '--vg', 'vg1'],
            adpt_fx=self.adpt, wrps=[self.dwrap])
        self.assertRaises(
            exc.UploadStdInRequiresSizeError, cp.obj_type.execute)
        vdisk_mock.assert_not_called()

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'LogicalVolume')

    @mock.patch('pypowervm.wrappers.storage.VG.search')
    @mock.patch('pypowervm.wrappers.storage.VG.update')
    def test_delete(self, vg_upd_mock, vg_srch_mock):
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['lv', 'delete', '--help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] LogicalVolume delete "
                      "{--parent-id | -p} <VirtualIOServer field>=value "
                      "{--child-id | -c} <VolumeGroup field>=value "
                      "{--object-id | -i} <VirtualDisk field>=value {args}",
                      help_out.getvalue())

        vg_wrap = stor.VG.wrap(self.resp)
        # No object-id (required, argparse catches with SystemExit)
        self.assertRaises(SystemExit, test_util.setup_mock,
                          self, args=['lv', 'delete', '-c', 'name=rootvg'],
                          adpt_fx=self.adpt, c_wrps=[vg_wrap],
                          wrps=[vg_wrap.virtual_disks[0]])
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # No child-id (required, argparse catches with SystemExit)
        self.assertRaises(SystemExit, test_util.setup_mock,
                          self, args=['lv', 'delete', '-i', 'name=asdf'],
                          adpt_fx=self.adpt, c_wrps=[vg_wrap],
                          wrps=[vg_wrap.virtual_disks[0]])
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Multiple children
        cp = test_util.setup_mock(self, args=['lv', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=asdf'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap, vg_wrap],
                                  wrps=[vg_wrap.virtual_disks[0]])
        self.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Multiple wrappers
        cp = test_util.setup_mock(self, args=['lv', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=asdf'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap],
                                  wrps=[vg_wrap.virtual_disks[0],
                                        vg_wrap.virtual_disks[0]])
        self.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Minimum args
        sample_vios_w = test_util.mock_vios(self.adpt)
        cp = test_util.setup_mock(self, args=['lv', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=asdf'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap],
                                  wrps=[vg_wrap.virtual_disks[0]],
                                  p_wrps=[sample_vios_w])
        self.assertEqual(vg_wrap.virtual_disks[0].name, 'asdf')
        self.assertEqual(len(vg_wrap.virtual_disks), 1)
        cp.obj_type.execute()
        self.assertEqual(len(vg_wrap.virtual_disks), 0)
        vg_srch_mock.assert_called_with(
            cp.obj_type.ec.sm.adapter,
            parent_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            parent_type='VirtualIOServer', xag=[], name='rootvg')
        vg_upd_mock.assert_called_once_with()
        vg_upd_mock.reset_mock()

    @mock.patch('pypowervm.wrappers.storage.VG.update')
    def test_update(self, vg_upd_mock):
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['lv', 'update', '--help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] LogicalVolume update "
                      "{--parent-id | -p} <VirtualIOServer field>=value "
                      "{--child-id | -c} <VolumeGroup field>=value "
                      "{--object-id | -i} <VirtualDisk field>=value {args}",
                      help_out.getvalue())

        vg_wrap = stor.VG.wrap(self.resp)

        cp = test_util.setup_mock(self, args=['lv', 'update', '-c',
                                              'name=rootvg', '-i',
                                              'name=asdf', '-s',
                                              'read_iops_limit=22',
                                              'write_iops_limit=33'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap],
                                  wrps=[vg_wrap.virtual_disks[0]])
        self.assertEqual(vg_wrap.virtual_disks[0].read_iops_limit, 44)
        self.assertEqual(vg_wrap.virtual_disks[0].write_iops_limit, 55)
        cp.obj_type.execute()
        self.assertEqual(vg_wrap.virtual_disks[0].read_iops_limit, 22)
        self.assertEqual(vg_wrap.virtual_disks[0].write_iops_limit, 33)
        vg_upd_mock.assert_called_once_with()
        vg_upd_mock.reset_mock()
