# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.shared_processor_pool as spp_c
import pvmcli.help.help_shared_proc_pool as help_spp
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.tests.test_utils.test_wrapper_abc as twa
from pypowervm.wrappers import base_partition as bp_w
from pypowervm.wrappers import logical_partition as lpar_w
from pypowervm.wrappers import managed_system as ms_w
from pypowervm.wrappers import shared_proc_pool


LIST_USAGE_HELP = (
    ["usage: pvmctl [help | --help] SharedProcessorPool list {--object-id | "
     "-i} <SharedProcessorPool field>=value {args}"])


UPDATE_USAGE_HELP = (["usage: pvmctl [help | --help] SharedProcessorPool "
                      "update {--object-id | -i} <SharedProcessorPool field>"
                      "=value {args}"])


DELETE_USAGE_HELP = (["usage: pvmctl [help | --help] SharedProcessorPool "
                      "delete {--object-id | -i} <SharedProcessorPool "
                      "field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    SharedProcessorPool.avail_proc_units\n"
     "    SharedProcessorPool.curr_rsrv_proc_units\n"
     "    SharedProcessorPool.id\n"
     "    SharedProcessorPool.is_default\n"
     "    SharedProcessorPool.max_proc_units\n"
     "    SharedProcessorPool.name\n"
     "    SharedProcessorPool.pend_rsrv_proc_units\n"
     "    SharedProcessorPool.uuid\n"])

SET_FIELD_HELP = (
    ["--set-fields options:\n"
     "    SharedProcessorPool.max_proc_units\n"
     "    SharedProcessorPool.name\n"
     "    SharedProcessorPool.pend_rsrv_proc_units\n"])


class SharedProcessorPoolTest(twa.TestWrapper):
    ''' Test framework for SharedProcessorPool Object Type. '''
    @property
    def file(self):
        return 'shrprocpool.txt'

    @property
    def wrapper_class_to_test(self):
        return shared_proc_pool.SharedProcPool

    def test_shared_storage_pool(self):
        spp = spp_c.SharedProcessorPool()

        test_util.test_obj_type(self, spp, 'SharedProcessorPool',
                                shared_proc_pool.SharedProcPool,
                                field_models=[shared_proc_pool.SharedProcPool],
                                name_to_ops={'list', 'update',
                                             'delete', 'create'},
                                exp_help=help_spp.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={spp.name_to_ops['list']: [ms_w.System],
                                    spp.name_to_ops['update']: [ms_w.System],
                                    spp.name_to_ops['delete']: [ms_w.System],
                                    spp.name_to_ops['create']: [ms_w.System]},
            child_model_ops={spp.name_to_ops['list']: None,
                             spp.name_to_ops['update']: None,
                             spp.name_to_ops['delete']: None,
                             spp.name_to_ops['create']: None},
            elem_wrp_func_ops={spp.name_to_ops['list']: None,
                               spp.name_to_ops['update']: None,
                               spp.name_to_ops['delete']: None,
                               spp.name_to_ops['create']: None})

    def test_list(self):
        spp_wrps = shared_proc_pool.SharedProcPool.wrap(self.resp)
        lp_wrp_str = "pypowervm.wrappers.logical_partition.LPAR.wrap"
        v_wrp_str = "pypowervm.wrappers.virtual_io_server.VIOS.wrap"
        with mock.patch(lp_wrp_str) as lp_mk, mock.patch(v_wrp_str) as vi_mk:
            lp_0 = lpar_w.LPAR.bld(
                mock.Mock(), "test_lpar_0",
                bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
                bp_w.PartitionProcessorConfiguration.bld_shared(mock.Mock(),
                                                                1, 1))
            lp_0.proc_config.shared_proc_cfg.pool_id = 0
            lp_1 = lpar_w.LPAR.bld(
                mock.Mock(), "test_lpar_1",
                bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
                bp_w.PartitionProcessorConfiguration.bld_shared(mock.Mock(),
                                                                1, 1))
            lp_1.proc_config.shared_proc_cfg.pool_id = 1
            lp_2 = lpar_w.LPAR.bld(
                mock.Mock(), "test_lpar_2",
                bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
                bp_w.PartitionProcessorConfiguration.bld_shared(mock.Mock(),
                                                                1, 1))
            lp_2.proc_config.shared_proc_cfg.pool_id = 2
            lp_3 = lpar_w.LPAR.bld(
                mock.Mock(), "test_lpar_3",
                bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
                bp_w.PartitionProcessorConfiguration.bld_shared(mock.Mock(),
                                                                1, 1))
            lp_3.proc_config.shared_proc_cfg.pool_id = 1
            lp_mk.return_value = [lp_0, lp_1, lp_2, lp_3]
            vi_mk.return_value = []

            test_util.test_list(
                self, 'SharedProcessorPool',
                LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
                ["Shared Processor Pools\n"
                 "+----+--------------+-----------+-------+----------+"
                 "--------------+\n"
                 "| ID |     Name     | Available |  Max  | Reserved |"
                 "  Partitions  |\n"
                 "+----+--------------+-----------+-------+----------+"
                 "--------------+\n"
                 "| 0  | DefaultPool  |    ----   |  ---- |   ----   |"
                 " test_lpar_0  |\n"
                 "| 1  | SharedPool01 |    5.05   | 10.25 |   6.15   |"
                 " test_lpar_1, |\n"
                 "|    |              |           |       |          |"
                 " test_lpar_3  |\n"
                 "+----+--------------+-----------+-------+----------+"
                 "--------------+\n"],
                ["-----------------------\n"
                 "| SharedProcessorPool |\n"
                 "-----------------------\n"
                 "  avail_proc_units=0.0\n"
                 "  curr_rsrv_proc_units=0\n"
                 "  id=0\n"
                 "  is_default=True\n"
                 "  max_proc_units=0\n"
                 "  name=DefaultPool\n"
                 "  pend_rsrv_proc_units=0\n"
                 "  uuid=4b854a40-4f9a-3848-bc1f-e278b4d8ce04\n"
                 "-----------------------\n"
                 "| SharedProcessorPool |\n"
                 "-----------------------\n"
                 "  avail_proc_units=5.05\n"
                 "  curr_rsrv_proc_units=5.35\n"
                 "  id=1\n"
                 "  is_default=False\n"
                 "  max_proc_units=10.25\n"
                 "  name=SharedPool01\n"
                 "  pend_rsrv_proc_units=6.15\n"
                 "  uuid=f55a0681-aed0-321a-9df0-3968a1dfe45d\n"],
                ['<SharedProcessorPool:SharedProcessorPool xmlns:SharedProce'],
                ['SharedProcessorPool.id', 'is_default'],
                ["id=1,is_default=False\n"],
                ["1,False\n"],
                ["id___1,is_default___False\n"],
                ["id=1___is_default=False\n"],
                ["id=1,is_default=False\n"],
                ["id___1___is_default___False\n"],
                {'root_type': 'ManagedSystem',
                 'child_type': 'SharedProcessorPool',
                 'root_id': 'caae9209-25e5-35cd-a71a-ed55c03f294d'},
                wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]],
                p_wrps=[test_util.mock_sys(self.adpt)])

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_update(self, ew_update_mock):
        help_strs = UPDATE_USAGE_HELP + SET_FIELD_HELP
        test_util.test_update(
            self, ew_update_mock, 'SharedProcessorPool',
            ["name=abcd", "SharedProcessorPool.max_proc_units=3"],
            "SharedProcessorPool.notathing=123", help_strs,
            "Set field SharedProcessorPool.notathing=123 has an improper "
            "format. Please specify the set field as <Field>=<Value>. Run the "
            "command 'pvmctl SharedProcessorPool update help' for a list of "
            "valid set fields.")

    @mock.patch('pypowervm.wrappers.shared_proc_pool.SharedProcPool.get')
    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_create(self, ew_mock_update, spp_get):
        ''' Test create operation on SharedProcessorPool ObjectType.

        :param ew_mock_update: Mock for update call as a create operation on
                               SPP is updating "hidden" wrappers.
        '''
        spp_wrps = shared_proc_pool.SharedProcPool.wrap(self.resp)
        spp_get.return_value = [spp_wrps[0], spp_wrps[1], spp_wrps[2]]
        # pvmctl spp create help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['spp', 'create', 'help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] SharedProcessorPool "
                      "create {--object-id | -i} <SharedProcessorPool "
                      "field>=value {args}", help_out.getvalue())
        ew_mock_update.assert_not_called()

        # pvmctl spp create --name spp_name --max 4
        cp = test_util.setup_mock(
            self, ['spp', 'create', '--name', 'spp_name', '--max', '4'],
            self.adpt, wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]])

        cp.obj_type.execute()
        self.assertEqual(spp_wrps[2].name, 'spp_name')
        self.assertEqual(spp_wrps[2].max_proc_units, 4)
        ew_mock_update.assert_called_once_with()
        ew_mock_update.reset_mock()

        # On second attempt, we will fail because there are no "free" SPPs
        # pvmctl spp create --name spp_name --max 4
        cp = test_util.setup_mock(
            self, ['spp', 'create', '--name', 'spp_name', '--max', '4'],
            self.adpt, wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]])

        self.assertRaises(exc.NoFreeSPPError, cp.obj_type.execute)

        # Make spp_wrps[2] available for create
        spp_wrps[2].max_proc_units = 0

        # pvmctl spp create --name spp_name --max 2 --reserved 1
        cp = test_util.setup_mock(
            self, ['spp', 'create', '--name', 'spp_name', '--max', '2',
                   '--reserved', '1'],
            self.adpt, wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]])

        cp.obj_type.execute()
        self.assertEqual(spp_wrps[2].name, 'spp_name')
        self.assertEqual(spp_wrps[2].max_proc_units, 2)
        self.assertEqual(spp_wrps[2].pend_rsrv_proc_units, 1)
        ew_mock_update.assert_called_once_with()
        ew_mock_update.reset_mock()

        # Make spp_wrps[2] available for create
        spp_wrps[2].max_proc_units = 0

        # pvmctl spp create --name spp_name --max 5 --reserved 3 --spp-id 9
        cp = test_util.setup_mock(
            self, ['spp', 'create', '--name', 'spp_name', '--max', '5',
                   '--reserved', '3', '--spp-id', '9'],
            self.adpt, wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]])

        self.assertRaises(exc.SpecifiedSPPNotFreeError, cp.obj_type.execute)

        # Make spp_wrps[1] available for create
        spp_wrps[1].max_proc_units = 0

        # pvmctl spp create --name spp_name --max 5 --reserved 3 --spp-id 1
        cp = test_util.setup_mock(
            self, ['spp', 'create', '--name', 'spp_name', '--max', '5',
                   '--reserved', '3', '--spp-id', '1'],
            self.adpt, wrps=[spp_wrps[0], spp_wrps[1], spp_wrps[2]])

        cp.obj_type.execute()
        self.assertEqual(spp_wrps[1].name, 'spp_name')
        self.assertEqual(spp_wrps[1].max_proc_units, 5)
        self.assertEqual(spp_wrps[1].pend_rsrv_proc_units, 3)
        self.assertEqual(spp_wrps[1].id, 1)
        ew_mock_update.assert_called_once_with()
        ew_mock_update.reset_mock()

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_delete(self, ew_mock_update):
        ''' Test delete operation on SharedProcessorPool ObjectType.

        :param ew_mock_update: Mock for update call as a delete operation on
                               SPP is simply setting max_proc_units to 0.
        '''
        self.dwrap.pend_rsrv_proc_units = 1
        self.dwrap.max_proc_units = 1
        self.assertEqual(self.dwrap.pend_rsrv_proc_units, 1)
        self.assertEqual(self.dwrap.max_proc_units, 1)
        test_util.test_delete(self, ew_mock_update, 'SharedProcessorPool',
                              DELETE_USAGE_HELP, wrps=[self.dwrap],
                              p_wrps=[test_util.mock_sys(self.adpt)])
        self.assertEqual(self.dwrap.pend_rsrv_proc_units, 0)
        self.assertEqual(self.dwrap.max_proc_units, 0)
