# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.sriov_adapter as sriov_c
import pvmcli.help.help_sriov_adapter as help_sriov
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.iocard as iocard
import pypowervm.wrappers.managed_system as msys


LIST_USAGE_HELP = (
    ["usage: pvmctl [help | --help] SRIOVAdapter list {--object-id | "
     "-i} <SRIOVAdapter field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    PhysicalPort.allocated_capacity\n"
     "    PhysicalPort.cfg_lps\n"
     "    PhysicalPort.cfg_max_lps\n"
     "    PhysicalPort.curr_speed\n"
     "    PhysicalPort.flow_ctl\n"
     "    PhysicalPort.label\n"
     "    PhysicalPort.link_status\n"
     "    PhysicalPort.loc_code\n"
     "    PhysicalPort.min_granularity\n"
     "    PhysicalPort.mtu\n"
     "    PhysicalPort.port_id\n"
     "    PhysicalPort.sriov_adap_id\n"
     "    PhysicalPort.sublabel\n"
     "    PhysicalPort.supp_max_lps\n"
     "    PhysicalPort.switch_mode\n"
     "    SRIOVAdapter.description\n"
     "    SRIOVAdapter.dev_name\n"
     "    SRIOVAdapter.dev_type\n"
     "    SRIOVAdapter.drc_name\n"
     "    SRIOVAdapter.id\n"
     "    SRIOVAdapter.mode\n"
     "    SRIOVAdapter.personality\n"
     "    SRIOVAdapter.phys_loc_code\n"
     "    SRIOVAdapter.phys_ports\n"
     "    SRIOVAdapter.sriov_adap_id\n"
     "    SRIOVAdapter.state\n"
     "    SRIOVAdapter.udid\n"])

UPDATE_USAGE_HELP = (
    ["pvmctl [help | --help] SRIOVAdapter update {--object-id | -i} "
     "<SRIOVAdapter field>=value {args}"])

SET_FIELD_HELP = (
    ["--set-fields options:\n"
     "    PhysicalPort.cfg_max_lps\n"
     "    PhysicalPort.flow_ctl\n"
     "    PhysicalPort.label\n"
     "    PhysicalPort.mtu\n"
     "    PhysicalPort.sublabel\n"
     "    PhysicalPort.switch_mode\n"
     "    SRIOVAdapter.mode\n"])


class SRIOVAdapterTest(twa.TestWrapper):
    ''' Test framework for SRIOVAdapter Object Type. '''
    @property
    def file(self):
        return 'sys_with_sriov.txt'

    @property
    def wrapper_class_to_test(self):
        return msys.System

    def test_sriov_adapter(self):
        sriov = sriov_c.SRIOVAdapter()

        test_util.test_obj_type(self, sriov, 'SRIOVAdapter',
                                iocard.SRIOVAdapter,
                                field_models=[iocard.SRIOVAdapter,
                                              iocard.SRIOVConvPPort,
                                              iocard.SRIOVEthPPort],
                                name_to_ops={'list', 'update'},
                                exp_help=help_sriov.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={sriov.name_to_ops['list']: [msys.System]},
            child_model_ops={sriov.name_to_ops['list']: None},
            elem_wrp_func_ops={sriov.name_to_ops['list']: [
                msys.System.asio_config, msys.ASIOConfig.sriov_adapters]})

        test_util.test_ops(
            self,
            parent_model_ops={sriov.name_to_ops['update']: [msys.System]},
            child_model_ops={sriov.name_to_ops['update']: None},
            elem_wrp_func_ops={sriov.name_to_ops['update']: [
                msys.System.asio_config, msys.ASIOConfig.sriov_adapters]})

    def test_list(self):
        test_util.test_list(
            self, 'SRIOVAdapter',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["SR-IOV Adapters\n"
             "+-----+---------+-----------+----------------------------+"
             "-----------+----------+--------+--------+\n"
             "|  ID |  State  |    Mode   |          Loc Code          |"
             " Port Stat | Port Lbl | Av Cap | Av LPs |\n"
             "+-----+---------+-----------+----------------------------+"
             "-----------+----------+--------+--------+\n"
             "|  1  | Running |   Sriov   | U78C7.001.RCH0004-P1-C8-T1 |"
             "     Up    |   None   | 98.0%  |   18   |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C8-T2 |"
             "    Down   |   None   | 98.0%  |   20   |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C8-T3 |"
             "     Up    |   None   | 98.0%  |   3    |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C8-T4 |"
             "    Down   |   None   | 98.0%  |   4    |\n"
             "|  2  | Running |   Sriov   | U78C7.001.RCH0004-P1-C9-T1 |"
             "     Up    |   None   | 98.0%  |   18   |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C9-T2 |"
             "    Down   |   None   | 98.0%  |   20   |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C9-T3 |"
             "     Up    |   None   | 98.0%  |   3    |\n"
             "|     |         |           | U78C7.001.RCH0004-P1-C9-T4 |"
             "    Down   |   None   | 98.0%  |   4    |\n"
             "| N/A |         | Dedicated |  U78CB.001.WZS06RG-P1-C7   |"
             "           |          |        |        |\n"
             "+-----+---------+-----------+----------------------------+"
             "-----------+----------+--------+--------+\n"],
            ["----------------\n"
             "| SRIOVAdapter |\n"
             "----------------\n"
             "  description=PCIe2 4-port (10Gb FCoE & 1GbE) SR&RJ45 Adapter\n"
             "  id=21010020\n"
             "  mode=Sriov\n"
             "  personality=MIN_CAPACITY\n"
             "  phys_loc_code=U78C7.001.RCH0004-P1-C8\n"
             "  sriov_adap_id=1\n"
             "  state=Running\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=2\n"
             "    cfg_max_lps=20\n"
             "    curr_speed=E10Mbps\n"
             "    flow_ctl=False\n"
             "    link_status=Up\n"
             "    loc_code=U78C7.001.RCH0004-P1-C8-T1\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=0\n"
             "    sriov_adap_id=1\n"
             "    supp_max_lps=20\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=0\n"
             "    cfg_max_lps=20\n"
             "    curr_speed=Unknown\n"
             "    flow_ctl=False\n"
             "    link_status=Down\n"
             "    loc_code=U78C7.001.RCH0004-P1-C8-T2\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=1\n"
             "    sriov_adap_id=1\n"
             "    supp_max_lps=20\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=1\n"
             "    cfg_max_lps=4\n"
             "    curr_speed=E1Gbps\n"
             "    flow_ctl=False\n"
             "    link_status=Up\n"
             "    loc_code=U78C7.001.RCH0004-P1-C8-T3\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=2\n"
             "    sriov_adap_id=1\n"
             "    supp_max_lps=4\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=0\n"
             "    cfg_max_lps=4\n"
             "    curr_speed=Unknown\n"
             "    flow_ctl=False\n"
             "    link_status=Down\n"
             "    loc_code=U78C7.001.RCH0004-P1-C8-T4\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=3\n"
             "    sriov_adap_id=1\n"
             "    supp_max_lps=4\n"
             "    switch_mode=Veb\n"
             "----------------\n"
             "| SRIOVAdapter |\n"
             "----------------\n"
             "  description=PCIe2 4-port (10Gb FCoE & 1GbE) SR&RJ45 Adapter\n"
             "  id=21010023\n"
             "  mode=Sriov\n"
             "  personality=MAX_MIN_CAPACITY\n"
             "  phys_loc_code=U78C7.001.RCH0004-P1-C9\n"
             "  sriov_adap_id=2\n"
             "  state=Running\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=2\n"
             "    cfg_max_lps=20\n"
             "    curr_speed=E10Mbps\n"
             "    flow_ctl=False\n"
             "    link_status=Up\n"
             "    loc_code=U78C7.001.RCH0004-P1-C9-T1\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=0\n"
             "    sriov_adap_id=2\n"
             "    supp_max_lps=20\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=0\n"
             "    cfg_max_lps=20\n"
             "    curr_speed=Unknown\n"
             "    flow_ctl=False\n"
             "    link_status=Down\n"
             "    loc_code=U78C7.001.RCH0004-P1-C9-T2\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=1\n"
             "    sriov_adap_id=2\n"
             "    supp_max_lps=20\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=1\n"
             "    cfg_max_lps=4\n"
             "    curr_speed=E1Gbps\n"
             "    flow_ctl=False\n"
             "    link_status=Up\n"
             "    loc_code=U78C7.001.RCH0004-P1-C9-T3\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=2\n"
             "    sriov_adap_id=2\n"
             "    supp_max_lps=4\n"
             "    switch_mode=Veb\n"
             "  ----------------\n"
             "  | PhysicalPort |\n"
             "  ----------------\n"
             "    allocated_capacity=2.00%\n"
             "    cfg_lps=0\n"
             "    cfg_max_lps=4\n"
             "    curr_speed=Unknown\n"
             "    flow_ctl=False\n"
             "    link_status=Down\n"
             "    loc_code=U78C7.001.RCH0004-P1-C9-T4\n"
             "    min_granularity=0.02\n"
             "    mtu=E_1500\n"
             "    port_id=3\n"
             "    sriov_adap_id=2\n"
             "    supp_max_lps=4\n"
             "    switch_mode=Veb\n"
             "----------------\n"
             "| SRIOVAdapter |\n"
             "----------------\n"
             "  description=PCIe2 4-port (10Gb FCoE & 1GbE) SR&RJ45 Adapter\n"
             "  id=21010010\n"
             "  mode=Dedicated\n"
             "  phys_loc_code=U78CB.001.WZS06RG-P1-C7\n"
             "  state=NotConfigured\n"],
            ['<SRIOVAdapter xmlns="http://www.ibm.com/xmlns/systems/power/'],
            ['SRIOVAdapter.phys_loc_code', 'link_status',
             'PhysicalPort.cfg_max_lps'],
            ["phys_loc_code=U78C7.001.RCH0004-P1-C8,link_status=Up,"
             "cfg_max_lps=20\nphys_loc_code=U78C7.001.RCH0004-P1-C8,"
             "link_status=Down,cfg_max_lps=20\nphys_loc_code=U78C7.001"
             ".RCH0004-P1-C8,link_status=Up,cfg_max_lps=4\n"
             "phys_loc_code=U78C7.001.RCH0004-P1-C8,link_status=Down,"
             "cfg_max_lps=4\nphys_loc_code=U78C7.001.RCH0004-P1-C9,"
             "link_status=Up,cfg_max_lps=20\nphys_loc_code=U78C7.001."
             "RCH0004-P1-C9,link_status=Down,cfg_max_lps=20\n"
             "phys_loc_code=U78C7.001.RCH0004-P1-C9,link_status=Up,"
             "cfg_max_lps=4\nphys_loc_code=U78C7.001.RCH0004-P1-C9,"
             "link_status=Down,cfg_max_lps=4\n"],
            ["U78C7.001.RCH0004-P1-C8,Up,20\nU78C7.001.RCH0004-P1-C8,Down,20\n"
             "U78C7.001.RCH0004-P1-C8,Up,4\nU78C7.001.RCH0004-P1-C8,Down,4\n"
             "U78C7.001.RCH0004-P1-C9,Up,20\nU78C7.001.RCH0004-P1-C9,Down,20\n"
             "U78C7.001.RCH0004-P1-C9,Up,4\nU78C7.001.RCH0004-P1-C9,Down,4\n"],
            ["phys_loc_code___U78C7.001.RCH0004-P1-C8,link_status___Up,"
             "cfg_max_lps___20\nphys_loc_code___U78C7.001.RCH0004-P1-C8,"
             "link_status___Down,cfg_max_lps___20\nphys_loc_code___"
             "U78C7.001.RCH0004-P1-C8,link_status___Up,cfg_max_lps___4\n"
             "phys_loc_code___U78C7.001.RCH0004-P1-C8,link_status___Down,"
             "cfg_max_lps___4\nphys_loc_code___U78C7.001.RCH0004-P1-C9,"
             "link_status___Up,cfg_max_lps___20\nphys_loc_code___"
             "U78C7.001.RCH0004-P1-C9,link_status___Down,cfg_max_lps___20\n"
             "phys_loc_code___U78C7.001.RCH0004-P1-C9,link_status___Up,"
             "cfg_max_lps___4\nphys_loc_code___U78C7.001.RCH0004-P1-C9,"
             "link_status___Down,cfg_max_lps___4\n"],
            ["phys_loc_code=U78C7.001.RCH0004-P1-C8___link_status=Up___"
             "cfg_max_lps=20\nphys_loc_code=U78C7.001.RCH0004-P1-C8___"
             "link_status=Down___cfg_max_lps=20\n"
             "phys_loc_code=U78C7.001.RCH0004-P1-C8___link_status=Up___"
             "cfg_max_lps=4\nphys_loc_code=U78C7.001.RCH0004-P1-C8___"
             "link_status=Down___cfg_max_lps=4\n"
             "phys_loc_code=U78C7.001.RCH0004-P1-C9___link_status=Up___"
             "cfg_max_lps=20\nphys_loc_code=U78C7.001.RCH0004-P1-C9___"
             "link_status=Down___cfg_max_lps=20\n"
             "phys_loc_code=U78C7.001.RCH0004-P1-C9___link_status=Up"
             "___cfg_max_lps=4\nphys_loc_code=U78C7.001.RCH0004-P1-C9"
             "___link_status=Down___cfg_max_lps=4\n"],
            ["phys_loc_code=U78C7.001.RCH0004-P1-C8,link_status=Up,"
             "cfg_max_lps=20___phys_loc_code=U78C7.001.RCH0004-P1-C8,"
             "link_status=Down,cfg_max_lps=20___"
             "phys_loc_code=U78C7.001.RCH0004-P1-C8,link_status=Up,"
             "cfg_max_lps=4___phys_loc_code=U78C7.001.RCH0004-P1-C8,"
             "link_status=Down,cfg_max_lps=4___"
             "phys_loc_code=U78C7.001.RCH0004-P1-C9,link_status=Up,"
             "cfg_max_lps=20___phys_loc_code=U78C7.001.RCH0004-P1-C9,"
             "link_status=Down,cfg_max_lps=20___"
             "phys_loc_code=U78C7.001.RCH0004-P1-C9,link_status=Up,"
             "cfg_max_lps=4___phys_loc_code=U78C7.001.RCH0004-P1-C9,"
             "link_status=Down,cfg_max_lps=4___"
             "phys_loc_code=U78CB.001.WZS06RG-P1-C7,link_status=,"
             "cfg_max_lps=\n"],
            ["phys_loc_code___U78C7.001.RCH0004-P1-C8___link_status___Up___"
             "cfg_max_lps___20___phys_loc_code___U78C7.001.RCH0004-P1-C8___"
             "link_status___Down___cfg_max_lps___20___phys_loc_code___"
             "U78C7.001.RCH0004-P1-C8___link_status___Up___cfg_max_lps___4___"
             "phys_loc_code___U78C7.001.RCH0004-P1-C8___link_status___Down___"
             "cfg_max_lps___4___phys_loc_code___U78C7.001.RCH0004-P1-C9___"
             "link_status___Up___cfg_max_lps___20___phys_loc_code___"
             "U78C7.001.RCH0004-P1-C9___link_status___Down___cfg_max_lps___"
             "20___phys_loc_code___U78C7.001.RCH0004-P1-C9___link_status___"
             "Up___cfg_max_lps___4___phys_loc_code___U78C7.001.RCH0004-P1-C9"
             "___link_status___Down___cfg_max_lps___4___phys_loc_code___"
             "U78CB.001.WZS06RG-P1-C7___link_status______cfg_max_lps___\n"],
            'ManagedSystem',
            wrps=[self.dwrap.asio_config.sriov_adapters[0],
                  self.dwrap.asio_config.sriov_adapters[1],
                  self.dwrap.asio_config.sriov_adapters[2]])

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_update(self, ew_update_mock):
        wrps = [self.dwrap.asio_config.sriov_adapters[0]]
        p_wrps = [self.dwrap]
        help_strs = (UPDATE_USAGE_HELP + test_util.EXPECTED_HELP +
                     DISPLAY_FIELD_HELP + SET_FIELD_HELP)

        # pvmctl <obj_name> update help
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ["SRIOVAdapter", 'update', 'help'], wrps, help_out=help_out)
        for help_str in help_strs:
            self.assertIn(help_str, help_out.getvalue())
        ew_update_mock.assert_not_called()
        ew_update_mock.reset_mock()

        # Happy path test
        args = ['sriov', 'update', '--port-loc', "U78C7.001.RCH0004-P1-C8-T1",
                '-s', 'PhysicalPort.sublabel=bestlabelever']
        cp = test_util.setup_mock(
            self, args, self.adpt, wrps=wrps, p_wrps=p_wrps)

        cp.obj_type.execute()
        self.assertEqual(wrps[0].phys_ports[0].sublabel, "bestlabelever")
        ew_update_mock.assert_called_once_with()
        ew_update_mock.reset_mock()

        # 2 set fields
        args = ['sriov', 'update', '--port-loc', "U78C7.001.RCH0004-P1-C8-T1",
                '-s', 'PhysicalPort.sublabel=newsub', 'label=howdy']
        cp = test_util.setup_mock(
            self, args, self.adpt, wrps=wrps, p_wrps=p_wrps)

        cp.obj_type.execute()
        self.assertEqual(wrps[0].phys_ports[0].sublabel, "newsub")
        self.assertEqual(wrps[0].phys_ports[0].label, "howdy")
        ew_update_mock.assert_called_once_with()
        ew_update_mock.reset_mock()

        # No loc, no physical port set field
        args = ['sriov', 'update', '-s', 'mode=Dedicated']
        cp = test_util.setup_mock(
            self, args, self.adpt, wrps=wrps, p_wrps=p_wrps)

        self.assertEqual(wrps[0].mode, "Sriov")
        cp.obj_type.execute()
        self.assertEqual(wrps[0].mode, "Dedicated")
        ew_update_mock.assert_called_once_with()
        ew_update_mock.reset_mock()

        # No loc, physical port set field
        args = ['sriov', 'update', '-s', 'mode=Sriov', 'label=woops']
        cp = test_util.setup_mock(
            self, args, self.adpt, wrps=wrps, p_wrps=p_wrps)

        port_err = self.assertRaises(
            exc.PPortUpdateRequiresLocError, cp.obj_type.execute)
        self.assertEquals(
            six.text_type(port_err),
            "--port-loc argument required to update a physical port.")
        self.assertEqual(wrps[0].mode, "Dedicated")
        ew_update_mock.assert_not_called()
        ew_update_mock.reset_mock()

        # loc filtering
        wrps = self.dwrap.asio_config.sriov_adapters[0:2]
        args = ['sriov', 'update', '--port-loc', "U78C7.001.RCH0004-P1-C8-T1",
                '-s', 'PhysicalPort.sublabel=filteringworks']
        cp = test_util.setup_mock(
            self, args, self.adpt, wrps=wrps, p_wrps=p_wrps)

        self.assertEqual(len(cp.obj_type.ec.wrappers), 2)
        cp.obj_type.execute()
        self.assertEqual(len(cp.obj_type.ec.wrappers), 1)
        self.assertEqual(wrps[0].phys_ports[0].sublabel, "filteringworks")
        ew_update_mock.assert_called_once_with()
        ew_update_mock.reset_mock()
