# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock

import pvmcli.cli_parser as cli_parser
import pvmcli.command.virtual_ethernet_adapter as vea_c
import pvmcli.command.operation as op
import pvmcli.help.help_virtual_ethernet_adapter as help_vea
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.network as n
import pypowervm.wrappers.logical_partition as l
import pypowervm.wrappers.virtual_io_server as v


LIST_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualEthernetAdapter "
                    "list {--parent-id | -p} <VirtualIOServer field>=value"
                    " {--object-id | -i} <ClientNetworkAdapter field>=value"
                    " {args}"])

UPDATE_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualEthernetAdapter "
                      "update {--parent-id | -p} <VirtualIOServer field>="
                      "value {--object-id | -i} <ClientNetworkAdapter "
                      "field>=value {args}"])

DELETE_USAGE_HELP = (["usage: pvmctl [help | --help] VirtualEthernetAdapter "
                      "delete {--parent-id | -p} <VirtualIOServer field>="
                      "value {--object-id | -i} <ClientNetworkAdapter "
                      "field>=value {args}"])

# TODO (IBM): Custom display field for vswitch_name
DISPLAY_FIELD_HELP = (
    ["    VirtualEthernetAdapter.configured_mtu\n"
     "    VirtualEthernetAdapter.enabled\n"
     "    VirtualEthernetAdapter.gateway\n"
     "    VirtualEthernetAdapter.ip_address\n"
     "    VirtualEthernetAdapter.is_tagged_vlan_supported\n"
     "    VirtualEthernetAdapter.is_trunk\n"
     "    VirtualEthernetAdapter.loc_code\n"
     "    VirtualEthernetAdapter.lpar_id\n"
     "    VirtualEthernetAdapter.mac\n"
     "    VirtualEthernetAdapter.ovs_bridge\n"
     "    VirtualEthernetAdapter.ovs_ext_ids\n"
     "    VirtualEthernetAdapter.pvid\n"
     "    VirtualEthernetAdapter.slot\n"
     "    VirtualEthernetAdapter.subnet_mask\n"
     "    VirtualEthernetAdapter.tagged_vlans\n"
     "    VirtualEthernetAdapter.trunk_pri\n"
     "    VirtualEthernetAdapter.uuid\n"
     "    VirtualEthernetAdapter.vsi_type_id\n"
     "    VirtualEthernetAdapter.vsi_type_manager_id\n"
     "    VirtualEthernetAdapter.vsi_type_version\n"
     "    VirtualEthernetAdapter.vswitch_id\n\n"])


SET_FIELD_HELP = (
    ["--set-fields options:\n"
     "    VirtualEthernetAdapter.is_tagged_vlan_supported\n"
     "    VirtualEthernetAdapter.pvid\n"
     "    VirtualEthernetAdapter.tagged_vlans\n"])


ALL_OUT_STR = (
    "--------------------------\n"
    "| VirtualEthernetAdapter |\n"
    "--------------------------\n"
    "  configured_mtu=1500\n"
    "  enabled=True\n"
    "  gateway=192.168.2.0\n"
    "  ip_address=192.168.2.6\n"
    "  is_tagged_vlan_supported=True\n"
    "  is_trunk=True\n"
    "  loc_code=U8246.L2C.0604C7A-V24-C32\n"
    "  lpar_id=3\n"
    "  mac=FAD4433ED120\n"
    "  ovs_bridge=br-int\n"
    "  ovs_ext_ids=iface-id=ba9d8ec3-64b2-47fe-9f50-e12ba373814c,"
    "iface-status=active,attached-mac=fa:e6:c8:3f:80:20,"
    "vm-uuid=64443c49-920d-47d7-9b78-1216845c51f5\n"
    "  pvid=100\n"
    "  slot=32\n"
    "  subnet_mask=255.255.255.0\n"
    "  tagged_vlans=[53, 54, 55]\n"
    "  trunk_pri=1\n"
    "  uuid=6445b54b-b9dc-3bc2-b1d3-f8cc22ba95b8\n"
    "  vsi_type_id=VSITID\n"
    "  vsi_type_manager_id=VSIMID\n"
    "  vsi_type_version=77.99\n"
    "  vswitch_id=0\n")


class VirtualEthernetAdapterTest(twa.TestWrapper):
    ''' Test framework for VirtualEthernetAdapter Object Type. '''
    @property
    def file(self):
        return 'fake_cna.txt'

    @property
    def wrapper_class_to_test(self):
        return n.CNA

    def test_virtual_ethernet_adapter(self):
        vea = vea_c.VirtualEthernetAdapter()

        test_util.test_obj_type(self, vea, 'VirtualEthernetAdapter',
                                n.CNA, field_models=[n.CNA],
                                name_to_ops={op.CREATE_NAME,
                                             op.LIST_NAME,
                                             op.UPDATE_NAME,
                                             op.DELETE_NAME},
                                exp_help=help_vea.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={vea.name_to_ops['list']: [l.LPAR, v.VIOS],
                                    vea.name_to_ops['update']: [l.LPAR,
                                                                v.VIOS],
                                    vea.name_to_ops['delete']: [l.LPAR,
                                                                v.VIOS],
                                    vea.name_to_ops['create']:
                                    [l.LPAR, v.VIOS]},
            child_model_ops={vea.name_to_ops['list']: None,
                             vea.name_to_ops['update']: None,
                             vea.name_to_ops['delete']: None,
                             vea.name_to_ops['create']: None},
            elem_wrp_func_ops={vea.name_to_ops['list']: None,
                               vea.name_to_ops['update']: None,
                               vea.name_to_ops['delete']: None,
                               vea.name_to_ops['create']: None})

    def test_overrides(self):
        ''' Test the overridden Field objects for VEA ObjectType '''
        vea = vea_c.VirtualEthernetAdapter()

        self.assertEqual(vea.get_schema(n.CNA.schema_type), vea_c.VEA_SCHEMA)
        self.assertEqual(
            vea.name_to_ops[op.UPDATE_NAME].op_args[op.SF_NAME].fields[
                'VirtualEthernetAdapter.tagged_vlans'].field_type, list)

    @mock.patch("pypowervm.wrappers.network.VSwitch.wrap")
    def test_list(self, vsw_wrap_mock):
        vsw_wrap_mock.return_value = n.VSwitch.bld(mock.Mock(), "test_vs")
        test_util.test_list(
            self, 'VirtualEthernetAdapter',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Virtual Ethernet Adapters\n"
             "+------+---------+------+--------------+"
             "------+-------+--------------+\n"
             "| PVID | VSwitch | LPAR |     MAC      |"
             " Slot | Trunk | Tagged VLANs |\n"
             "+------+---------+------+--------------+"
             "------+-------+--------------+\n"
             "| 100  | test_vs |  3   | FAD4433ED120 |"
             "  32  |  True |   53,54,55   |\n"
             "+------+---------+------+--------------+"
             "------+-------+--------------+\n"],
            [ALL_OUT_STR],
            ["<ClientNetworkAdapter:ClientNetworkAdapter xmlns:ClientN"],
            ['is_trunk', 'mac'],
            ["is_trunk=True,mac=FAD4433ED120\n"],
            ["True,FAD4433ED120\n"],
            ["is_trunk___True,mac___FAD4433ED120\n"],
            ["is_trunk=True___mac=FAD4433ED120\n"],
            ["is_trunk=True,mac=FAD4433ED120\n"],
            ["is_trunk___True___mac___FAD4433ED120\n"],
            {'child_type': 'ClientNetworkAdapter',
             'root_id': '014C1F78-B210-40B2-A12E-C82316B878CA',
             'root_type': 'VirtualIOServer'},
            p_wrps=[test_util.mock_vios(self.adpt)])

    def test_create(self):
        parent_models = [l.LPAR, v.VIOS]
        cp = test_util.setup_mock(self, args=['VirtualEthernetAdapter',
                                              'create', '--pvid', '1',
                                              '--vswitch', 'ETH0',
                                              '-p', 'name=lpar_name'],
                                  adpt_fx=self.adpt)

        self.assertEquals(cp.obj_type.ec.exec_op.parent_models,
                          parent_models)

        cp = test_util.setup_mock(self, args=['VirtualEthernetAdapter',
                                              'create', '--pvid', '1',
                                              '--vswitch', 'ETH0', '--slot',
                                              '2', '--mac', '123', '--vlans',
                                              '1', '2', '-p',
                                              'name=lpar_name'],
                                  adpt_fx=self.adpt)

        # Test trunk priority
        cp = test_util.setup_mock(self, args=['VirtualEthernetAdapter',
                                              'create', '--pvid', '1',
                                              '--vswitch', 'ETH0', '--slot',
                                              '2', '--mac', '123',
                                              '--vlans', '1', '2',
                                              '-p', 'name=lpar_name',
                                              '--trunk-pri', '2'],
                                  adpt_fx=self.adpt)

        # Test invalid trunk priority
        cp = cli_parser.CLIParser(args=['VirtualEthernetAdapter', 'create',
                                        '--pvid', '1', '--vswitch', 'ETH0',
                                        '--slot', '2', '--mac', '123',
                                        '--vlans', '1', '2',
                                        '-p', 'name=lpar_name',
                                        '--trunk-pri', 'abc'])
        se_exc = self.assertRaises(SystemExit, cp.process)
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # No pvid argument
        cp = cli_parser.CLIParser(args=['VirtualEthernetAdapter', 'create',
                                        '--vswitch', 'ETH0', '--slot', '2',
                                        '--mac', '123', '--vlans', '1', '2',
                                        '-p', 'name=lpar_name'])
        se_exc = self.assertRaises(SystemExit, cp.process)
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # No vswitch argument
        cp = cli_parser.CLIParser(args=['VirtualEthernetAdapter', 'create',
                                        '--pvid', '1', '--slot', '2',
                                        '--mac', '123', '--vlans', '1', '2',
                                        '-p', 'name=lpar_name'])
        se_exc = self.assertRaises(SystemExit, cp.process)
        self.assertEquals(se_exc.code, exc.ARGPARSE_ERR_RET_CODE)

        # Test trunk priority, no parent id
        self.assertRaises(SystemExit, test_util.setup_mock, self,
                          args=['VirtualEthernetAdapter',
                                'create', '--pvid', '1',
                                '--vswitch', 'ETH0', '--slot',
                                '2', '--mac', '123',
                                '--vlans', '1', '2',
                                '--trunk-pri', '2'], adpt_fx=self.adpt)

    def _update_mock_ret_value(self, cp):
        return cp.obj_type.ec.wrappers[0]

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.update')
    def test_update(self, ew_update_mock):
        ''' Test update operation on VirtualEthernetAdapter ObjectType. '''
        args = ['VirtualEthernetAdapter', 'update', '--set-fields']
        args.extend(['VirtualEthernetAdapter.tagged_vlans=',
                     'VirtualEthernetAdapter.is_tagged_vlan_supported=False'])
        cp = test_util.setup_mock(self, args, self.adpt, wrps=[self.dwrap])

        ew_update_mock.return_value = self._update_mock_ret_value(cp)
        self.assertEqual(
            cp.obj_type.ec.wrappers[0].tagged_vlans, [53, 54, 55])
        self.assertTrue(
            cp.obj_type.ec.wrappers[0].is_tagged_vlan_supported)

        cp.obj_type.execute()

        ew_update_mock.assert_called_once_with()
        self.assertEqual(
            cp.obj_type.ec.wrappers[0].tagged_vlans, [])
        self.assertFalse(
            cp.obj_type.ec.wrappers[0].is_tagged_vlan_supported)
        ew_update_mock.reset_mock()

        args = ['VirtualEthernetAdapter', 'update', '--set-fields']
        args.append('VirtualEthernetAdapter.tagged_vlans=1 2 3')
        cp = test_util.setup_mock(self, args, self.adpt, wrps=[self.dwrap])

        ew_update_mock.return_value = self._update_mock_ret_value(cp)
        self.assertEqual(
            cp.obj_type.ec.wrappers[0].tagged_vlans, [])
        self.assertFalse(
            cp.obj_type.ec.wrappers[0].is_tagged_vlan_supported)

        cp.obj_type.execute()

        ew_update_mock.assert_called_once_with()
        self.assertEqual(
            cp.obj_type.ec.wrappers[0].tagged_vlans, [1, 2, 3])
        self.assertTrue(
            cp.obj_type.ec.wrappers[0].is_tagged_vlan_supported)
        ew_update_mock.reset_mock()

    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.search')
    @mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.delete')
    def test_delete(self, ew_delete_mock, search_mock):
        ''' Test delete operation on VirtualEthernetAdapter ObjectType. '''
        search_mock.return_value = [self.dwrap]
        test_util.test_delete(self, ew_delete_mock, 'VirtualEthernetAdapter',
                              DELETE_USAGE_HELP, wrps=[self.dwrap],
                              p_wrps=[test_util.mock_vios(self.adpt)])

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'VirtualEthernetAdapter')
