# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock

import pvmcli.command.virtual_network as vnet_c
import pvmcli.help.help_virtual_network as help_vnet
import pvmcli.tests.util.test_util as test_util

import pypowervm.tests.test_utils.pvmhttp as ph
import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.managed_system as ms_w
import pypowervm.wrappers.network as n
import pypowervm.wrappers.virtual_io_server as v


LIST_USAGE_HELP = (
    ["usage: pvmctl [help | --help] VirtualNetwork list {--object-id | -i} "
     "<VirtualNetwork field>=value {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    VirtualNetwork.associated_switch_uri\n"
     "    VirtualNetwork.name\n"
     "    VirtualNetwork.tagged\n"
     "    VirtualNetwork.uuid\n"
     "    VirtualNetwork.vlan\n"
     "    VirtualNetwork.vswitch_id\n"])


class VirtualNetworkTest(twa.TestWrapper):
    ''' Test framework for VirtualNetwork Object Type. '''
    @property
    def file(self):
        return 'fake_virtual_network_feed.txt'

    @property
    def wrapper_class_to_test(self):
        return n.VNet

    def test_virtual_network(self):
        vnet = vnet_c.VirtualNetwork()

        test_util.test_obj_type(self, vnet, 'VirtualNetwork',
                                n.VNet, field_models=[n.VNet],
                                name_to_ops={'list'},
                                exp_help=help_vnet.help_text)

        # Test Operations
        test_util.test_ops(
            self, parent_model_ops={vnet.name_to_ops['list']: [ms_w.System]},
            child_model_ops={vnet.name_to_ops['list']: None},
            elem_wrp_func_ops={vnet.name_to_ops['list']: None})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'VirtualNetwork')

    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.wrap")
    @mock.patch("pypowervm.wrappers.network.VSwitch.wrap")
    def test_list(self, vsw_wrap_mock, vios_wrap_mock):
        vsw_wrap_mock.return_value = n.VSwitch.bld(mock.Mock(), "test_vs")
        vnet_2 = n.VNet.bld(mock.Mock(), 'test_vlan', 123, 'vnet_uri', True)

        ret_vios = v.VIOS.bld(
            mock.Mock(), "test_vios",
            bp.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        vios_wrap_mock.return_value = ret_vios

        nbr = self._setup_mock_nbr()[0]
        with mock.patch("pypowervm.wrappers.network.NetBridge.wrap",
                        return_value=nbr):
            test_util.test_list(
                self, 'VirtualNetwork',
                LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
                ["Virtual Networks\n"
                 "+------+---------+--------+-----------+----------+------+\n"
                 "| VLAN | VSwitch | Tagged |    VIOS   | Back Dev | SEA  |\n"
                 "+------+---------+--------+-----------+----------+------+\n"
                 "| 123  | test_vs |  True  |           |          |      |\n"
                 "| 2227 | test_vs | False  | test_vios |   ent0   | ent8 |\n"
                 "|      |         |        | test_vios |   ent0   | ent8 |\n"
                 "+------+---------+--------+-----------+----------+------+"
                 "\n"],
                ["------------------\n"
                 "| VirtualNetwork |\n"
                 "------------------\n"
                 "  associated_switch_uri=https://9.1.2.3:12443/rest/api/uom/"
                 "ManagedSystem/67dca605-3923-34da-bd8f-26a378fc817f/VirtualS"
                 "witch/ec8aaa54-9837-3c23-a541-a4e4be3ae489\n"
                 "  name=VLAN2227-ETHERNET0\n"
                 "  tagged=False\n"
                 "  uuid=20714b36-f66c-329f-b1c2-763c1c1a65a2\n"
                 "  vlan=2227\n"
                 "  vswitch_id=0\n"],
                ['<VirtualNetwork:VirtualNetwork xmlns:VirtualNetwork="'],
                ['VirtualNetwork.name', 'VirtualNetwork.vlan'],
                ["name=VLAN2227-ETHERNET0,vlan=2227\n"
                 "name=test_vlan,vlan=123\n"],
                ["VLAN2227-ETHERNET0,2227\ntest_vlan,123\n"],
                ["name___VLAN2227-ETHERNET0,vlan___2227\n"
                 "name___test_vlan,vlan___123\n"],
                ["name=VLAN2227-ETHERNET0___vlan=2227\n"
                 "name=test_vlan___vlan=123\n"],
                ["name=VLAN2227-ETHERNET0,vlan=2227___"
                 "name=test_vlan,vlan=123\n"],
                ["name___VLAN2227-ETHERNET0___vlan___2227___"
                 "name___test_vlan___vlan___123\n"],
                {'root_type': 'ManagedSystem',
                 'child_type': 'VirtualNetwork',
                 'root_id': 'caae9209-25e5-35cd-a71a-ed55c03f294d'},
                wrps=[self.dwrap, vnet_2],
                p_wrps=[test_util.mock_sys(self.adpt)])

    def _setup_mock_nbr(self):
        resp = ph.PVMResp(
            pvmfile=ph.PVMFile('fake_network_bridge.txt'),
            adapter=self.adpt).get_response()
        return n.NetBridge.wrap(resp)
