# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2016
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
import six

import pvmcli.command.virtual_optical_media as vom_c
import pvmcli.help.help_virtual_optical_media as help_vopt
import pvmcli.tests.util.test_util as test_util
import pvmcli.util.exceptions as exc

import pypowervm.tests.test_utils.test_wrapper_abc as twa
import pypowervm.wrappers.base_partition as bp_w
import pypowervm.wrappers.storage as stw
from pypowervm.wrappers import virtual_io_server as vios_w


LIST_USAGE_HELP = (
    ["usage: pvmctl [help | --help] VirtualOpticalMedia list {--parent-id | "
     "-p} <VirtualIOServer field>=value {--child-id | -c} <VolumeGroup "
     "field>=value {--object-id | -i} <VirtualOpticalMedia field>=value"
     " {args}"])

DISPLAY_FIELD_HELP = (
    ["--display-fields options:\n"
     "    VirtualOpticalMedia.media_name\n"
     "    VirtualOpticalMedia.mount_type\n"
     "    VirtualOpticalMedia.name\n"
     "    VirtualOpticalMedia.size\n"
     "    VirtualOpticalMedia.udid\n"])


class VirtualOpticalMediaTest(twa.TestWrapper):
    ''' Test framework for VirtualOpticalMedia Object Type. '''
    @property
    def file(self):
        return 'fake_volume_group.txt'

    @property
    def wrapper_class_to_test(self):
        return stw.VMediaRepos

    def test_virtual_optical_media(self):
        vom = vom_c.VirtualOpticalMedia()

        test_util.test_obj_type(self, vom, 'VirtualOpticalMedia',
                                stw.VOptMedia,
                                field_models=[stw.VOptMedia],
                                name_to_ops={'list', 'upload', 'delete'},
                                exp_help=help_vopt.help_text)

        # Test Operations
        opt_m = stw.VMediaRepos.optical_media
        test_util.test_ops(
            self, parent_model_ops={vom.name_to_ops['list']: [vios_w.VIOS],
                                    vom.name_to_ops['upload']: None,
                                    vom.name_to_ops['delete']: [vios_w.VIOS]},
            child_model_ops={vom.name_to_ops['list']: [stw.VG],
                             vom.name_to_ops['upload']: None,
                             vom.name_to_ops['delete']: [stw.VG]},
            elem_wrp_func_ops={vom.name_to_ops['list']: [stw.VG.vmedia_repos,
                                                         opt_m],
                               vom.name_to_ops['upload']: None,
                               vom.name_to_ops['delete']: [stw.VG.vmedia_repos,
                                                           opt_m]})

    def test_help(self):
        ''' Test help arguments for an object type. '''
        test_util.test_help(self, 'VirtualOpticalMedia')

    @mock.patch("pypowervm.wrappers.virtual_io_server.VIOS.search")
    def test_list(self, vios_search_mock):
        sample_vios_w = test_util.mock_vios(self.adpt)
        vios_search_mock.return_value = [sample_vios_w]
        vg_wrap = stw.VG.wrap(self.resp)
        test_util.test_list(
            self, 'VirtualOpticalMedia',
            LIST_USAGE_HELP + test_util.EXPECTED_HELP + DISPLAY_FIELD_HELP,
            ["Virtual Optical Media\n"
             "+-------------+------------+---------------+-----------+"
             "---------------+--------------+----+\n"
             "|     VIOS    |     VG     | VMR Size (GB) | Free (GB) |"
             "    OM Name    | OM Size (GB) | rw |\n"
             "+-------------+------------+---------------+-----------+"
             "---------------+--------------+----+\n"
             "| vios_1_Drew | image_pool |      11.0     |  10.8535  |"
             "  blank_media1 |    0.0977    | rw |\n"
             "|             |            |               |           |"
             " blank_media_2 |    0.0488    | rw |\n"
             "+-------------+------------+---------------+-----------+"
             "---------------+--------------+----+\n"],
            ["-----------------------\n"
             "| VirtualOpticalMedia |\n"
             "-----------------------\n"
             "  media_name=blank_media1\n"
             "  mount_type=rw\n"
             "  name=blank_media1\n"
             "  size=0.0977\n"
             "  udid=0eblank_media1\n"],
            ['<VirtualOpticalMedia xmlns="http://www.ibm.com/xmlns/systems'],
            ['VirtualOpticalMedia.name', 'VirtualOpticalMedia.mount_type'],
            ["name=blank_media1,mount_type=rw\n"],
            ["blank_media1,rw\n"],
            ["name___blank_media1,mount_type___rw\n"],
            ["name=blank_media1___mount_type=rw\n"],
            ["name=blank_media1,mount_type=rw\n"],
            ["name___blank_media1___mount_type___rw\n"],
            {'root_type': 'VirtualIOServer', 'child_type': 'VolumeGroup',
             'root_id': '014C1F78-B210-40B2-A12E-C82316B878CA'},
            c_wrps=[vg_wrap], wrps=[vg_wrap.vmedia_repos[0].optical_media[0]],
            p_wrps=[sample_vios_w])

    @mock.patch('pypowervm.wrappers.virtual_io_server.VIOS.search')
    @mock.patch('pypowervm.tasks.storage.upload_vopt')
    def test_upload(self, vopt_mock, vios_search_mock):
        ret_vios = vios_w.VIOS.bld(
            mock.Mock(), "test_vios",
            bp_w.PartitionMemoryConfiguration.bld(mock.Mock(), 1024),
            bp_w.PartitionProcessorConfiguration.bld_dedicated(mock.Mock(), 1))
        ret_vios.set_parm_value(bp_w._BP_STATE, bp_w.LPARState.RUNNING)
        ret_vios.set_parm_value(bp_w._BP_RMC_STATE, bp_w.RMCState.ACTIVE)
        vios_search_mock.return_value = [ret_vios]

        # Minimum args
        cp = test_util.setup_mock(self, args=['vom', 'upload', '--vios',
                                              'test_vios', '--name', 'vo_name',
                                              '--data', 'vo_data'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])
        q_w_str = 'pvmcli.command.object_type.ExecutionContext.query_wrappers'
        with mock.patch(test_util.mock_open_str()) as mock_open:
            with mock.patch('os.path.getsize') as mock_getsize:
                with mock.patch(q_w_str) as q_wrps_mock:
                    cp.obj_type.execute()
                    vopt_mock.assert_called_once_with(
                        cp.obj_type.ec.sm.adapter, None,
                        mock_open().__enter__(), 'vo_name',
                        f_size=mock_getsize(), sha_chksum=None)
                    q_wrps_mock.asssert_not_called()
        vopt_mock.reset_mock()

        # All args
        cp = test_util.setup_mock(self, args=['vom', 'upload', '--vios',
                                              'test_vios', '--name', 'vo_name',
                                              '--data', 'vo_data',
                                              '--checksum', 'abc'],
                                  adpt_fx=self.adpt, wrps=[self.dwrap])

        with mock.patch(test_util.mock_open_str()) as mock_open:
            with mock.patch('os.path.getsize') as mock_getsize:
                cp.obj_type.execute()
                vopt_mock.assert_called_once_with(
                    cp.obj_type.ec.sm.adapter, None, mock_open().__enter__(),
                    'vo_name', f_size=mock_getsize(), sha_chksum='abc')
        vopt_mock.reset_mock()

        # No args
        self.assertRaises(SystemExit, test_util.setup_mock, self,
                          args=['vom', 'upload'], adpt_fx=self.adpt,
                          wrps=[self.dwrap])
        vopt_mock.assert_not_called()

    @mock.patch('pypowervm.wrappers.storage.VG.search')
    @mock.patch('pypowervm.wrappers.storage.VG.update')
    def test_delete(self, vg_upd_mock, vg_srch_mock):
        help_out = six.StringIO()
        test_util.setup_and_exec(
            self, ['vom', 'delete', '--help'], help_out=help_out)
        self.assertIn("usage: pvmctl [help | --help] VirtualOpticalMedia "
                      "delete {--parent-id | -p} <VirtualIOServer "
                      "field>=value {--child-id | -c} <VolumeGroup "
                      "field>=value {--object-id | -i} <VirtualOpticalMedia "
                      "field>=value {args}", help_out.getvalue())

        vg_wrap = stw.VG.wrap(self.resp)
        om_wrp = vg_wrap.vmedia_repos[0].optical_media[0]
        # No object-id (required, argparse catches with SystemExit)
        self.assertRaises(SystemExit, test_util.setup_mock,
                          self, args=['vom', 'delete', '-c', 'name=rootvg'],
                          adpt_fx=self.adpt, c_wrps=[vg_wrap], wrps=[om_wrp])
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # No child-id (required, argparse catches with SystemExit)
        self.assertRaises(SystemExit, test_util.setup_mock,
                          self, args=['vom', 'delete', '-i',
                                      'name=blank_media1'],
                          adpt_fx=self.adpt, c_wrps=[vg_wrap], wrps=[om_wrp])
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Multiple children
        cp = test_util.setup_mock(self, args=['vom', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=blank_media1'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap, vg_wrap],
                                  wrps=[om_wrp])
        self.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Multiple wrappers
        cp = test_util.setup_mock(self, args=['vom', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=blank_media1'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap],
                                  wrps=[om_wrp, om_wrp])
        self.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)
        vg_srch_mock.assert_not_called()
        vg_upd_mock.assert_not_called()

        # Minimum args
        sample_vios_w = test_util.mock_vios(self.adpt)
        cp = test_util.setup_mock(self, args=['vom', 'delete', '-c',
                                              'name=rootvg', '-i',
                                              'name=blank_media1'],
                                  adpt_fx=self.adpt, c_wrps=[vg_wrap],
                                  wrps=[om_wrp], p_wrps=[sample_vios_w])
        self.assertEqual(vg_wrap.vmedia_repos[0].optical_media[0].name,
                         'blank_media1')
        self.assertEqual(len(vg_wrap.vmedia_repos), 1)
        self.assertEqual(len(vg_wrap.vmedia_repos[0].optical_media), 2)
        cp.obj_type.execute()
        self.assertEqual(len(vg_wrap.vmedia_repos[0].optical_media), 1)
        vg_srch_mock.assert_called_with(
            cp.obj_type.ec.sm.adapter,
            parent_uuid='014C1F78-B210-40B2-A12E-C82316B878CA',
            parent_type='VirtualIOServer', xag=[], name='rootvg')
        vg_upd_mock.assert_called_once_with()
        vg_upd_mock.reset_mock()
