# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2016, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import mock
from mock import Mock
import six
import sys

import pvmcli.cli_parser as cli_parser
import pvmcli.util.exceptions as exc
import pvmcli.util.util as util

from pypowervm import adapter as adpt
from pypowervm import const as c
from pypowervm import exceptions as pvm_exc
import pypowervm.tasks.power as pwr
from pypowervm.tests.test_utils import pvmhttp
import pypowervm.wrappers.iocard as iocard
import pypowervm.wrappers.logical_partition as lpar_w
import pypowervm.wrappers.managed_system as msys_w
import pypowervm.wrappers.virtual_io_server as vios_w


HELP_USAGE = "usage: pvmctl [help | --help] %(name)s {operation [args]}\n\n"

EXPECTED_HELP = (
    ["--display-fields option [option ...], -d option [option ...]",
     "--object-id id=value, -i id=value",
     "--hide-label",
     "--label-char LABEL_CHAR, -l LABEL_CHAR",
     "--where WHERE [WHERE ...], -w WHERE [WHERE ...]",
     "--xml                 Display raw response XML for all query responses.",
     "--field-delim FIELD_DELIM, -f FIELD_DELIM",
     "--record-delim RECORD_DELIM, -r RECORD_DELIM",
     "--object-id id=value, -i id=value",
     "--help                show this help message and exit",
     "help                  show this help message and exit",
     "--display-fields options:"])

LPAR_OBJ_NAME = 'lpar'
VIOS_OBJ_NAME = 'vios'


def mock_open_str():
    return '__builtin__.open' if six.PY2 else 'builtins.open'


def test_help(test_obj, obj_name):
    ''' Test --help argument for an object type. '''
    # pvmctl <obj_name>
    obj_help_out = six.StringIO()
    cp = cli_parser.CLIParser([obj_name],
                              help_out=obj_help_out)
    test_obj.assertEqual(cp.process(), exc.OK_RETURN_CODE)
    test_obj.assertIsNotNone(obj_help_out)
    test_obj.assertIsNotNone(obj_help_out.getvalue())

    # pvmctl <obj_name> --help
    dash_help_out = six.StringIO()
    cp = cli_parser.CLIParser([obj_name, '--help'],
                              help_out=dash_help_out)
    test_obj.assertEqual(cp.process(), exc.OK_RETURN_CODE)
    test_obj.assertIsNotNone(dash_help_out)
    test_obj.assertIsNotNone(dash_help_out.getvalue())

    # pvmctl <obj_name> help
    help_out = six.StringIO()
    cp = cli_parser.CLIParser([obj_name, 'help'],
                              help_out=help_out)
    test_obj.assertEqual(cp.process(), exc.OK_RETURN_CODE)
    test_obj.assertIsNotNone(help_out)
    test_obj.assertIsNotNone(help_out.getvalue())

    # pvmctl help <obj_name>
    pre_help_out = six.StringIO()
    cp = cli_parser.CLIParser(['help', obj_name],
                              help_out=pre_help_out)
    test_obj.assertEqual(cp.process(), exc.OK_RETURN_CODE)
    test_obj.assertIsNotNone(pre_help_out)
    test_obj.assertIsNotNone(pre_help_out.getvalue())

    # pvmctl --help <obj_name>
    pre_dash_help_out = six.StringIO()
    cp = cli_parser.CLIParser(['--help', obj_name],
                              help_out=pre_dash_help_out)
    test_obj.assertEqual(cp.process(), exc.OK_RETURN_CODE)
    test_obj.assertIsNotNone(pre_dash_help_out)
    test_obj.assertIsNotNone(pre_dash_help_out.getvalue())

    test_obj.assertEqual(obj_help_out.getvalue(), dash_help_out.getvalue())
    test_obj.assertEqual(dash_help_out.getvalue(), help_out.getvalue())
    test_obj.assertEqual(help_out.getvalue(), pre_help_out.getvalue())
    test_obj.assertEqual(pre_help_out.getvalue(),
                         pre_dash_help_out.getvalue())


def test_obj_type(test_obj, obj_type, name, model, field_models=[],
                  name_to_ops=[], schema_remap={}, exp_help=HELP_USAGE):
    ''' Test display models and registered operations for an ObjectType. '''
    test_obj.assertEqual(obj_type.model, model)
    test_obj.assertEqual(obj_type.name, name)
    test_obj.assertEqual(obj_type.field_models, field_models)
    test_obj.assertEqual(set(obj_type.name_to_ops.keys()), name_to_ops)
    for sch in schema_remap:
        test_obj.assertEqual(obj_type.get_schema(sch), schema_remap(sch))

    # Test output
    cp_no_arg_out = six.StringIO()
    setup_mock(test_obj, [obj_type.name], test_obj.adpt,
               wrps=[test_obj.dwrap], help_out=cp_no_arg_out)
    test_obj.assertIn(exp_help % {'name': obj_type.name},
                      cp_no_arg_out.getvalue())

    cp_help_out = six.StringIO()
    setup_mock(test_obj, [obj_type.name, 'help'], test_obj.adpt,
               wrps=[test_obj.dwrap], help_out=cp_help_out)
    test_obj.assertIn(exp_help % {'name': obj_type.name},
                      cp_help_out.getvalue())

    cp_dashhelp_out = six.StringIO()
    setup_mock(test_obj, [obj_type.name, '--help'], test_obj.adpt,
               wrps=[test_obj.dwrap], help_out=cp_dashhelp_out)
    test_obj.assertIn(exp_help % {'name': obj_type.name},
                      cp_dashhelp_out.getvalue())

    test_obj.assertEqual(cp_no_arg_out.getvalue(), cp_help_out.getvalue())
    test_obj.assertEqual(cp_help_out.getvalue(), cp_dashhelp_out.getvalue())


def test_ops(test_obj, parent_model_ops, child_model_ops, elem_wrp_func_ops):
    ''' Test operations on an object type. '''
    for parent_model_op in parent_model_ops:
        test_obj.assertEqual(parent_model_op.parent_models,
                             parent_model_ops[parent_model_op])
    for child_model_op in child_model_ops:
        test_obj.assertEqual(child_model_op.child_models,
                             child_model_ops[child_model_op])
    for elem_wrp_func_op in elem_wrp_func_ops:
        test_obj.assertEqual(elem_wrp_func_op.elem_wrp_funcs,
                             elem_wrp_func_ops[elem_wrp_func_op])


def setup_and_exec(test_obj, op_args, wrps=[], p_wrps=[], c_wrps=[],
                   help_out=sys.stdout, msys_mock_file='managedsystem.txt'):
    cp = setup_mock(test_obj, op_args, test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                    c_wrps=c_wrps, help_out=help_out,
                    msys_mock_file=msys_mock_file)

    test_obj.assertEqual(cp.obj_type.cli_err.getvalue(), '')
    test_obj.assertEqual(cp.obj_type.cli_out.getvalue(), '')

    if cp.obj_type.parsed_args:
        cp.obj_type.execute()
        cp.obj_type.sess_man.create_session_adapter.assert_called_once_with()

    return cp


def test_list(test_obj, obj_name, expected_help, default_strs, all_out_strs,
              xml_strs, df_args, df_strs, hide_label_strs,
              label_char_strs, field_delim_strs, record_delim_strs,
              combo_arg_strs, expected_read_args, wrps=None, p_wrps=[],
              c_wrps=[], msys_mock_file='managedsystem.txt'):
    if not wrps:
        wrps = [test_obj.dwrap]

    # pvmctl <obj_name> list help
    help_out = six.StringIO()
    setup_and_exec(test_obj, [obj_name, 'list', 'help'], wrps,
                   p_wrps, c_wrps, help_out=help_out,
                   msys_mock_file=msys_mock_file)
    for help_str in expected_help:
        test_obj.assertIn(help_str, help_out.getvalue())

    # pvmctl <obj_name> list
    _test_list_valid(test_obj, obj_name, [obj_name, 'list'], default_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --all-out
    _test_list_valid(test_obj, obj_name, [obj_name, 'list', '--all-out'],
                     all_out_strs, expected_read_args,
                     wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --xml
    _test_list_valid(test_obj, obj_name, [obj_name, 'list', '--xml'], xml_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --display-fields <df_args>
    specified_args = [obj_name, 'list', '--display-fields']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, df_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list -d <df_args> --hide-label
    specified_args = [obj_name, 'list', '--hide-label', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, hide_label_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --hide-label --label-char ___ -d <df_args>
    specified_args = [obj_name, 'list', '--hide-label', '--label-char',
                      '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, hide_label_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --label-char ___ -d <df_args>
    specified_args = [obj_name, 'list', '--label-char', '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, label_char_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --field-delim ___ -d <df_args>
    specified_args = [obj_name, 'list', '--field-delim', '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, field_delim_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --record-delim ___ -d <df_args>
    specified_args = [obj_name, 'list', '--record-delim', '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, record_delim_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # pvmctl <obj_name> list --label-char ___ --field-delim ___ \
    # --record-delim ___ -d <df_args>
    specified_args = [obj_name, 'list', '--label-char', '___', '--field-delim',
                      '___', '--record-delim', '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, combo_arg_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # Test short arg names
    # pvmctl <obj_name> list -l ___ -f ___ -r ___ -d <df_args>
    specified_args = [obj_name, 'list', '-l', '___', '-f',
                      '___', '-r', '___', '-d']
    specified_args.extend(df_args)
    _test_list_valid(test_obj, obj_name, specified_args, combo_arg_strs,
                     expected_read_args, wrp_list=[wrps, p_wrps, c_wrps],
                     msys_mock_file=msys_mock_file)

    # Test invalid arguments
    # label-char no arg
    # pvmctl <obj_name> list --label-char
    _test_list_invalid(test_obj, [obj_name, 'list', '--label-char'], wrps,
                       p_wrps, c_wrps)

    # field-delim no arg
    # pvmctl <obj_name> list --field-delim
    _test_list_invalid(test_obj, [obj_name, 'list', '--field-delim'], wrps,
                       p_wrps, c_wrps)

    # record-delim no arg
    # pvmctl <obj_name> list --record-delim
    _test_list_invalid(test_obj, [obj_name, 'list', '--record-delim'], wrps,
                       p_wrps, c_wrps)

    # display-field no arg
    # pvmctl <obj_name> list --display-fields
    _test_list_invalid(test_obj, [obj_name, 'list', '--display-fields'], wrps,
                       p_wrps, c_wrps)


def _test_list_valid(test_obj, obj_name, spec_args, expected_strs,
                     expected_read_args, wrp_list, msys_mock_file):
    wrps = wrp_list[0]
    p_wrps = wrp_list[1]
    c_wrps = wrp_list[2]

    spec_args.extend(['--max-width', '200'])
    cp = setup_and_exec(test_obj, spec_args, wrps, p_wrps, c_wrps,
                        msys_mock_file=msys_mock_file)

    test_obj.assertEqual(cp.obj_type.cli_err.getvalue(), '')
    for expected_str in expected_strs:
        test_obj.assertIn(expected_str, cp.obj_type.cli_out.getvalue())
    if isinstance(expected_read_args, dict):
        cp.obj_type.ec.sm.adapter.read.assert_any_call(xag=[],
                                                       **expected_read_args)
    else:
        cp.obj_type.ec.sm.adapter.read.assert_any_call(expected_read_args,
                                                       xag=[])


def _test_list_invalid(test_obj, spec_args, wrps, p_wrps=[], c_wrps=[]):
    test_obj.assertRaises(SystemExit, setup_and_exec, test_obj, spec_args,
                          wrps, sys.stdout)


def _test_up_exc(update_mock):
    resp = adpt.Response('reqmethod', 'reqpath',
                         c.HTTPStatus.ETAG_MISMATCH, 'reason',
                         'headers', None)
    return pvm_exc.HttpError(resp)


def _test_up_no_exc(update_mock):
    return update_mock


def test_update(test_obj, ew_update_mock, obj_name, set_fields,
                invalid_set_field, expected_help_strs, expected_err_str):
    wrps = [test_obj.dwrap]
    # pvmctl <obj_name> update help
    help_out = six.StringIO()
    setup_and_exec(test_obj, [obj_name, 'update', 'help'], wrps,
                   help_out=help_out)
    for help_str in expected_help_strs:
        test_obj.assertIn(help_str, help_out.getvalue())
    ew_update_mock.assert_not_called()
    ew_update_mock.reset_mock()

    # pvmctl <obj_name> update --set-fields <set_fields>[0]
    args = [obj_name, 'update', '--set-fields']
    args.append(set_fields[0])
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=[test_obj.dwrap])

    cp.obj_type.execute()

    ew_update_mock.assert_called_once_with()
    ew_update_mock.reset_mock()

    # pvmctl <obj_name> update -s <set_fields>
    args = [obj_name, 'update', '-s']
    args.extend(set_fields)
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=[test_obj.dwrap])

    cp.obj_type.execute()

    ew_update_mock.assert_called_once_with()
    ew_update_mock.reset_mock()

    # Invalid args
    # pvmctl <obj_name> update (requires --set-fields)
    test_obj.assertRaises(SystemExit, setup_and_exec, test_obj,
                          [obj_name, 'update'], wrps)

    ew_update_mock.assert_not_called()
    ew_update_mock.reset_mock()

    with mock.patch('sys.stderr', new_callable=six.StringIO) as err_mock:
        # pvmctl <obj_name> update -s <invalid_set_field>
        args.append(invalid_set_field)
        cp = setup_mock(
            test_obj, [obj_name, 'update', '--set-fields', invalid_set_field],
            test_obj.adpt, wrps=[test_obj.dwrap], proc_ret_code=1)
        test_obj.assertIn(expected_err_str, err_mock.getvalue())
        ew_update_mock.assert_not_called()
        ew_update_mock.reset_mock()

    # Try update with 0 wrappers
    args = [obj_name, 'update', '-s']
    args.extend(set_fields)
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=[test_obj.dwrap])

    cp.obj_type.ec._set_wrappers = Mock(side_effect=_set_wrappers(cp))

    test_obj.assertRaises(exc.NoEntriesFoundError, cp.obj_type.execute)

    ew_update_mock.assert_not_called()
    ew_update_mock.reset_mock()

    # Try update with 2 wrappers
    args = [obj_name, 'update', '-s']
    args.extend(set_fields)
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=[test_obj.dwrap])

    cp.obj_type.ec._set_wrappers = Mock(side_effect=_set_wrappers(
        cp, wrps=[test_obj.dwrap, test_obj.dwrap]))

    test_obj.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)

    ew_update_mock.assert_not_called()
    ew_update_mock.reset_mock()

    # Test retry: pvmctl <obj_name> update -s <set_fields>
    with mock.patch('pypowervm.wrappers.entry_wrapper.EntryWrapper.refresh',
                    return_value=test_obj.dwrap) as ref_mock:
        ew_update_mock.side_effect = [up_fn(ew_update_mock) for up_fn in [
            _test_up_exc, _test_up_exc, _test_up_no_exc]]
        args = ['ManagedSystem', 'update', '-s']
        args.extend(["number.model=abcd", "machine_type=xyz",
                     "ManagedSystem.max_vcpus_per_aix_linux_lpar=4"])
        cp = setup_mock(test_obj, args, test_obj.adpt, wrps=[test_obj.dwrap])
        cp.obj_type.execute()
        test_obj.assertEqual(ew_update_mock.call_count, 3)
        test_obj.assertEqual(ref_mock.call_count, 2)


def test_power_on(test_obj, power_on_mock, obj_name):
    cp = setup_mock(test_obj, [obj_name, 'power-on'], test_obj.adpt,
                    [test_obj.dwrap])
    cp.obj_type.execute()
    power_on_mock.assert_called_once_with(test_obj.dwrap,
                                          cp.obj_type.ec.msys.uuid,
                                          add_parms={})
    power_on_mock.reset_mock()

    cp = setup_mock(test_obj, [obj_name, 'power-on', '--bootstring', 'abcd'],
                    test_obj.adpt, [test_obj.dwrap])
    cp.obj_type.execute()
    power_on_mock.assert_called_once_with(
        test_obj.dwrap, cp.obj_type.ec.msys.uuid,
        add_parms={'bootstring': 'abcd'})
    power_on_mock.reset_mock()

    for b_opt in pwr.BootMode.ALL_VALUES:
        cp = setup_mock(test_obj, [obj_name, 'power-on', '--bootmode', b_opt],
                        test_obj.adpt, [test_obj.dwrap])
        cp.obj_type.execute()
        power_on_mock.assert_called_once_with(
            test_obj.dwrap, cp.obj_type.ec.msys.uuid,
            add_parms={pwr.BootMode.KEY: b_opt})
        power_on_mock.reset_mock()

        cp = setup_mock(test_obj, [obj_name, 'power-on',
                                   '--bootstring', 'abcd',
                                   '--bootmode', b_opt], test_obj.adpt,
                        [test_obj.dwrap])
        cp.obj_type.execute()
        power_on_mock.assert_called_once_with(
            test_obj.dwrap, cp.obj_type.ec.msys.uuid,
            add_parms={pwr.BootMode.KEY: b_opt, 'bootstring': 'abcd'})
        power_on_mock.reset_mock()

    for b_opt in pwr.KeylockPos.ALL_VALUES:
        cp = setup_mock(test_obj, [obj_name, 'power-on', '--keylock', b_opt],
                        test_obj.adpt, [test_obj.dwrap])
        cp.obj_type.execute()
        power_on_mock.assert_called_once_with(
            test_obj.dwrap, cp.obj_type.ec.msys.uuid,
            add_parms={pwr.KeylockPos.KEY: b_opt})
        power_on_mock.reset_mock()

        cp = setup_mock(test_obj, [obj_name, 'power-on',
                                   '--bootstring', 'abcd',
                                   '--keylock', b_opt], test_obj.adpt,
                        [test_obj.dwrap])
        cp.obj_type.execute()
        power_on_mock.assert_called_once_with(
            test_obj.dwrap, cp.obj_type.ec.msys.uuid,
            add_parms={pwr.KeylockPos.KEY: b_opt, 'bootstring': 'abcd'})
        power_on_mock.reset_mock()

    test_obj.assertRaises(SystemExit, setup_mock, test_obj,
                          [obj_name, 'power-on', '--bootstring'],
                          test_obj.adpt)

    test_obj.assertRaises(SystemExit, setup_mock, test_obj,
                          [obj_name, 'power-on', '--bootmode'],
                          test_obj.adpt)

    test_obj.assertRaises(SystemExit, setup_mock, test_obj,
                          [obj_name, 'power-on', '--bootmode' '--bootstring'],
                          test_obj.adpt)

    test_obj.assertRaises(SystemExit, setup_mock, test_obj,
                          [obj_name, 'power-on', '--bootmode', 'abcd'],
                          test_obj.adpt)

    cp = setup_mock(test_obj, [obj_name, 'power-on'], test_obj.adpt, [])
    test_obj.assertRaises(exc.NoEntriesFoundError, cp.obj_type.execute)

    cp = setup_mock(test_obj, [obj_name, 'power-on'], test_obj.adpt,
                    [test_obj.dwrap, test_obj.dwrap])
    test_obj.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)


def validate_input_called(obj_name, input_mock, op_name):
    if obj_name == LPAR_OBJ_NAME and op_name == 'power-off':
        input_mock.assert_called_once_with(
            "Powering off the management partition will leave you without "
            "control of your system. You will need to restart the entire "
            "system to get back control. Are you sure you wish to power off "
            "the management partition (Y/N)?")
    elif obj_name == VIOS_OBJ_NAME and op_name == 'power-off':
        input_mock.assert_called_once_with(
            "Powering off the VIOS will leave running partitions without "
            "access to their storage and/or network unless you have a "
            "redundant VIOS that is up and running. Are you sure you wish to "
            "power off the VIOS (Y/N)?")
    if obj_name == LPAR_OBJ_NAME and op_name in ['restart', 'dumprestart']:
        input_mock.assert_called_once_with("Are you sure you wish to restart "
                                           "the management partition "
                                           "(Y/N)?")
    elif obj_name == VIOS_OBJ_NAME and op_name == 'restart':
        input_mock.assert_called_once_with(
            "Restarting the VIOS will leave running partitions without "
            "access to their storage and/or network unless you have a "
            "redundant VIOS that is up and running. Are you sure you wish to "
            "restart the VIOS (Y/N)?")
    input_mock.reset_mock()


@mock.patch("pvmcli.util.util.input", return_value="Y")
def test_power_off(test_obj, input_mock, power_off_mock, obj_name):
    p_opt_o = 'pypowervm.tasks.power_opts.PowerOffOpts.operation'
    p_opt_i = 'pypowervm.tasks.power_opts.PowerOffOpts.vsp_hard'
    p_opt_r = 'pypowervm.tasks.power_opts.PowerOffOpts.restart'
    with mock.patch(p_opt_o) as ptc_o, mock.patch(p_opt_i) as ptc_i,\
            mock.patch(p_opt_r) as ptc_r:
        for op_name in ['power-off', 'restart']:
            cp = setup_mock(test_obj, [obj_name, op_name],
                            test_obj.adpt, [test_obj.dwrap])
            cp.obj_type.execute()
            ptc_o.assert_not_called()
            ptc_i.assert_not_called()
            if op_name == 'power-off':
                ptc_r.assert_not_called()
            elif op_name == 'restart':
                ptc_r.assert_called_once_with()
                ptc_r.reset_mock()
            power_off_mock.assert_called()
            validate_input_called(obj_name, input_mock, op_name)
            power_off_mock.reset_mock()

            cp = setup_mock(test_obj,
                            [obj_name, op_name, '--no-prompt'],
                            test_obj.adpt, [test_obj.dwrap])
            cp.obj_type.execute()
            ptc_o.assert_not_called()
            ptc_i.assert_not_called()
            if op_name == 'power-off':
                ptc_r.assert_not_called()
            elif op_name == 'restart':
                ptc_r.assert_called_once_with()
                ptc_r.reset_mock()
            power_off_mock.assert_called()
            input_mock.assert_not_called()
            power_off_mock.reset_mock()

        cp = setup_mock(test_obj, [obj_name, 'power-off', '--hard'],
                        test_obj.adpt, [test_obj.dwrap])
        cp.obj_type.execute()
        ptc_o.assert_not_called()
        ptc_i.assert_called_once_with()
        ptc_r.assert_not_called()
        ptc_i.reset_mock()
        power_off_mock.assert_called()
        validate_input_called(obj_name, input_mock, 'power-off')
        power_off_mock.reset_mock()

        cp = setup_mock(test_obj, [obj_name, 'power-off', '--restart'],
                        test_obj.adpt, [test_obj.dwrap])
        cp.obj_type.execute()
        ptc_o.assert_not_called()
        ptc_i.assert_not_called()
        ptc_r.assert_called_once_with()
        power_off_mock.assert_called()
        validate_input_called(obj_name, input_mock, 'power-off')
        ptc_r.reset_mock()
        power_off_mock.reset_mock()

        for op_name in ['power-off', 'restart']:
            # Test lower case 'y'
            input_mock.return_value = 'y'
            cp = setup_mock(test_obj, [obj_name, op_name], test_obj.adpt,
                            [test_obj.dwrap])
            cp.obj_type.execute()
            power_off_mock.assert_called()
            power_off_mock.reset_mock()

            cp = setup_mock(test_obj, [obj_name, op_name], test_obj.adpt,
                            [])
            test_obj.assertRaises(exc.NoEntriesFoundError,
                                  cp.obj_type.execute)

            cp = setup_mock(test_obj, [obj_name, op_name], test_obj.adpt,
                            [test_obj.dwrap, test_obj.dwrap])
            test_obj.assertRaises(exc.MultipleEntriesError,
                                  cp.obj_type.execute)

            input_mock.return_value = "no"
            cp = setup_mock(test_obj, [obj_name, op_name], test_obj.adpt,
                            [test_obj.dwrap])
            cp.obj_type.execute()
            power_off_mock.assert_not_called()


def test_delete(test_obj, ew_delete_mock, obj_name, expected_help_strs,
                wrps=None, p_wrps=None, c_wrps=None):
    # pvmctl <obj_name> delete help
    help_out = six.StringIO()
    setup_and_exec(test_obj, [obj_name, 'delete', 'help'], wrps=wrps,
                   p_wrps=p_wrps, c_wrps=c_wrps, help_out=help_out)
    for help_str in expected_help_strs:
        test_obj.assertIn(help_str, help_out.getvalue())
    ew_delete_mock.assert_not_called()
    ew_delete_mock.reset_mock()

    # By default we are guaranteed to have 1 wrapper to delete
    # pvmctl <obj_name> delete -i uuid=abcd
    args = [obj_name, 'delete', '-i', 'uuid=abcd']
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                    c_wrps=c_wrps)
    cp.obj_type.execute()
    ew_delete_mock.assert_called_once_with()
    ew_delete_mock.reset_mock()

    # Try invalid lookup field pvmctl <obj_name> delete -i abcd
    args = [obj_name, 'delete', '-i', 'abcd']
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                    c_wrps=c_wrps)

    test_obj.assertRaises(exc.InvalidIDArgumentError, cp.obj_type.execute)

    ew_delete_mock.assert_not_called()
    ew_delete_mock.reset_mock()

    # Try deleting with 0 wrappers
    args = [obj_name, 'delete', '-i', 'uuid=abcd']
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                    c_wrps=c_wrps)

    cp.obj_type.ec._set_wrappers = Mock(side_effect=_set_wrappers(cp))

    test_obj.assertRaises(exc.NoEntriesFoundError, cp.obj_type.execute)

    ew_delete_mock.assert_not_called()
    ew_delete_mock.reset_mock()

    # Try deleting with 2 wrappers
    args = [obj_name, 'delete', '-i', 'uuid=abcd']
    cp = setup_mock(test_obj, args, test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                    c_wrps=c_wrps)

    cp.obj_type.ec._set_wrappers = Mock(side_effect=_set_wrappers(
        cp, wrps=[test_obj.dwrap, test_obj.dwrap]))

    test_obj.assertRaises(exc.MultipleEntriesError, cp.obj_type.execute)

    ew_delete_mock.assert_not_called()
    ew_delete_mock.reset_mock()

    # pvmctl <obj_name> delete
    args = [obj_name, 'delete']

    test_obj.assertRaises(SystemExit, setup_mock, test_obj, args,
                          test_obj.adpt, wrps=wrps, p_wrps=p_wrps,
                          c_wrps=c_wrps)

    ew_delete_mock.assert_not_called()
    ew_delete_mock.reset_mock()


def _set_wrappers(cp, wrps=[], c_wrps=[], p_wrps=[]):
    cp.obj_type.ec.wrappers = wrps
    cp.obj_type.ec.elem_p_wrappers = p_wrps
    cp.obj_type.ec.elem_c_wrappers = c_wrps


def _setup_mock_msys(adpt, msys_mock_file):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile(msys_mock_file),
                           adapter=adpt).get_response()
    return msys_w.System.wrap(resp)[0]


def setup_mock(test_obj, args, adpt_fx, wrps=[], c_wrps=[], p_wrps=[],
               help_out=sys.stdout, proc_ret_code=0,
               msys_mock_file='managedsystem.txt'):
    ''' Setup a mocked ObjectType based on the provided arguments.

    :param args: The CLI arguments to be processed.
    :returns: The CLIParser object for this execution test attempt.
    '''
    cp = cli_parser.CLIParser(args, help_out=help_out)

    util.lpar_cache = {}
    util.vios_cache = {}
    util.vswitch_cache = {}
    util.name_cache = {}

    test_obj.assertEqual(cp.process(auto_exec=False), proc_ret_code)

    if cp.obj_type.sess_man:
        cp.obj_type.sess_man.create_session_adapter = Mock()
        cp.obj_type.sess_man._session_adapter = Mock()

    if cp.obj_type.ec:
        cp.obj_type.ec.parent_wrp_to_children = {}
        cp.obj_type.ec.sm.adapter.read = Mock(return_value=test_obj.resp)
        cp.obj_type.ec._set_wrappers = Mock(side_effect=_set_wrappers(
            cp, wrps=wrps, c_wrps=c_wrps, p_wrps=p_wrps))
        cp.obj_type.ec._add_child_wrappers = Mock(side_effect=_set_wrappers(
            cp, wrps=wrps, c_wrps=c_wrps, p_wrps=p_wrps))
        cp.obj_type.ec._add_parent_wrappers = Mock(side_effect=_set_wrappers(
            cp, wrps=wrps, c_wrps=c_wrps, p_wrps=p_wrps))
        cp.obj_type.ec.sm.adapter.update = Mock(return_value=test_obj.resp)
        cp.obj_type.ec.find_managed_system = Mock(
            return_value=_setup_mock_msys(adpt_fx, msys_mock_file))

    cp.obj_type.cli_out = six.StringIO()
    cp.obj_type.cli_err = six.StringIO()

    return cp


def mock_vios(mv_adpt):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile('fake_vios_mappings.txt'),
                           adapter=mv_adpt).get_response()
    return vios_w.VIOS.wrap(resp)


def mock_lpar(ml_adpt):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile('lpar.txt'),
                           adapter=ml_adpt).get_response()
    return lpar_w.LPAR.wrap(resp)[0]


def mock_vnic(ml_adpt):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile('vnic_feed.txt'),
                           adapter=ml_adpt).get_response()
    return iocard.VNIC.wrap(resp)[0]


def mock_sys_with_sriov(m_adpt):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile('sys_with_sriov.txt'),
                           adapter=m_adpt).get_response()
    return msys_w.System.wrap(resp)[0]


def mock_sys(m_adpt):
    resp = pvmhttp.PVMResp(pvmfile=pvmhttp.PVMFile('managedsystem.txt'),
                           adapter=m_adpt).get_response()
    return msys_w.System.wrap(resp)[0]
