# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


from pypowervm.exceptions import AbstractMsgFmtError
from pvmcli.util.i18n import _


''' Return codes
    0: OK
    1: General error
    2: argparse provided generic error
    3: Keyboard interruption (e.g. user pressed ctrl+c)
    40 - 63: Predefined HTTP Error
    254: Unmapped HTTP Error
'''
OK_RETURN_CODE = 0
GENERAL_ERROR_RETURN_CODE = 1
ARGPARSE_ERR_RET_CODE = 2
KEYBOARD_INTTERUPT_ERR_RET_CODE = 3
HTTP_ERROR_RET_CODE_MAP = {400: 40,
                           401: 41,
                           402: 42,
                           403: 43,
                           404: 44,
                           405: 45,
                           406: 46,
                           407: 47,
                           408: 48,
                           409: 49,
                           410: 50,
                           411: 51,
                           412: 52,
                           413: 53,
                           414: 54,
                           415: 55,
                           416: 56,
                           417: 57,
                           500: 58,
                           501: 59,
                           502: 60,
                           503: 61,
                           504: 62,
                           505: 63}
HTTP_UNMAPPED_RETURN_CODE = 254


def http_err_ret_code(http_error):
    ''' Define error codes for HTTP errors here.

    :param http_error: HTTPError object that caused the error
    '''
    # TODO (clbush): Help text for all error return codes
    if http_error.response.status in HTTP_ERROR_RET_CODE_MAP:
        return HTTP_ERROR_RET_CODE_MAP[http_error.response.status]
    else:
        return HTTP_UNMAPPED_RETURN_CODE


class CLIAbstractMsgFmtError(AbstractMsgFmtError):
    ''' Used for CLI Exceptions. Includes a return code to send to the user.

        Extends the AbstractMsgFmtError, so all subclasses must set the
        msg_fmt class variable. The consumer should instantiate the subclass
        with **kwargs appropriate to its msg_fmt.
    '''
    def __init__(self, response=None, ret_code=GENERAL_ERROR_RETURN_CODE,
                 **kwargs):
        super(CLIAbstractMsgFmtError, self).__init__(response=response,
                                                     **kwargs)
        self.ret_code = ret_code


class InvalidSetFieldElemNotFoundError(CLIAbstractMsgFmtError):
    ''' Used for instances where an element may not exist on a retrieved
        payload, but a user can attempt to update it anyway.
    '''
    msg_fmt = _("Element with schema type %(schema)s does not exist on "
                "retrieved object type %(obj_type)s.")


class ObjectTypeValidationError(CLIAbstractMsgFmtError):
    msg_fmt = _("Validation failed for %(func_name)s during %(object_type)s"
                " processing.")


class NoEntriesFoundError(CLIAbstractMsgFmtError):
    # TODO(clbush): Don't error here, need return code of 0, no stdout
    def __init__(self, ret_code=OK_RETURN_CODE, **kwargs):
        super(NoEntriesFoundError, self).__init__(
            response=None, ret_code=ret_code, **kwargs)
    msg_fmt = _("Failed to find any %(object_type)s entries.")


class NetBridgeDeleteNoEntriesFoundError(CLIAbstractMsgFmtError):
    msg_fmt = _("No entries of type Network Bridge with PVID %(pvid)d.")


class LPARNoEntriesFoundError(CLIAbstractMsgFmtError):
    # TODO (clbush): Combine this and similar VIOS error into 1 error.
    msg_fmt = _("No entries of type Logical Partition with Name %(name)s.")


class LPARMultipleEntriesFoundError(CLIAbstractMsgFmtError):
    # TODO (clbush): Combine this and similar VIOS error into 1 error.
    # This will probably never happen... if it does, something went terribly
    # wrong.
    msg_fmt = _("Multiple entries of type Logical Partition with "
                "Name %(name)s.")


class VIOSNoEntriesFoundError(CLIAbstractMsgFmtError):
    # TODO (clbush): Combine this and similar LPAR error into 1 error.
    msg_fmt = _("No entries of type Logical Partition with Name %(name)s.")


class VIOSMultipleEntriesFoundError(CLIAbstractMsgFmtError):
    # TODO (clbush): Combine this and similar LPAR error into 1 error.
    # This will probably never happen... if it does, something went terribly
    # wrong.
    msg_fmt = _("Multiple entries of type Virtual I/O Server with Name "
                "%(name)s.")


class UpdateProcessFailedError(CLIAbstractMsgFmtError):
    msg_fmt = _("Update failed on %(object_type)s. No response retrieved.")


class MultipleEntriesError(CLIAbstractMsgFmtError):
    msg_fmt = _("Cannot %(operation)s multiple entries simultaneously.")


class NoFoundDisksError(CLIAbstractMsgFmtError):
    msg_fmt = _("Couldn't find %(storage_type)s with %(stor_id_val)s "
                "%(stor_id_type)s.")


class MultipleFoundDisksError(CLIAbstractMsgFmtError):
    msg_fmt = _("Found multiple %(storage_type)s entries with a "
                "%(stor_id_type)s of %(stor_id_val)s. Narrow your search by "
                "identifying volume groups (--vg) or shared storage pools "
                "(--ssp) in your command, or pass the UDID to the --stor-id "
                "argument.")


class MultipleEntriesOnParentError(CLIAbstractMsgFmtError):
    msg_fmt = _("Cannot %(operation)s %(c_type)s entry on multiple "
                "%(p_type)s entries simultaneously.")


class InvalidSetFieldsError(CLIAbstractMsgFmtError):
    msg_fmt = _("Must specify at least one valid field to set for "
                "%(operation)s operation on %(object_type)s.")


class DeleteProcessFailedError(CLIAbstractMsgFmtError):
    msg_fmt = _("Delete failed on %(object_type)s. No response retrieved.")


class UUIDMustBeSetError(CLIAbstractMsgFmtError):
    msg_fmt = _("UUID must be specified to perform %(operation)s "
                "on %(object_type)s.")


class ParentMustBeSetError(CLIAbstractMsgFmtError):
    msg_fmt = _("Parent Id must be specified to perform operations on "
                "%(object_type)s.")


class ChildMustBeSetError(CLIAbstractMsgFmtError):
    msg_fmt = _("Child Id must be specified to perform operations on "
                "%(object_type)s.")


class AuthenticationError(CLIAbstractMsgFmtError):
    msg_fmt = _("Authentication failure, please specify the %(auth_arg)s. "
                "command line option.")


class CreateProcessFailedError(CLIAbstractMsgFmtError):
    msg_fmt = _("Create failed on %(object_type)s.")


class MissingArgumentError(CLIAbstractMsgFmtError):
    msg_fmt = _("Argument %(arg)s is required.")


class LparCreateRequiresProcCfgError(CLIAbstractMsgFmtError):
    msg_fmt = _("Logical Partition creation requires a valid "
                "processor configuration.")


class UUIDLookupFailedError(CLIAbstractMsgFmtError):
    msg_fmt = _("Looking up UUID on %(object_type)s failed.")


class VIOSAdaptersPairError(CLIAbstractMsgFmtError):
    msg_fmt = _("Every Virtual I/O Server specified must also identify "
                "the backing physical Ethernet adapter.")


class VIOSAdaptersQuantityError(CLIAbstractMsgFmtError):
    msg_fmt = _("A NetworkBridge supports a maximum of two Virtual "
                "I/O Server / Device pairs.")


class InvalidVIOSNameError(CLIAbstractMsgFmtError):
    msg_fmt = _("Virtual I/O Server with name %(name)s was not found.")


class AmbiguousFieldNameError(CLIAbstractMsgFmtError):
    msg_fmt = _("Ambiguous display field name: '%(given_field)s' could match "
                "any of %(matches)s.")


class CreateIOSlotOnPartitionError(CLIAbstractMsgFmtError):
    msg_fmt = _("Creation of I/O Slot with name %(name)s failed. If this was "
                "part of a partition create operation, your partition was "
                "most likely created. Please check the provided name and try "
                "creating the I/O Slot again using the IOSlot create "
                "operation.")


class UploadLVRequiresVGError(CLIAbstractMsgFmtError):
    msg_fmt = _("Uploading a file to a Logical Volume requires a Volume Group "
                "to be specified with the --vg argument.")


class UploadLURequiresSSPError(CLIAbstractMsgFmtError):
    msg_fmt = _("Uploading a file to a Logical Unit requires a Shared Storage "
                "Pool to be specified with the --ssp argument.")


class PriOutputMinLengthError(CLIAbstractMsgFmtError):
    msg_fmt = _("The minimum length of the output table must be 80.")


class InvalidDrcNameError(CLIAbstractMsgFmtError):
    msg_fmt = _("No IO slots were found for drc name %(name)s.")


class InvalidSetFieldFormatError(CLIAbstractMsgFmtError):
    msg_fmt = _("Set field %(field)s has an improper format. Please specify "
                "the set field as <Field>=<Value>. Run the command 'pvmctl "
                "%(obj_type)s update help' for a list of valid set fields.")


class InvalidSetFieldValueError(CLIAbstractMsgFmtError):
    msg_fmt = _("Set field %(field)s has an improper value specified. Please "
                "check your arguments and try again.")


class InvalidIDArgumentError(CLIAbstractMsgFmtError):
    msg_fmt = _("Specified identifier argument %(id_arg)s is invalid. ID "
                "arguments must be of the format <Field>=<Value>. Valid "
                "Fields: %(valid_fields)s")


class InvalidSplitArgError(CLIAbstractMsgFmtError):
    msg_fmt = _("Search argument %(arg)s is invalid. Search arguments "
                "should be of the format <Field>=<Value>.")


class NoKeysFoundError(CLIAbstractMsgFmtError):
    def __init__(self, response=None, **kwargs):
        super(NoKeysFoundError, self).__init__(
            response=response, ret_code=OK_RETURN_CODE, kwargs=kwargs)
    msg_fmt = _("No keys were found.")


class EmptyKeyError(CLIAbstractMsgFmtError):
    msg_fmt = _("Authorized keys cannot be empty.")


class UploadStdInRequiresSizeError(CLIAbstractMsgFmtError):
    msg_fmt = _("Uploading from stdin requires --data-size to be specified.")


class NoRunningVIOSFoundError(CLIAbstractMsgFmtError):
    msg_fmt = _("No running VIOS with an active RMC state was found.")


class InvalidKeyValue(CLIAbstractMsgFmtError):
    msg_fmt = _("Parameters to the %(arg)s argument must be of the "
                "format <Key>=<Value>.")


class InvalidRESTLogType(CLIAbstractMsgFmtError):
    msg_fmt = _("Invalid REST log type '%(logtype)s'.  Valid types: "
                "%(valid)s")


class InvalidRESTLogLevel(CLIAbstractMsgFmtError):
    msg_fmt = _("Invalid REST log level '%(level)s'.  Valid levels: "
                "%(valid)s")


class NoFreeSPPError(CLIAbstractMsgFmtError):
    msg_fmt = _("All shared processor pools are currently in use.")


class SpecifiedSPPNotFreeError(CLIAbstractMsgFmtError):
    msg_fmt = _("The specified shared processor pool with ID %(spp_id)d is "
                "currently in use.")


class InvalidSharingModeError(CLIAbstractMsgFmtError):
    msg_fmt = _("Invalid sharing mode for processor type `%(proc_type)s`. "
                "Valid sharing modes: %(shr_modes)s")


class PPortUpdateRequiresLocError(CLIAbstractMsgFmtError):
    msg_fmt = _("--port-loc argument required to update a physical port.")


class CapacityOutOfBoundsError(CLIAbstractMsgFmtError):
    msg_fmt = _("--%(arg)s argument must be between 0.00 and 100.00.")


class NoPPortFoundFromLocError(CLIAbstractMsgFmtError):
    msg_fmt = _("No physical ports with location code %(loc)s found.")


class InvalidBackDevError(CLIAbstractMsgFmtError):
    '''
    Invalid Backing Device format error.
    spec_err is used to provide specific information as to why the backing
    device is invalid.
    '''
    msg_fmt = _("%(spec_err)s The specified backing device "
                "strings '%(back_dev_str)s' are invalid. If the VIOS name "
                "contains spaces, make sure it is passed in surrounded "
                "with \". All backing device strings must meet the following "
                "format: vios=<VIOS name> loc=<SR-IOV Phys Port Loc Code> "
                "[cap=<capacity percentage value>] [pri=<priority failover "
                "integer value]. Example: --back-dev vios=vios1 "
                "loc=U78CB.001.WZS0485-P1-C5-T3 cap=42.42%% pri=10.")


class MappingNoEntriesModifiedError(CLIAbstractMsgFmtError):
    msg_fmt = _("No %(type)s mappings were modified. Please check your "
                "provided args with `pvmctl %(type)s list` and try again")


class MappingNoEntriesDeletedError(CLIAbstractMsgFmtError):
    msg_fmt = _("No %(type)s mappings were deleted. Please check your "
                "provided args with `pvmctl %(type)s list` and try again.")


class MappingTypedDeleteRequiresLPAR(CLIAbstractMsgFmtError):
    msg_fmt = _("--lpar argument is required for all scsi delete operations "
                "with a provided --type.")


class SingleMappingEntryRemapError(CLIAbstractMsgFmtError):
    msg_fmt = _("Unable to remap %(type)s mapping. Expected to find exactly "
                "one matching mapping, found %(num_mappings)d.")


class SingleMappingEntryDeleteError(CLIAbstractMsgFmtError):
    msg_fmt = _("Unable to delete %(type)s mapping. Expected to find exactly "
                "one matching mapping, found %(num_mappings)d.")


class SCSIInvalidStorIDPrefixError(CLIAbstractMsgFmtError):
    msg_fmt = _("Invalid prefix specified to --stor-id. Expected: %(exp_pre)s")


class VNICBackDevUpdateLocError(CLIAbstractMsgFmtError):
    msg_fmt = _("Error finding physical ports with provided --loc argument "
                "%(loc)s.")


class VNICBackDevLocRequiredError(CLIAbstractMsgFmtError):
    msg_fmt = _("To update a VNIC backing device, --loc is required.")


class SCSITypeDeleteRequiresStorIdError(CLIAbstractMsgFmtError):
    msg_fmt = _("VirtualSCSIMapping delete with a --type argument requires "
                "a user provided --stor-id argument.")


class NoPFCPortFoundError(CLIAbstractMsgFmtError):
    msg_fmt = _("Couldn't find pfc port with %(arg_type)s %(port_arg)s on "
                "VIOS %(vios_name)s.")
