# Licensed Materials - Property of IBM
#
# COPYRIGHT International Business Machines Corp. 2015, 2017
# All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.


import logging
import re
import os
import six
import sys

import pvmcli.util.exceptions as exc
import pvmcli.util.op_generator as op_gen
import pvmcli.util.util as util

import pypowervm.wrappers.entry_wrapper as ew


LOG = logging.getLogger('pvmctl')


def hex_no_prefix(value):
    ''' Convert specified value to hexadecimal, dropping the 0x prefix. '''
    return format(int(value), 'x')


def four_char_hex(value):
    ''' Convert the specified value to hexadecimal, and format to 4 chars. '''
    return "%04X" % value


def two_char_hex(value):
    ''' Convert the specified value to hexadecimal, and format to 2 chars. '''
    return "%02X" % value


def process_http_error(obj_type, e):
    ''' Write HTTP Response error status and reason to stderr.

    :param e: The HttpError object to process.
    '''
    cli_err = sys.stderr
    if obj_type:
        cli_err = obj_type.cli_err
    util.print_(six.text_type(e), outfile=cli_err)


def is_field_displayable(wrp, attr):
    ''' Determine if the specified attribute is displayable for a wrapper.

    :param wrp: The wrapper that contains the attribute to display
    :param attr: The string name of the attribute on the wrapper to display
    '''
    return ((attr not in op_gen.NON_USER_WRAPPER_FIELD_NAMES
             or (wrp.schema_type in op_gen.FIELD_OVERRIDES
                 and attr in op_gen.FIELD_OVERRIDES[wrp.schema_type]))
            and not attr.startswith('_'))


def raw_xml_display(obj_type):
    ''' Output a newline delimited string of xml output for each wrapper.

    Example input:
        pvmctl ManagedSystem list --xml

    :param obj_type: The ObjectType being processed for output.
    '''
    # Use a StringIO buffer stream because this output can become massive.
    xml_str_io = six.StringIO()

    for xml_str in [wrp.toxmlstring() for wrp in obj_type.ec.wrappers]:
        xml_str_io.write("%(value)s%(nl)s" %
                         {'value': xml_str,
                          'nl': os.linesep})
    util.print_(xml_str_io.getvalue(), outfile=obj_type.cli_out)

    xml_str_io.close()


def where_maps(where_clauses):
    ''' Get maps from display field name to exclusion/inclusion values.

    :param where_clauses: User specified command line where clauses.'''
    equal_map = {}
    not_equal_map = {}

    if not where_clauses:
        return equal_map, not_equal_map

    for where_clause in where_clauses:
        split_clause = re.split('(\s*!?=\s*)', where_clause, 1)
        if split_clause[1].strip(' ') == '!=':
            not_equal_map[split_clause[0].strip(' ')] = split_clause[2]
        elif split_clause[1].strip(' ') == '=':
            equal_map[split_clause[0].strip(' ')] = split_clause[2]

    return equal_map, not_equal_map


def graph_conflict(out_graph, obj_type):
    ''' Determine if the specified output graph has a where conflict.

    :param out_graph: OutputGraph object to inspect.
    :param obj_type: The object type for this execution.
    '''
    where_clauses = obj_type.ec.parsed_args.where
    if not where_clauses:
        return False

    equal_map, not_equal_map = where_maps(where_clauses)
    for eq_df_name in equal_map:
        d_fields = obj_type.get_disp_fields()
        if eq_df_name not in d_fields:
            raise exc.InvalidIDArgumentError(
                id_arg=eq_df_name, valid_fields=", ".join(d_fields))
        df = obj_type.get_disp_fields()[eq_df_name]
        if (eq_df_name in out_graph.attr_vals
                and equal_map[eq_df_name] !=
                six.text_type(df.out_func(out_graph.attr_vals[eq_df_name]))):
                    return True

    for ne_df_name in not_equal_map:
        df = obj_type.get_disp_fields()[ne_df_name]
        if (ne_df_name in out_graph.attr_vals
                and not_equal_map[ne_df_name] ==
                six.text_type(df.out_func(out_graph.attr_vals[ne_df_name]))):
                    return True

    return False


def _gen_header_strs(wrp_name, indent_level=0):
    out_strs = []
    out_strs.append("%(in)s%(ds)s" % {'in': '  ' * indent_level,
                                      'ds': ((len(wrp_name) + 4) * '-')})
    out_strs.append("%(in)s| %(schema)s |" % {'in': '  ' * indent_level,
                                              'schema': wrp_name})
    out_strs.append("%(in)s%(ds)s" % {'in': '  ' * indent_level,
                                      'ds': ((len(wrp_name) + 4) * '-')})

    return out_strs


def _all_disp_graph_out(obj_type, out_graph, out_strs, indent_level=0):
    ''' Recursively build the output string for a specified output graph.

    :param obj_type: ObjectType containing user arguments.
    :param out_graph: OutputGraph object for this output.
    :param out_strs: List of strings to output, built recursively.
    :param indent_level: Current indent level of the string being built.
    :returns: False if where clauses on the provided obj_type have a conflict,
              otherwise the list of --all-out output strings. '''

    # Base case: If there is no output graph to process, return out_strs
    if not out_graph:
        return out_strs

    # If we find a conflict, return False. This is only seen mid-recursion so
    # we avoid adding an entire wrapper if something where-clause related
    # conflicts.
    if graph_conflict(out_graph, obj_type):
        return False

    # Generate the header for this graph's output.
    out_strs.extend(_gen_header_strs(out_graph.sch_type,
                                     indent_level=indent_level))
    # Indent one to align label/value text with the text in a header
    indent_level += 1

    # Walk all attribute values for this output graph
    for out_val in sorted(out_graph.attr_vals):
        if out_val in obj_type.get_disp_fields():
            d_field = obj_type.get_disp_fields()[out_val]
            hide_label = obj_type.ec.parsed_args.hide_label
            label_char = obj_type.ec.parsed_args.label_char
            if not isinstance(out_graph.attr_vals[out_val], ew.Wrapper):
                out_strs.append(
                    "%(indent)s%(label)s%(value)s" %
                    {'indent': '  ' * indent_level,
                     'label': '' if hide_label else d_field.label + label_char,
                     'value': d_field.out_func(out_graph.attr_vals[out_val])})

    # Walk all child graphs, building up the output strings based
    for child_graph in out_graph.child_graphs:
        out_strs = _all_disp_graph_out(obj_type, child_graph, out_strs,
                                       indent_level)
        # If any recursive step comes back with a where conflict, break
        # recursion by returning False
        if out_strs is False:
            return False

    return out_strs


def raw_all_display(obj_type):
    ''' Output function for ObjectType list operation with --all-out argument.

    Example input:
        pvmctl ManagedSystem list --all-out

    Example output:
        -----------------
        | ManagedSystem |
        -----------------
          system_name: value
          ...
          -----------------------------------
          | MachineTypeModelAndSerialNumber |
          -----------------------------------
            machine_type: value
            model: value
            mtms_str: value
            serial: value


    :param obj_type: The ObjectType being processed for output.
    '''
    all_out_strs = []

    for wrp in obj_type.ec.wrappers:
        out_graph = OutputGraph(obj_type, wrp)
        out_strs = _all_disp_graph_out(obj_type, out_graph, out_strs=[])
        # out_strs will be False if there is a where clause conflict
        if out_strs is not False:
            all_out_strs.extend(out_strs)
    util.print_(os.linesep.join(all_out_strs), outfile=obj_type.cli_out)


def df_sch_type(disp_field_name):
    return disp_field_name.split('.', 1)[0]


sch_dfs = {}


def _get_dfs_for_sch_type(sch_type, obj_type):
    # If we've already built up the fields for this sch_type, return from cache
    if sch_type in sch_dfs:
        return sch_dfs[sch_type]

    sch_dfs[sch_type] = []
    # Walk all display fields on given object type
    for potential_df in obj_type.get_disp_fields():
        # If the display field name starts with this schema type
        if potential_df.startswith(sch_type + '.'):
            # Add it to the schema/display field cache map
            sch_dfs[sch_type].append(potential_df)

    return sch_dfs[sch_type]


def _is_wrapper(out_val, df_name):
    return out_val and df_name in out_val.attr_vals and isinstance(
        out_val.attr_vals[df_name], ew.Wrapper)


def _get_display_vals(d_field, out_val, df_name):
    df_vals = []
    # If our display field value is a Wrapper, we need to output all children
    if _is_wrapper(out_val, df_name):
        # Walk all display fields on this wrapper
        ov_sch = out_val.attr_vals[df_name].schema_type
        for w_df in sorted(_get_dfs_for_sch_type(ov_sch, out_val.obj_type)):
            # Get the output graph for this display field
            o_ls = out_val.get_spec_disp_lists([w_df], None)
            # Get the value for this display field
            d_val = out_val.obj_type.dfield_val(out_val.attr_vals[df_name],
                                                w_df)
            # TODO (clbush): Issue involving some specified display
            #                lists returning as [[None]].
            #                Example: proc_config.sharing_mode
            # If the output value is not a wrapper itself, include
            # TODO (clbush): This is failing for PartitionMemoryConfiguration.
            #                ppt_ratio in LPAR unit tests. Had to add a
            #                display field existence check (w_df in...)
            #                to get passing. In production, seems to cause
            #                no issue.
            if not o_ls[0][0] or (
                    w_df in o_ls[0][0].attr_vals and not isinstance(
                        o_ls[0][0].attr_vals[w_df], ew.Wrapper)):
                d_lb = out_val.obj_type.get_disp_fields()[w_df].label
                df_vals.append((d_lb, d_val))
    else:
        df_out_val = (d_field.out_func(out_val.attr_vals[df_name])
                      if out_val is not None and df_name in out_val.attr_vals
                      else '')
        df_vals.append((d_field.label, df_out_val))

    return df_vals


def specified_display(obj_type):
    ''' Output function for ObjectType list operation with user specified
        --display-fields argument.

    Example input:
        pvmctl ManagedSystem list --display_fields ManagedSystem.system_name \
        IOSlot.dyn_reconfig_conn_index

    Example output:
        system_name=name1,dyn_reconfig_conn_index=index1
        system_name=name1,dyn_reconfig_conn_index=index2
        ...
        system_name=name1,dyn_reconfig_conn_index=indexN

    :param obj_type: The ObjectType being processed for output.
    '''
    out_list = []

    labelchar = obj_type.ec.parsed_args.label_char
    delim = obj_type.ec.parsed_args.field_delim
    separator = obj_type.ec.parsed_args.record_delim
    hide_label = obj_type.ec.parsed_args.hide_label

    out_records = []
    for wrp in obj_type.ec.wrappers:
        out_graph = OutputGraph(obj_type, wrp)

        out_lists = out_graph.get_spec_disp_lists(
            obj_type.ec.parsed_args.display_fields, [])

        for out_list in out_lists:
            has_conflict = False
            for out_gph in out_list:
                if out_gph and out_gph.has_conflict():
                    has_conflict = True

            if not has_conflict:
                out_vals = []
                df_names = obj_type.ec.parsed_args.display_fields
                for out_val, df in zip(out_list, df_names):
                    d_field = obj_type.get_disp_fields()[df]
                    for d_lb, d_val in _get_display_vals(d_field, out_val, df):
                        out_vals.append(
                            "%(lb)s%(value)s"
                            % {'lb': ('' if hide_label
                                      else d_lb + labelchar),
                               'value': d_val})
                out_records.append(delim.join(out_vals))
    util.print_(separator.join(out_records), outfile=obj_type.cli_out)


def default_display(obj_type):
    ''' Non-specialized list operation function for pvmctl <ObjectType> list.

    :param obj_type: The ObjectType being processed for output.
    '''
    # TODO (clbush): Prioritized output per each ObjectType
    #                Remove this conditional and add abstract checks to
    #                ObjectType once every type has pri_output defined
    if 'pri_output' in dir(obj_type):
        util.print_(obj_type.pri_output(), outfile=obj_type.cli_out)
    else:
        raw_all_display(obj_type)


class OutputGraph(object):
    ''' A bidirected graph representation of a pypowervm wrapper object.

    This is used to drive output and filter displayed values based on
    specified display fields and where clauses. Architecturally, a wrapper
    can contain any number of atomic fields, child wrappers, or list/tuple of
    child wrappers. This class builds a graph that considers wrappers as nodes.
    These nodes are then used as the records when doing any output processing.
    '''
    sch_to_out_graph = {}

    def __init__(self, obj_type, wrp, parent_graph=None, init_cls_sch=True):
        if init_cls_sch:
            OutputGraph.sch_to_out_graph = {}
        self.obj_type = obj_type
        self.wrp = wrp
        self.parent_graph = parent_graph
        self.attr_vals = {}
        self.child_graphs = []

        if (obj_type.get_schema(wrp.schema_type)
                in self.sch_to_out_graph):
                    self.sch_to_out_graph[obj_type.get_schema(
                        wrp.schema_type)].append(self)
        else:
            self.sch_to_out_graph[obj_type.get_schema(
                wrp.schema_type)] = [self]

        for attr in (a for a in dir(wrp) if is_field_displayable(wrp, a)):
            f_name = ("%(schema)s.%(attrname)s" %
                      {'schema': obj_type.get_schema(wrp.schema_type),
                       'attrname': attr})
            if f_name not in obj_type.exc_disp_fields:
                attr_val = getattr(wrp, attr)
                # Walk all atomic attrs
                if isinstance(attr_val, op_gen.atomic_f_types):
                    # Generate field name for this attr
                    self.attr_vals[f_name] = attr_val
                elif (isinstance(attr_val, ew.Wrapper)):
                    self.attr_vals[f_name] = attr_val
                    self.child_graphs.append(OutputGraph(
                        obj_type, attr_val,
                        parent_graph=self, init_cls_sch=False))
                elif isinstance(attr_val, (list, tuple)):
                    if attr_val:
                        for attr_wrp in attr_val:
                            if isinstance(attr_wrp, ew.Wrapper):
                                self.child_graphs.append(
                                    OutputGraph(obj_type, attr_wrp,
                                                parent_graph=self,
                                                init_cls_sch=False))
                            else:
                                w_st = obj_type.get_schema(wrp.schema_type)
                                f_name = (
                                    "%(sch)s.%(aname)s" %
                                    {'sch': w_st, 'aname': attr})
                                self.attr_vals[f_name] = attr_val

        self.where_conflict = graph_conflict(self, obj_type)

    @property
    def connected_graphs(self):
        ''' Return a list of all connected graphs (child and parents). '''
        conn_graphs = self.child_graphs[:]
        if self.parent_graph:
            conn_graphs.append(self.parent_graph)
        return conn_graphs

    @property
    def all_child_graphs(self):
        ''' Return a list of all child graphs at all levels. '''
        if not self.child_graphs:
            return []

        child_graph_bldr = self.child_graphs[:]
        for child_graph in self.child_graphs:
            child_graph_bldr.extend(child_graph.all_child_graphs)

        return child_graph_bldr

    @property
    def all_parent_graphs(self):
        ''' Return a list of parent graphs at all levels. '''
        if not self.parent_graph:
            return []

        parent_graph_bldr = [self.parent_graph]
        parent_graph_bldr.extend(self.parent_graph.all_parent_graphs)
        return parent_graph_bldr

    def has_conflict(self):
        ''' Check for any where clause conflicts with this graph node.

        Will return True if this node or a parent graph node conflicts. Will
        also return True if the where clause contains a must equal clause and
        there is no graph to match that schema type.
        '''
        curr_graph = self

        while curr_graph.parent_graph is not None:
            if curr_graph.where_conflict:
                return True
            curr_graph = curr_graph.parent_graph

        # Check for None case
        eq_where_map = where_maps(self.obj_type.ec.parsed_args.where)[0]
        for eq_where_df in eq_where_map:
            if df_sch_type(eq_where_df) not in self.sch_to_out_graph:
                return True

        eq_map, ne_map = where_maps(self.obj_type.ec.parsed_args.where)
        for eq_df in eq_map:
            for assoc_graph in self.get_assoc_graphs(eq_df, [self]):
                if assoc_graph.where_conflict:
                    return True
        for ne_df in ne_map:
            for assoc_graph in self.get_assoc_graphs(ne_df, [self]):
                if assoc_graph.where_conflict:
                    return True

        return self.where_conflict

    def get_spec_disp_lists(self, d_fields, root_graphs):
        ''' Get all graphs for the specified display fields.

            This will walk the output graph, returning a list of lists of
            connected graph nodes given those specified display fields.

        :param d_fields: Display fields to output
        :param root_graphs: Current root graphs to consider when determining
                            attached graphs. Primarily used for the recursive
                            step of this method.
        '''
        # If we have no display fields, return whatever the root graphs are.
        if not d_fields:
            return root_graphs

        # If the root graphs are empty, we need to seed the values we will be
        # using to build out the output lists
        if not root_graphs:
            root_graphs = []
            df_root = d_fields[0]
            # Build out a list of lists of all root display field nodes
            if df_sch_type(df_root) in self.sch_to_out_graph:
                for og in self.sch_to_out_graph[df_sch_type(df_root)]:
                    root_graphs.append([og])
            else:
                # If we don't have any of the root display fields in this
                # graph, this level of the display fields is None
                root_graphs = [[None]]

            # Build the remaining lists based on the remaining display fields
            rem_disp_fields = []
            if len(d_fields) > 1:
                rem_disp_fields = d_fields[1:]
            return self.get_spec_disp_lists(rem_disp_fields, root_graphs)

        # If we have root graphs to append to, build up the display list values
        # based on their associated graphs
        new_graphs = []
        for curr_graph in root_graphs:
            assoc_graphs = self.get_assoc_graphs(d_fields[0], curr_graph[:])

            if not assoc_graphs:
                new_graph = curr_graph[:]
                new_graph.append(None)
                new_graphs.append(new_graph)
            else:
                for a_graph in assoc_graphs:
                    new_graph = curr_graph[:]
                    new_graph.append(a_graph)
                    new_graphs.append(new_graph)

        # If we have display fields left to process, recurse with the new
        # graph list and a trimmed display field list
        if len(d_fields) > 1:
            return self.get_spec_disp_lists(d_fields[1:], new_graphs)
        else:
            return new_graphs

    def _check_assoc_graph(self, graphs_to_find, visited_graphs=None):
        if not visited_graphs:
            visited_graphs = []

        visited_graphs.append(self)
        visited_sch_types = set([vg.sch_type for vg in visited_graphs])

        if self in graphs_to_find:
            graphs_to_find.remove(self)

        if not graphs_to_find:
            return True

        for graph_to_check in self.connected_graphs:
            if (graph_to_check not in visited_graphs
                    and graph_to_check.sch_type not in visited_sch_types):
                        graph_to_check._check_assoc_graph(graphs_to_find,
                                                          visited_graphs[:])

        return False if graphs_to_find else True

    def get_assoc_graphs(self, d_field, curr_graphs):
        ''' Generate all graphs associated with this one based on the given
        display field and the graphs we are associating it with.

        :param d_field: Display field name we are associating.
        :param curr_graphs: A list of OutputGraph objects that the d_field
                            should be associated with.
        '''
        if not d_field:
            return None

        # Get list of all potential matches
        potential_graphs = []
        if df_sch_type(d_field) in self.sch_to_out_graph:
            potential_graphs = self.sch_to_out_graph[df_sch_type(d_field)]
        graph_matches = []
        for p_graph in potential_graphs:
            if p_graph._check_assoc_graph(set([g for g in curr_graphs if g])):
                graph_matches.append(p_graph)

        return graph_matches

    @property
    def sch_type(self):
        ''' Return the schema type for this graph's wrapper. '''
        return self.obj_type.get_schema(self.wrp.schema_type)
